/* Rutinas estadsticas:

	ini_random: inicializar generador de nmeros aleatorios.
	uniforme: generador de nmeros aleatorios [0, 1]
	qwantilae: hallar cuantilas de muestra aleatoria
	qwantilae_ordena: hallar cuantilas de muestra aleatoria (ordenndola)
			    Nota: algunos parametros de esta rutina -> double
	descriptiva: calcula los valores de media muestral y variancia
		muestral de x (tam. n)
	covarianza: calcula la covarianza muestral entre dos muestras x e y.
	correlacion: correlacion muestral entre dos muestras x e y.
	kolmogorov: hace una prueba de Kolmogorov-Smirnov entre una
		muestra de tam. n (vector x) y una distribucin por
		rectngulos, dada por los elementos:
			k: artculos (es "bloques" + 1)
			p: probabilidades (k-1) por bloque
			q: cuantilas (k)
		Devuelve los valores obs observado y nvl p-value.
		En esta y otras rutinas, las cuantilas no vienen dadas
		incrementalmente, sino en modo absoluto.
	rect_esperanza: esperanza para una distribucin rectangular.
	rect_variancia: variancia para una distribucin rectangular.
	rect_covarianza: covarianza para dos distribuciones rectangulares;
		precisa de la distribucin conjunta.
	F_distri: para calcular la funcion de distribucion de una 
		var rectangular.
		Nota: esta rutina ha sido pasada a tipo (double)
		para poder ser utilizada por `sim4'.  =24-I-95=
*/

#define MAX_ENTERO 2147483647
#define ABS(x) ((x)>0? (x): -(x))
#define REDOND(x) (int)(floor((double)(x)))

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <math.h>

float N_signif_Kolmo();
double correlacion();
float rect_esperanza();
float rect_variancia();
float rect_covarianza();
float rect_esperanza_();
float rect_variancia_();
float rect_covarianza_();
double F_distri();
double rect_esperanza_D();
double rect_variancia_D();

int fcompara(i,j)	/* rutina para qsort que ordena reales */
const void *i, *j;
{
	int f;
	float	*x, *y;

	x = (float *)i;
	y = (float *)j;
	f = (*x == *y)? 0: ((*x < *y)? -1: 1);
	return f;
}

f_ordena(vec, tam)
float	vec[];
int	tam;
{
	qsort(vec, tam, sizeof(float), fcompara);
}

/* ini_random()
{
	time_t	null = 13579;
	srand(time(&null));
}

double uniforme()
{
	return (double)rand()/MAX_ENTERO;
}
*/

ini_random()		/*	VERSION SUNIX */
{
	struct	timeval tp;
	FILE 	*in, *out;
	long	z;

	in = fopen("Semilla.2", "r");		/* Si existe un fichero */
	if (in) {				/* llamado 'Semilla.2'  */
		fscanf(in, "%ld", &z);		/* utiliza el valor que */
		fclose(in);			/* se encuentra alli y  */
		srand48(z);			/* retorna. */
		return;
	} 
	gettimeofday(&tp, NULL);		/* En otro caso, genera */
	z = tp.tv_sec+tp.tv_usec;		/* un valor nuevo con   */
	srand48(z);				/* la hora actual, y lo */
	out = fopen("Semilla", "w");		/* guarda en 'Semilla'  */
	fprintf(out, "%ld\n", z);		/* para uso futuro. */
	fclose(out);
}

double uniforme()
{
	double l = (double)lrand48();
	return l/MAX_ENTERO;
} 


/* Esta rutina se basa en otra escrita en fortran, "Kwantilae", de la 
   cual se diferencia adems por:
	el vector de trabajo toma memoria dinmicamente (no se debe
	destruir el orden del vector de muestra, ya que se rompe la 
	relacin con otras posibles variables)
	la versin de fortran trabajaba en la base de que las probab.
	por bloque eran iguales
*/

qwantilae(z, n, q, p, k)
float	z[], q[], p[];
int	n, k;
{
	int	i, j;
	float	y, h, alfa, *x;

	x = (float *) malloc(n*sizeof(float));
	for (i=0; i<n; i++) x[i] = z[i];
	f_ordena(x,n);

	q[0] = x[0];
	q[k-1] = x[n-1];
	for  (i=0, y=0.0; i<k-2; i++) {
         	y += p[i];
         	h = (n-1)*y;
         	j = REDOND(h-0.5);
         	alfa = h-j;
         	q[i+1] = (1.0-alfa)*x[j] + alfa*x[j+1];
	}
	free(x);
}

qwantilae_ordena(x, n, q, p, k)
float	x[];
double 	q[], p[];
int	n, k;
{
	int	i, j;
	double	y, h, alfa;

	f_ordena(x,n);

	q[0] = x[0];
	q[k-1] = x[n-1];
	for  (i=0, y=0.0; i<k-2; i++) {
         	y += p[i];
         	h = (n-1)*y;
         	j = REDOND(h-0.5);
         	alfa = h-j;
         	q[i+1] = (1.0-alfa)*x[j] + alfa*x[j+1];
	}
}

descriptiva(x, n, m, v)
float	x[], *m, *v;
int	n;
{
	double	sc, t, d;
	int	j;

	t = (double)x[0];
	sc = (double)0.0;
	for (j=1; j<n; j++) {
		t += (double)x[j];
		d = (double)((j+1)*x[j] - t);
		sc += (double)d*d/j/(j+1);
	}
	*m = (float)t/n;
	*v = (float)sc/(n-1);
}

descriptiva_2(x, n, m, v)	/* Esta rutina se duplica para `sim4' */
double	x[], *m, *v;		/* ya que hay muchas otras que utilizan */
int	n;			/* la version (float).   =25-I-95=    */
{
	double	sc, t, d;
	int	j;

	t = x[0];
	sc = (double)0.0;
	for (j=1; j<n; j++) {
		t += x[j];
		d = (j+1)*x[j] - t;
		sc += d*d/j/(j+1);
	}
	*m = t/n;
	*v = sc/(n-1);
}

float covarianza(x, y, n)
float	x[], y[];
int	n;
{
	/* Sxy = SUMA { (xi - x_) (yi - y_) } / (n-1) 
	   = SUMA {xi yi} / (n-1) - SUMA {xi} SUMA {yi} /(n-1)/n */
	int i;
	float sx, sy, sprod, cv;

	sx = sy = sprod = 0.0;
	for (i=0; i<n; i++) {
		sx += x[i];
		sy += y[i];
		sprod += x[i]*y[i];
	}
	cv = ( sprod - sx*sy/n )/(n-1);
	return cv;
}

double correlacion(x, y, n)
double	x[], y[];
int	n;
{
	int i;
	double	s1, s2, s3, s4, s5;

	s1=s2=s3=s4=s5=(double)0.0;
	for (i=0; i<n; i++) {
		s1 += x[i]*y[i];
		s2 += x[i];
		s3 += y[i];
		s4 += x[i]*x[i];
		s5 += y[i]*y[i];
	}
	return (s1-s2*s3/n)/sqrt((s4-s2*s2/n)*(s5-s3*s3/n));
}

kolmogorov(x, n, p, q, obs, nvl)
float	x[], p[], q[], *obs, *nvl;
int	n;
{
	float	discre, sump, b, f0;
	int	i, j;

	f_ordena(x,n);
	discre = 0.0;			/* lo mnimo */
	for (i=0; i<n; i++) {
		sump = 0.0;
		b = (i+0.5)/n;		/* F(x[i]) */
		j = 0;
		while (q[j] < x[i] && q[j+1] <= x[i]) 
			sump += p[j++];
		/* x[i] == q[j] || x[i] <= q[j+1] */
/* si x == q[j], p.ej: q=(7,7,8,9) y x=7 --> aqu j = 0
   q=(7,8,8,9) y x=8, --> j = 1; q=(7,8,9,9), x=9 --> j = 2 */
 
		if (q[j] == q[j+1])
			f0 = b;
		else
			f0 = sump + (x[i] - q[j])*p[j]/(q[j+1]-q[j]);

		if (ABS(b-f0) > discre) discre = ABS(b-f0);
	}
	*obs = discre;				/* mxima discrepancia */
	*nvl = N_signif_Kolmo(discre, n);
}

float N_signif_Kolmo(obs, n)
float	obs;
int	n;
{
#define EPSILON2 (double)0.00005
	int	i,k,sgn;
	double	lambda, ns;

	if (obs < 1.0e-8) return 0.0;
	lambda = obs*sqrt((double)n);
	k = (int)ceil(sqrt((double)(-0.5)*log(EPSILON2))/lambda);
	ns = (double)1.0;
	for (i=1, sgn = -1; i<=k; sgn = -sgn, i++)
		ns += 2.0*sgn*exp((double)(-2.0)*i*i*lambda*lambda);
	return (float)ns;
}

double F_distri(q, p, k, x)     /* Con flujos NO acumulados */
double q[], p[], x;
int   k;
{
	double alfa = (double)0.0, t;
	int	i=0;

	if (x <= q[0]) return (double)0.0;

	t = q[0];
	while (i<k-1 && x>=t+q[i+1]) {
		alfa += p[i];
		t += q[++i];
	}
	if (i<k-1 && q[i+1]>0.0)  {
		alfa += p[i]*(x-t)/q[i+1];
		return alfa;
	}
	else if (i==k-1) return (double)1.0;
}

double rect_esperanza_D(p, q, k)
double	p[], q[];
int	k;
{
	int i;
	double e = (double)0.0;

	for (i=1; i<k; i++) 
		e += p[i-1]*(q[i] + q[i-1]);
	return e/(double)2.0;
}

double rect_variancia_D(p, q, k)
double	p[], q[];
int	k;
{
	int i;
	double e = (double)0.0, mu = rect_esperanza_D(p, q, k), *q0;

	q0 = (double *)malloc(sizeof(double)*k);
	for (i=0; i<k; i++) q0[i] = q[i]-mu;
	for (i=1; i<k; i++) 
		e += p[i-1]*(q0[i]*q0[i] + q0[i-1]*q0[i-1] + q0[i-1]*q0[i]);
	free(q0);
	return e/(double)3.0;
}

float rect_esperanza(p, q, k)
float	p[], q[];
int	k;
{
	int i;
	float e = 0.0;

	for (i=1; i<k; i++) 
		e += p[i-1]*(q[i] + q[i-1]);
	return e/2.0;
}

float rect_variancia(p, q, k)
float	p[], q[];
int	k;
{
	int i;
	float e = 0.0, mu = rect_esperanza(p, q, k), *q0;

	q0 = (float *)malloc(sizeof(float)*k);
	for (i=0; i<k; i++) q0[i] = q[i]-mu;
	for (i=1; i<k; i++) 
		e += p[i-1]*(q0[i]*q0[i] + q0[i-1]*q0[i-1] + q0[i-1]*q0[i]);
	free(q0);
	return e/3.0;
}

float rect_covarianza(p, qx, qy, dc, k)
float	p[], qx[], qy[], *dc;
int	k;
{
	/* dc(,): funcin de densidad conjunta (x,y) 
		(i,j) == (k-1)*i + j; 	i = 0..k-2, j = 0..k-2 */

	int i, j, l;
	float cv = 0.0, mux, muy, ax[10], ay[10];

	mux = rect_esperanza(p, qx, k);
	muy = rect_esperanza(p, qy, k);
	for (i=1; i<k; i++) {
		ax[i-1] = (qx[i] - qx[i-1])*(qx[i] + qx[i-1] - 2.0*mux);
		ay[i-1] = (qy[i] - qy[i-1])*(qy[i] + qy[i-1] - 2.0*muy);
	}
	for (i=0, l=0; i<k-1; i++)
		for (j=0; j<k-1; j++, l++)
			cv += dc[l]*ax[i]*ay[j];	/*recorr. filas*/

	return cv/4.0;
}

float rect_esperanza_(p, q, k)	/* Interficie con fortran */
float	p[], q[];
int	k;
{
	return rect_esperanza(p,q,k);
}

float rect_variancia_(p, q, k)	/* Interficie con fortran */
float	p[], q[];
int	k;
{
	return rect_variancia(p,q,k);
}

descriptiva_(x, n, m, v)	/* Interficie con fortran */
float	x[], *m, *v;
int	n;
{
	descriptiva(x, n, m, v);
}

qwantilae_(z, n, q, p, k)	/* Interficie con fortran */
float	z[], q[], p[];
int	n, k;
{
	qwantilae(z, n, q, p, k);
}

qwantilae_ordena_(z, n, q, p, k)	/* Interficie con fortran */
float	z[], q[], p[];
int	n, k;
{
	qwantilae_ordena(z, n, q, p, k);
}

float rect_covarianza_(p, qx, qy, dc, k) /* Interficie con fortran */
float	p[], qx[], qy[], *dc;
int	k;
{
	return rect_covarianza(p, qx, qy, dc, k);
}
