c***********************************************************************
c     Routine for reading model dimensions
c
c     INPUT parameters:
c     model:  name of the model
c     nmodel: lenght of the model name
c     nunit:  number of a free unit to open files from fortran
c     
c     OUTPUT parameters:
c     nint:  number of intervals
c     nat:   number of commodities (for hydrogeneration)
c     ncomb: number of available fuels
c     nterm: number of thermal units plus contracts
c     ncontr:number of contracts
c     neb:   number of reservoirs
c     ncc:   number of discharge and pumping arcs
c     nvv:   number of spillage arcs
c     ntv:   number of transhipment arcs
c     nar:   number of arcs (nar=ncc+nvv+ntv)
c     iexit: if iexit= 0 routine exited normally
c            if iexit= -1 data file (.dat) could not be opened
c            if iexit= -2 fuels file (.cmb) could not be opened
c            if iexit= -3 thermal units file (.trm) could not be opened
c            if iexit= -4 hydro model data file could not be opened
c***********************************************************************
      subroutine get_dimensions(model,nmodel,nunit,nint,nat,ncomb,nterm,
     $     ncontr,neb,ncc,nvv,ntv,nar,iexit)
      implicit none
      include 'path_hydromod.h'
c     arguments
      integer*4 nmodel,nunit,nint,nat,ncomb,nterm,ncontr,neb,
     $     ncc,nvv,ntv,iexit
      character*1 model(nmodel)
c     local variables
      integer*4 i,j,k,iaux,in,nar
      real*8 raux
      character*256 namemod,finput,fcmb,ftrm,hydromodel,fhydro
      character*3 straux
      character*1 arctype,t_o_c

      iexit= 0

      in= nunit
      do i= 1,nmodel
         namemod(i:i)= model(i)
      end do
      do i= nmodel+1,256
         namemod(i:i)= ' '
      end do
      
c     reading from .dat file: nint, nat and the name of the fuels 
c     and thermal units files (processed below)
      in= nunit
      call concat(finput,namemod,'.dat')
      open(unit=in,err=901,file=finput,status='old')

      read(in,*) hydromodel
      read(in,*) nint
      read(in,*) nat
      read(in,2000) (raux,i=1,nat-1) 
      read(in,*) (iaux, i=1,nint)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*) 
c*      read(in,*) (raux,i=1,nint)
c*      read(in,*) (iaux,i=1,nint)
      read(in,*) fcmb
      read(in,*) ftrm
      close(in)

      if (index(ftrm,'.').eq.0) then 
         call concat(ftrm,ftrm,'.trm')
      endif
      if (index(fcmb,'.').eq.0) then 
         call concat(fcmb,fcmb,'.cmb')
      endif
C==== Puede interesar crear los ficheros .trm y .cmb para leerlos despues?
C	Por si acaso, hay esta el codigo:
C
      call creat_trm_cmb(namemod, ftrm, fcmb, nint, iexit)
      if (iexit.ne.0) return
C......

c     	reading from fuels file: ncomb
      open(unit=in,err=902,file=fcmb,status='old')
      read(in,*) ncomb
      close(in)

c     	reading from thermal units file: nterm
      open(unit=in,err=903,file=ftrm,status='old')
      read(in,*) nterm
      ncontr= 0
      do i=1,nterm
         read(in,97) t_o_c,straux,raux,raux,straux,j
         do k=1,j
            read(in,*) 
         end do
         call minuscules(t_o_c,t_o_c)
         if (t_o_c.eq.'c') ncontr= ncontr+1
      end do
      close(in)
      
c     reading from hydro model: neb, ncc, nvv, ntv
      if (direstruct) then
         call prefix(hydromodel,fhydro)
         call concat(fhydro,datapath,fhydro)
         call concat(fhydro,fhydro,'/')
      else
         fhydro= datapath
      end if
      call concat(fhydro,fhydro,hydromodel)
      call concat(fhydro,fhydro,'.hydro')
!!!!!!      call minuscules(fhydro,fhydro)
      open (unit=in,err=904,file=fhydro,status='old')
      read(in,*) neb
      do i= 1, neb
         read(in,*)
         read(in,2010) straux,raux,raux,iaux,iaux
         read(in,2030) (raux,j=1,nat)
         read(in,2000) raux, raux, raux, raux 
         read(in,2030) raux
      end do
      read(in,*) nar
      ncc= 0
      nvv= 0
      ntv= 0
      do i= 1,nar
         read(in,2020) arctype, straux, straux, iaux, raux, raux,
     $        iaux, iaux
         read(in,2030) (raux, j=1,nat)
         call minuscules(arctype, arctype)
         if ( (arctype.eq.'d').or.
     $        (arctype.eq.'e').or.
     $        (arctype.eq.'f').or.
     $        (arctype.eq.'g')    ) then 
            read (in,2030) raux, raux
         end if	      	      
         if ( (arctype.eq.'s').or.
     $        (arctype.eq.'t').or.
     $        (arctype.eq.'u').or.
     $        (arctype.eq.'r').or.
     $        (arctype.eq.'b').or.
     $        (arctype.eq.'c').or.
     $        (arctype.eq.'w').or.
     $        (arctype.eq.'x')    ) then 
            read (in,2030) raux
         end if
c     update ncc, nvv, ntv
         if ( (arctype.eq.'d') .or.
     $        (arctype.eq.'e') .or.
     $        (arctype.eq.'f') .or.
     $        (arctype.eq.'g') .or.
     $        (arctype.eq.'s') .or.
     $        (arctype.eq.'t') .or.
     $        (arctype.eq.'u') .or.
     $        (arctype.eq.'r') .or.
     $        (arctype.eq.'b') .or.
     $        (arctype.eq.'c') .or.
     $        (arctype.eq.'w') .or.
     $        (arctype.eq.'x')      ) ncc= ncc+1
         if ( (arctype.eq.'v') .or.
     $        (arctype.eq.'o') .or.
     $        (arctype.eq.'p') .or.
     $        (arctype.eq.'q')      ) nvv= nvv+1  ![[ Ampliado por JAG ]]
         if ( (arctype.eq.'i')      ) ntv= ntv+1
      end do
      close(in)
      
      return
 901  iexit= -1
      return
 902  iexit= -2
      return
 903  iexit= -3
      return
 904  iexit= -4
      return

97	format(x, a1, x, a3, x, f7.1, x, f7.3, x, a12, i1)
 2000 format(5 (1x,e13.7))
 2010 format(1x,a3,1x,2f10.4,1x,i2,1x,i2)
 2020 format(1x,a1,1x,a3,1x,a3,1x,i5,1x,2f10.4,1x,i2,1x,i2)
 2030 format(1x,7f10.4)
               
      end

      !! En esta rutina hay comentarios marcados con una 'C' o con 'C===='
      !! que son mios [JAG, VIII/98]
c***********************************************************************
c     Routine for loading the thermal system structure and some 
c     general data. Performs a simple check of the input data.
c     Values are stored in commons "thermal_system" and "general_data".
c
c     INPUT parameters:
c     model:  name of the model
c     nmodel: lenght of the model name
c     nunit:  number of a free unit to open files from fortran
c     
c     OUTPUT parameters:
c     iexit: if iexit= 0 routine exited normally
c            if iexit= -1 data file (.dat) could not be opened
c            if iexit= -2 fuels file (.cmb) could not be opened
c            if iexit= -3 thermal units file (.trm) could not be opened
c            if iexit= -5 too many intervals (nint > MAXINT)
c            if iexit= -6 too many hydro commodities (nat > MAXNAT)
c            if iexit= -7 too many fuels (ncomb > MAXCOMB)
c            if iexit= -8 too many thermal units (nterm > MAXTERM)
c            if iexit= -9 a thermal unit uses more than available fuels
c            if iexit= -10 a thermal unit uses an unknown fuel
c            if iexit= -11 error in initial or final interval for some
c                          some contract
C====	     if iexit= -17 load file (.load) could not be opened
C====	     if iexit= -18 not enough hours in load file 
c***********************************************************************
      subroutine get_gen_and_thermal_data(model,nmodel,
     $     nunit,iexit)
      implicit none

      include 'htcoor_f.h'
c     arguments
      integer*4 	nmodel, nunit, iexit
      character*1 	model(nmodel)
c     functions
      integer*4 	busca_str
c     local variables
      integer*4 	i, j, k, in, toth
      character*256 	namemod, finput, fcmb, ftrm
      character 	straux*10, t_o_c
      real		Load(MAXLOADS*MAXINT)

      iexit= 0

      do i= 1,nmodel
         namemod(i:i)= model(i)
      end do
      do i= nmodel+1,256
         namemod(i:i)= ' '
      end do

c     reading from .dat file: general data
      in= nunit
      call concat(finput,namemod,'.dat')
      open(unit=in,err=901,file=finput,status='old')

      read(in,*)
      read(in,*) nint
      if (nint.gt.MAXINT) then
         iexit= -5
         return
      endif
      read(in,*) nat
      if (nat.gt.MAXNAT) then
         iexit= -6
         return
      end if
      read(in,2000) (probhd(i),i=2,nat) 
      read(in,*) (nhores(k),k=1,nint)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*)
      read(in,*) phdmax
c*      read(in,*) (ppunta(k),k=1,nint)	    !! OBSOLETO [jag]
c*      read(in,*) (nforma(k),k=1,nint)	    !! OBSOLETO [jag]
                                !! quedara como reminiscencia: nforma(k) = k
      read(in,*) fcmb
      read(in,*) ftrm
      read(in,*) ffor			!! OBSOLETO [jag]
      close (in)

C==== Marca de JAG ----
C	A continuacion incluyo (comentado) codigo que debe sustituir 
C	alguna parte del codigo antiguo; no es definitivo todavia,
C	aun se leen ficheros para obtener ciertas cosas.
C
C	Las cargas horarias estan en un fichero <modelo>.load
C
      call concat(finput,namemod,'.load')
      open(unit=in,err=904,file=finput,status='old')
      i=0
      do while (.true.)
	  i=i+1
	  read(in, *, end=300) Load(i)
      end do
 300  close(in)
      toth = 0
      do k=1, nint
	  toth = toth + nhores(k)
      end do
      if (i .lt. toth) then
	  iexit = -18
	  return
      end if
C
C	Se supone que el primer dia de cargas se corresponde con el primer
C	dia del periodo de optimizacion (a la BD se le pediria un vector de
C	cargas a partir de un dia especifico)
C
C	Obtencion de formas y potencias punta de las monotonas (esto ya 
C	es definitivo)
C
      call getLDCshapes(nhores, nint, NQT, NQP, monot, NQT, Load, iexit)
      if (iexit.ne.0) return
      call getLDCpeaks(nhores, nint, Load, ppunta)

      do k=1, nint
	  nforma(k) = k
      end do
C......

c     reading from .cmb fuels file
      if (index(fcmb,'.').eq.0) then 
         call concat(fcmb,fcmb,'.cmb')
      endif
      open(unit=in,err=902,file=fcmb,status='old')
      read(in,*) ncomb
      if (ncomb.ge.MAXCOMB) then 
         iexit= -7
         return
      endif
      do i=1, ncomb
         read(in,96) preu(i),totalcomb(i),apreciacio(i),nomcomb(i),
     >		    minimcomb(i)
         call minuscules(nomcomb(i),nomcomb(i))
      end do
      read(in,*) preuext
      close(in)

c     reading from .trm thermal units file. Contracts are considered
c     as special thermal units.
      if (index(ftrm,'.').eq.0) then 
         call concat(ftrm,ftrm,'.trm')
      endif
      open(unit=in,err=903,file=ftrm,status='old')

      ncontr= 0
      read(in,*) nterm, nman
      if (nterm.gt.MAXTERM) then
         iexit= -8
         return
      end if
      do i=1,nterm
         read(in,97) t_o_c,idterm(i),pmaxterm(i),panes(i),nomterm(i),
     $        nartterm(i)
         call minuscules(nomterm(i),nomterm(i))
         call minuscules(idterm(i),idterm(i))
c     see if it is a contract or a thermal unit
         call minuscules(t_o_c,t_o_c)
         if (t_o_c.eq.'t') then
c     deal with thermal units
            es_contracte(i)= .false.
            if (nartterm(i).gt.ncomb) then
               iexit= -9
               return
            end if
            do j=1,nartterm(i)
               read(in,98) straux,rendiment_g(i,j),potcomprar(i,j),
     $              potguardar(i,j)
               call minuscules(straux,straux)
               k= busca_str(nomcomb,ncomb,straux)
               if (k.eq.-1) then
                  iexit= -10
                  return
               endif
c               if (j.gt.1) then
c                  if (k.lt.artterm(i,j-1)) then
c                     write(*,*) 'fuels for this thermal unit are not',
c     $                    ' in the same order than in fuels file.'
c                     stop 
c                  end if
c               end if
               artterm(i,j)=k
            end do
         else if (t_o_c.eq.'c') then
c     deal with contract
            ncontr= ncontr+1
            es_contracte(i)= .true.     
            read(in,*) cost_quad(i),cost_lin(i),ini_contract(i),
     $           fi_contract(i)
            if ((ini_contract(i).lt.1).or.(fi_contract(i).gt.nint).or.
     $           (ini_contract(i).gt.fi_contract(i))) then
               iexit= -11
               return
            end if
         end if
      end do
C====			Included in this file info about maintenances
      do i=1, nman
	 read(in, 99) thu_man(i), pmx_man(i), from(i), until(i)
      end do
C.....
      close (in)
C====			Special inputs

      call bring_minpower(pminterm, nterm)

      return
 901  iexit= -1
      return
 902  iexit= -2
      return
 903  iexit= -3
      return
 904  iexit= -17
      return

96	format(x, f11.5, x, e13.7, x, f5.3, x, a12, x, e13.7)
97	format(x, a1, x, a3, x, f7.1, x, f7.3, x, a12, i1)
98	format(x, a12, x, f9.7, x, e13.7, x, e13.7)
99	format(x, a12, x, f7.1, x, i4, x, i4)
 2000 format(5 (1x,e13.7))

      end


c***********************************************************************
c     Transforms the original data of the thermal system, computing
c     the values required by the optimization phase (e.g., information
c     related with the Bezier curves).
c     Values are stored in commons "thermal_system", "general_data",
c     and "bezier_curves_gen_opt".
c
c     INPUT parameters:
c     
c     OUTPUT parameters:
c     iexit: if iexit= 0 routine exited normally
c***********************************************************************
      subroutine deal_with_gen_and_thermal_data(iexit)
      implicit none
      include 'htcoor_f.h'
      include 'TadBezier.h'
c     arguments
      integer*4 iexit
c     local variables
      integer*4 i,j,k
      real*8 racum,tothores,totpotcomprar,limitstmp(MAXTERM)
      record /corbabez/ cbez
      
      iexit= 0

c     convert availability in % to outage probability
      do i= 1,nterm
         panes(i)= 1.0d0-panes(i)/100.0d0
      end do

c     compute some values for contracts
      do i= 1, nterm
         if (es_contracte(i)) then
c     ficticious fuel 
            nartterm(i)= 1   
c     ncomb+1= index of ficticious fuel
            artterm(i,1)= ncomb+1 
c     ficticious efficiency for this fuel
            rendiment_g(i,1)= 100.0 
c     estimate total fuel that can be purchased with this contract
            racum= 0.0d0
            do j= ini_contract(i),fi_contract(i)
               racum= racum + nhores(j)
            end do
            potcomprar(i,1)= racum*pmaxterm(i)
            potguardar(i,1)= potcomprar(i,1)
         end if
      end do
      
c     adding new ficticious fuel for contracts
      if (ncontr.gt.0) then
         ncomb= ncomb+1
         nomcomb(ncomb)= 'contracts'
         racum= 0.0d0
         do i=1,nterm
            if (es_contracte(i)) then
               racum= racum+potcomprar(i,1)
            end if
         end do
         totalcomb(ncomb)= racum
      end if

c     estimating which thermal units have fuel limits
c     (contracts never have fuel limits)
      tothores= 0.0d0
      do i=1,nint
         tothores= tothores+nhores(i)
      end do
      do j= 1,nterm
         if (es_contracte(j)) then
            te_limits_comb(j)= .false.
	    pmaxlim(j) = 0.d0
         else
            totpotcomprar= 0.0d0
            do k= 1,nartterm(j)
               totpotcomprar= totpotcomprar+potcomprar(j,k)
            end do
            te_limits_comb(j)= 
     +           (totpotcomprar .lt. (pmaxterm(j)*tothores))
	    pmaxlim(j)=min(pmaxterm(j),totpotcomprar/tothores)
         end if
      end do
      
c     computing Bezier curves and scale factors for time, power and
c     energy
      do j= 1,nterm
         limitstmp(j)= 0.0d0
         do k= 1,nartterm(j)
            limitstmp(j)= limitstmp(j)+potcomprar(j,k)
         end do
      end do
      call carregar_bateries(nint,nterm,phdmax,nforma,ppunta,nhores,
     +     panes,pmaxterm,limitstmp,es_contracte,pmaxlim,te_limits_comb,
     +	   ini_contract, fi_contract)
      call analisis_monotonas(escal_potencia,escal_hores,monot,NQT)
      escal_energia= escal_potencia*escal_hores      

c     set scale factor for prices
      ordre_preu= 1000.0d0

c     scaling powers and energies through scale factors
      phdmax= phdmax/escal_potencia
      do i=1,ncomb
         totalcomb(i)= totalcomb(i)/escal_energia
         preu(i)= preu(i)/ordre_preu
      end do
      preuext=preuext/ordre_preu
      do i=1,nint
         ppunta(i)= ppunta(i)/escal_potencia
         nhores(i)= nhores(i)/escal_hores
      end do
      do i=1,nterm
         pmaxterm(i)= pmaxterm(i)/escal_potencia
         pminterm(i)= pminterm(i)/escal_potencia
         do k=1,nartterm(i)
            potguardar(i,k)= potguardar(i,k)/escal_energia
            potcomprar(i,k)= potcomprar(i,k)/escal_energia
         end do
         if (es_contracte(i)) then
            cost_lin(i)= cost_lin(i)/ordre_preu
            cost_quad(i)= cost_quad(i)/ordre_preu
         end if
      end do

c     compute number of thermal commodities
      nthrmarticles= 0
      do j=1,nterm
         nthrmarticles= nthrmarticles+nartterm(j)
      end do
      
c     get data from Bezier curve structures; most of this data will be
c     used during the optimization phase
      do i= 1,nint
         call obtener_curva(i,cbez)
         energia(i)= cbez.energia
         e0a(i)= cbez.energy.a
         e0b(i)= cbez.energy.b
         e0c(i)= cbez.energy.c
         t0a(i)= cbez.time.a
         t0b(i)= cbez.time.b
         t0c(i)= cbez.time.c
         pmin(i)= cbez.pmin
         tbase(i)= cbez.tbase
         max_ho(i)= cbez.max_ho
         min_gt(i)= cbez.min_gt
      end do      
c      
c     write data to check them (this must not be done, it is just
c     useful to compare with data generated by the old genhidro code)
c      write(*,*) escal_potencia,escal_hores,escal_energia,ordre_preu
c      write(*,*) nterm
c      write(*,*) ncomb
c      write(*,*) phdmax
c      write(*,*) (preu(k),k=1,ncomb)
c      write(*,*) preuext
c      write(*,*) (apreciacio(k),k=1,ncomb)
c      write(*,*) (nartterm(j),j=1,nterm)
c      do j=1,nterm
c         write(*,*) (artterm(j,k),k=1,nartterm(j))
c      end do
c      write(*,'(5(a10))') (nomterm(j),j=1,nterm)
c      write(*,*) (pmaxterm(j),j=1,nterm)
c      write(*,*) (nforma(i),i=1,nint)
c      write(*,*) (ppunta(i),i=1,nint)
c      do i= 1,nint
c         write(*,*) energia(i)
c         write(*,*) e0a(i),e0b(i),e0c(i)
c         write(*,*) t0a(i),t0b(i),t0c(i)         
c         write(*,*) pmin(i),tbase(i)
c      end do
c      write(*,*) (es_contracte(j),j=1,nterm)
c      write(*,*) (cost_lin(j),j=1,nterm)
c      write(*,*) (cost_quad(j),j=1,nterm)
      end


c***********************************************************************
c     Routine for loading the hydro system structure. 
c
c     INPUT parameters:
c     model:  name of the model
c     nmodel: lenght of the model name
c     nunit:  number of a free unit to open files from fortran
c     
c     OUTPUT parameters:
c     iexit: if iexit= 0 routine exited normally
c            if iexit= -1 data file (.dat) could not be opened
c            if iexit= -14 file with coefficients of the fourth degree 
c                polynomial arc efficiency function could not be opened
c            otherwise =  error returned by lec_conca (see lec_conca)
c***********************************************************************
      subroutine get_hydro_system(model,nmodel,nunit,iexit,init_date,
     >				  inf_dtes)
      implicit none
      include 'path_hydromod.h'
      include 'htcoor_f.h'
      structure /Fecha/
	integer*2	d,m,a
      end structure

c     arguments
      integer*4 nmodel,nunit,iexit
      character*1 model(nmodel)
      record /Fecha/ init_date, inf_dtes(*)

c     local variables
      integer*4 i,in,iaux,mo,mua,muv,muc,ndias(MAXINT)
      integer*4	get_cdf_inflows
      real*8 raux
      character*256 namemod,finput,hydromodel

      iexit= 0

      do i= 1,nmodel
         namemod(i:i)= model(i)
      end do
      do i= nmodel+1,256
         namemod(i:i)= ' '
      end do

c     reading from .dat file the name of the hydromodel and some
c     information for locating the reservoirs and arcs data files
      in= nunit
      call concat(finput,namemod,'.dat')
      open(unit=in,err=901,file=finput,status='old')
      read(in,*) hydromodel
      read(in,*) 
      read(in,*) 
      read(in,2000) (raux,i=1,nat-1) 
      read(in,*) (iaux, i=1,nint)
      read(in,*) mo
      read(in,*) mua
      read(in,*) muv
      read(in,*) muc
      read(in,*) pvi
      close(in)

C====	Below we call to the function "get_cdf_inflows", which computes
C	the block distributions of the natural inflows. The result is
C	stored in a private array that can be accessed through an special
C	function "gimme_inflows"

      do i=1, nint
	ndias(i) = nhores(i)/24
      end do
      iexit= get_cdf_inflows(model, init_date, ndias, nint, inf_dtes,
     >			    nat, probhd)
      if (iexit.ne.0) call tractar_errors(%val(iexit))

c     reading the hydro model; see lec_conca for an explanation
c     of the parameters
      call lec_conca(hydromodel, neb, nar, nint, nat, 
     $     ncc, nvv, ntv,
     $     mo, mua, muv, muc, MAXNAR, MAXNEB, MAXNAT, MAXINT, 
     $     conca, capmut, capemb, vlx, aportacio, 
     $     nomembas, embas ,tiparc, limarc, limemb, volini, volfin, 
     $     minima, minime, sbx, ctx,
     $     nhores, retard, nunit,
     $     cvb, cvl, cvq, cvc, cvcu,
     $     iexit)
      if (iexit.ne.0) return
      
c     reading coefficients of the fourth degree polynomial 
c     arc efficiency function for each generation arc
C      call concat(finput,namemod,'.ren')
C      open(unit=in,err=914,file=finput,status='old')
C      do i=1, ncc
C         read (in,*) ehe4(i), ehe3(i), ehe2(i), ehe1(i), ehe0(i)
C      end do      

C	Reading coefficients of quadratic polynomial (two variables)
C	for generation arcs, or fourth degree polynomial for
C	pumping arcs

	call gimme_eff_polyn(ehe5, ehe4, ehe3, ehe2, ehe1, ehe0, ncc)

      return
 901  iexit= -1
      return
 914  iexit= -14
      return

 2000 format(5 (1x,e13.7))

      end


c***********************************************************************
c     routine for reading the hydromodel
c     (the units --of volumes, etc.-- are that of the original data
c     read; after their lecture, they are converted to the value that 
c     will be used during the optimization phase)
c     
c     INPUT parameters:
c     hydromodel:  name of the hydromodel
c     nat (=nks of original lec_conca routine): number of commodities 
c     nint (=nre of original lec_conca routine): number of intervals
c     mo,mua,muv,muc: information about which files must be read
c                     to load hydro system
c     MAXNAR: maximum number of arcs
c     MAXNEB: mabimum number of reservoirs
c     MAXNAT: maximum number of commodities
c     MAXINT: maximum number of intervals
c     nunit: number of a free unit to open files from fortran
c
c     OUTPUT parameters:
c     neb (=nnu of original lec_conca routine): number of reservoirs
c     nar: number of arcs (discharge and pumping+spill+transhipment)
c     ncc: number of discharge and pumping arcs
c     nvv: number of spillage arcs
c     ntv: nubmer of transhipment arcs
c     conca(nar,2): hydro network topology (for each arc, origin and
c                   destination reservoirs)
c     capmut(nar,nint): mutual capacities of the arcs in m3/s
c     capemb(neb,nint): reservoirs capacities in Hm3
c     vlx(neb): reservoir capacities again in Hm3 (auxiliary vector)
c     aportacio(nat,neb,nint): reservoir inflows in m3/s
c     nomembas(neb): names of the reservoirs
c     embas(neb): three character identifiers for the reservoirs
c     tiparc(nar): types of the arcs (spillage,disch./pump. or transh.)
c     limarc(nat,nar,nint): arc bounds for comm. and interval in m3/s
c     limemb(nat,neb,nint): reservoir capacity bounds for commodity 
c                           and interval in Hm3
c     volini(nat,neb): initial volumes for comm. and reservoir in Hm3 
c     volfi(nat,neb): final volumes for comm. and reservoir in Hm3
c     minima(nar,nint): minimum flows at arcs in m3/s
c     minime(neb,nint): minimum capacities at reservoirs in Hm3
c     sbx(nar): what's that???: it is required to generate snq(nar)
c     ctx(neb): what's that???: it is required to generate snq(nar)
c     nhores(nint): duration of each interval
c     retard(nar): arcs delays
c     cvb,cvl,cvq,cvc,cvcu (neb): coefficients of the fourth degree 
c                                 polynomial head volume function
c     iexit: if iexit= 0 routine exited normally
c            if iexit= -4 hydro model data file could not be opened
c            if iexit= -12 an inflows data file could not be opened
c                          for some reservoir    
c            if iexit= -13 a commodity volumes data file could not 
c                          be opened for some reservoir
c***********************************************************************
      subroutine lec_conca(hydromodel, neb, nar, nint, nat,
     $     ncc, nvv, ntv,
     $     mo, mua, muv, muc, MAXNAR, MAXNEB, MAXNAT, MAXINT, 
     $     conca, capmut, capemb, vlx, aportacio, 
     $     nomembas, embas, tiparc, limarc, limemb, volini, volfin, 
     $     minima, minime, sbx, ctx,
     $     nhores, retard, nunit,
     $     cvb, cvl, cvq, cvc, cvcu,
     $     iexit)
      implicit none
      include 'path_hydromod.h'
c     arguments
      integer*4 neb, nar, nint, nat, ncc, nvv, ntv, mo, mua, muv, muc, 
     $     MAXNAR, MAXNEB, MAXNAT, MAXINT, 
     $     conca(MAXNAR,2), retard(MAXNAR), nunit, iexit
      real*8 capmut(MAXNAR,MAXINT), capemb(MAXNEB,MAXINT-1),vlx(MAXNEB),
     $     aportacio(MAXNAT,MAXNEB,MAXINT), 
     $     limarc(MAXNAT,MAXNAR,MAXINT), limemb(MAXNAT,MAXNEB,MAXINT), 
     $     volini(MAXNAT,MAXNEB), volfin(MAXNAT,MAXNEB),
     $     minima(MAXNAR,MAXINT), minime(MAXNEB,MAXINT), 
     $     sbx(MAXNAR), ctx(MAXNEB), nhores(MAXINT), 
     $     cvb(MAXNEB), cvl(MAXNEB), cvq(MAXNEB), 
     $     cvc(MAXNEB), cvcu(MAXNEB)
      character*1 tiparc(MAXNAR) 
      character*3 embas(MAXNEB+1)
      character*10 nomembas(MAXNEB)
      character*256 hydromodel
      
c     local variables
      integer*4	i, j, k, fla, flb, canal
      real*8 raux, m3s_hm3h, mnme, cpmt, mnma
      character*1 anat
c      character*2 amo ,amua ,amuv ,amuc 
      character*3 org ,dst
c      character*10 nom10
c      character*12 nom12
      character*256 str

c     initialize iexit and read units
      iexit= 0
      canal= nunit

c     conversion coefficient from m3/s to Hm3/hour
      m3s_hm3h = 0.0036d0
            
c     open reservoirs and arcs data file 
      encode(1,240,anat) nat
      if (direstruct) then
         call prefix(hydromodel,str)
         call concat(str,datapath,str)
         call concat(str,str,'/')
      else
         str= datapath
      end if
      call concat(str,str,hydromodel)
      call concat(str,str,'.hydro')
!!!!!!      call minuscules(str,str)
      open (unit=canal,err=904,file=str,status='old')

c     reading reservoirs information (volumes in Hm3)
      read (canal,*) neb
c     Bring fourth coefficient 
C         cvcu(i)= 0.0d0 
	 call gimme_4th_cv(cvcu, neb)
C......................................
      do i=1,neb
         read (canal,*) nomembas(i)
         call minuscules (nomembas(i),nomembas(i))
         read (canal,210) embas(i) ,vlx(i) ,mnme ,fla ,flb
         call minuscules(embas(i),embas(i))
         read (canal,230) (limemb(k,i,1),k=1,nat)
         read (canal,*) cvb(i),cvl(i),cvq(i),cvc(i)

         read (canal,*) ctx(i)
         do j=1,nint
            capemb(i,j) = vlx(i)
            minime(i,j) = mnme
         end do
         do j=2,nint
            do k=1,nat
               limemb(k,i,j) = limemb(k,i,1)
            end do
         end do
c     Unlike the original model, there are not reservoir side constr.
c     Unlike the original model, there are not non uniform capacities
c     at some reservoirs.
      end do
      
c     reading arcs information (flows in m3/s)
      read (canal,*) nar
      do i=1,nar
         read (canal,220) tiparc(i),org,dst,retard(i),cpmt,mnma,fla,flb
         read (canal,230) (limarc(k,i,1),k=1,nat)
         call minuscules (tiparc(i),tiparc(i))
         call minuscules (org,org)
         call minuscules (dst,dst)
c     obtain maximum flow capacities in Hm3
         do j=1,nint
            capmut(i,j) = cpmt * m3s_hm3h * nhores(j)
            minima(i,j) = mnma * m3s_hm3h * nhores(j)
         end do
         do j=nint,1,-1 !reverse order necessary due to limarc(k,i,1)
            do k=1,nat
               limarc(k,i,j) = limarc(k,i,1) * m3s_hm3h * nhores(j)
            end do
         end do

c     create network topology structure (origin and destination 
c     reservoirs for each arc)
         j = 1
         do while ( (j.le.neb).and.(embas(j).ne.org) )
            j = j + 1
         end do
         conca(i,1) = j
         if (dst.eq.'zzz') then
            conca(i,2) = ( neb * nint ) + 1
         else
            k = 1
            do while ( (k.le.neb).and.(embas(k).ne.dst) )
               k = k + 1
            end do
            conca(i,2) = k
         end if
         
c     update ncc, ntv and nvv according to the arc type
         if ( (tiparc(i).eq.'d') .or.
     $        (tiparc(i).eq.'e') .or.
     $        (tiparc(i).eq.'f') .or.
     $        (tiparc(i).eq.'g') .or.
     $        (tiparc(i).eq.'s') .or.
     $        (tiparc(i).eq.'t') .or.
     $        (tiparc(i).eq.'u') .or.
     $        (tiparc(i).eq.'r') .or.
     $        (tiparc(i).eq.'b') .or.
     $        (tiparc(i).eq.'c') .or.
     $        (tiparc(i).eq.'w') .or.
     $        (tiparc(i).eq.'x')      ) ncc= ncc+1
         if ( (tiparc(i).eq.'v') .or.
     $        (tiparc(i).eq.'o') .or.
     $        (tiparc(i).eq.'p') .or.
     $        (tiparc(i).eq.'q')      ) nvv= nvv+1  ![[ Ampliado por JAG ]]  
         if ( (tiparc(i).eq.'i')      ) ntv= ntv+1

c     read additional information according to the arc type
         if ( (tiparc(i).eq.'d').or.  
     $        (tiparc(i).eq.'e').or.  
     $        (tiparc(i).eq.'f').or.  
     $        (tiparc(i).eq.'g')    ) then
            read(canal,230) raux, sbx(i) 
         end if
         
         if ( (tiparc(i).eq.'s').or.  
     $        (tiparc(i).eq.'t').or.  
     $        (tiparc(i).eq.'u').or.  
     $        (tiparc(i).eq.'r').or.  
     $        (tiparc(i).eq.'b').or.  
     $        (tiparc(i).eq.'c').or.  
     $        (tiparc(i).eq.'w').or.  
     $        (tiparc(i).eq.'x')     ) read (canal,*)
         
      end do
c     Unlike the original model, no general side constraints 
c     are considered
      close(canal)	
      
C=      encode(2,250,amo ) mo
C=      encode(2,250,amua) mua
C=      encode(2,250,amuv) muv
C=      encode(2,250,amuc) muc
      
C=      if ( amo(1:1).eq.' ')  amo(1:1)='0'
C=      if (amua(1:1).eq.' ') amua(1:1)='0'
C=      if (amuv(1:1).eq.' ') amuv(1:1)='0'
C=      if (amuc(1:1).eq.' ') amuc(1:1)='0'
      
c     read inflows information

	call gimme_inflows(aportacio, MAXNAT, MAXNEB, nat, neb, nint)
	do j=1,neb
	    do i=1,nint
		do k=1,nat
                    aportacio(k,j,i) = aportacio(k,j,i)*
     >					m3s_hm3h*nhores(i)
                end do
	    end do
	end do

C=      do j=1,neb
c     open inflow data files at reservoirs
C=         nom12 = embas(j) // anat // amua // amo // '.apo'
C=         if (direstruct) then
C=            call prefix(hydromodel,str)
C=            call concat(str,datapath,str)
C=            call concat(str,str,'/')
C=         else
C=            str= datapath
C=         end if
C=         call concat(str,str,nom12)
C=         call minuscules(str,str)
C=         open ( unit=canal , err=912, file=str ,status='old' )
         
c     read inflows (in m3/s)
C=         do i=1,nint
C=            read(canal,280) (aportacio(k,j,i),k=1,nat)
c     obtain total inflows per interval in Hm3
C=            do k=1,nat
C=               aportacio(k,j,i) = aportacio(k,j,i)*m3s_hm3h*nhores(i)
C=            end do
C=         end do
C=         close(canal)
C=      end do	   

c     reading initial and final reservoir volumes for each 
c     commodity in Hm3

	call gimme_inivol(volini, MAXNAT, nat, neb)
	call gimme_finvol(volfin, MAXNAT, nat, neb)

C=      do j=1,neb
C=         nom10 = embas(j) // anat // amuv // '.vol'
C=         if (direstruct) then
C=            call prefix(hydromodel,str)
C=            call concat(str,datapath,str)
C=            call concat(str,str,'/')
C=         else
C=            str= datapath
C=         end if
C=         call concat(str,str,nom10)
C=         call minuscules(str,str)
C=         open ( unit=canal, err=913, file=str ,status='old' )
C=         do i=1,nat
C=            read(canal,230) volini(i,j) ,volfin(i,j) 
C=         end do
C=         close(canal)
C=      end do
      
      return
 904  iexit= -4
      return
 912  iexit= -12
      return
 913  iexit= -13
      return

 200  format (10i6)
 210  format (1x,a3,1x,2f10.4,1x,i2,1x,i2)
 220  format (1x,a1,1x,a3,1x,a3,1x,i5,1x,2f10.4,1x,i2,1x,i2)
 230  format (1x,7f10.4)
 240  format (i1)
 250  format (i2)
 260  format (1x,i3,8(1x,f10.4))
 280  format (6f12.4)
 290  format (1x,i3,6(1x,f10.4))
 300  format (1x,a1,1x,i3)
 310  format (1x,i3,1x,i3,1x,f10.4)
 320  format (5x,i3,60(1x ,f10.4 ,' [' ,i3 ,'] '))
      end


c***********************************************************************
c     Transforms the original data of the hydro system, computing
c     the values required by the optimization phase.
c     Values are stored in common "hydro_system".
c
c     INPUT parameters:
c     
c     OUTPUT parameters:
c     iexit: if iexit= 0 routine exited normally
c***********************************************************************
      subroutine deal_with_hydro_system(iexit)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 iexit
c     local variables
      integer*4 i,j,xdimf
      real*8 xua

      iexit= 0

c     computes availability probabilities for hydro commodities.
c     original probabilites for each hydro block are at probhd(2:nat).
c     final probabilites are left at probhd(1:nat) (first position
c     corresponds to deterministic water and has prob= 1)
      probhd(1)= 0.0d0
      do i= 1,nat
         probhd(i)= probhd(i)/2.0d0
         do j= i+1,nat
            probhd(i)= probhd(i)+probhd(j)
         end do
      end do     

c     penalty parameter for spillage at non-full reservoirs
c     currently set to 0 to avoid ill-conditioning of funobj 
c     during optimization phase
      pvi= 0.0d0
      
c     create idenfifier for sink node
      embas(neb+1)= '...'

c     add initial and final volumes to inflows at reservoirs, at 
c     the first and last interval, to obtain the real injections 
c     at nodes of the multiinterval network 
      do j= 1,neb
         do i=1,nat
            aportacio(i,j,1)   = aportacio(i,j,1)   + volini(i,j)
            aportacio(i,j,nint) = aportacio(i,j,nint) - volfin(i,j)
         end do
      end do
      
c     according to minime(), we modify reservoir capacities to account
c     for minimum volumes to be guaranteed during all the intervals
      do j=1,neb
         do i=1,nint-1
            capemb(j,i)        = capemb(j,i)        - minime(j,i)
            limemb(1,j,i)      = limemb(1,j,i)      - minime(j,i)
            aportacio(1,j,i)   = aportacio(1,j,i)   - minime(j,i)
            aportacio(1,j,i+1) = aportacio(1,j,i+1) + minime(j,i)
         end do
      end do
      
c     according to minima(), we modify arc capacities to account
c     for minimum flows to be guaranteed during all the intervals
      do j=1,nar
         do i=1,nint
            capmut(j,i) = capmut(j,i) - minima(j,i)
            limarc(1,j,i) = limarc(1,j,i) - minima(j,i)
         end do
      end do
c     injections at initial and final nodes node for each arc and
c     interval must be still updated; this will be done later when
c     constructing the final multiinterval netwok, where these 
c     these node numbers will be known

c     generation of different required vectors
c     from now on, some of the called routines would use a different
c     notation (as nnu=neb, nre=nint, mdi=nhores, etc). Refer to
c     the calling routines to see the names used in the rest of the
c     code

      xdimf= (MAXNAR+MAXNEB)*MAXINT-MAXNEB
      
c     generating multiinterval hydro network
c     computing also nax (total number or arcs in the replicated 
c     network) and nnx (total number of nodes in the replicated network)
c     (nax= nar*nint+neb*(nint-1) )
c     (nnx= neb*nint+1)
      call gen_xarxa(neb, nar, nint, nnx, nax, nhores, escal_hores,
     $     MAXNAR, 2, conca,
     $     xdimf, 2, xarxa, 
     $     retard, tiparc, tiparx)

c     generate ipl and mnk vectors
      call gen_ipl_mnk ( nnx ,nax ,xdimf ,2 ,xarxa,
     $     ipl ,mnk )

c     generate rin and rfl vectors
      call gen_rin_rfl ( neb ,nat ,MAXNAT ,MAXNEB ,volini ,volfin,
     $     rin ,rfl )

c     generate jcl vector
      call gen_jcl ( nar ,tiparc ,jcl)	

c     generate jeo and jed vectors
      call gen_jeo_jed ( nar ,MAXNAR ,2 ,conca ,tiparc, 
     $     jeo ,jed )

c     generate ren and snq vectors
      call gen_snq( nar ,MAXNAR ,2 ,conca ,tiparc,
     $     cvb ,ctx ,sbx ,snq )
      
c     generate jll vector
      call gen_jll ( nar ,MAXNAR ,2 ,conca ,tiparc ,jll )	

c     adjust ehei (i=0..5) values using jll and escal_energia 
c     (previously computed)
      do j=1, ncc
         if (jll(j) .ge.0) then
            ehe0(j) = ehe0(j)/escal_energia
            ehe1(j) = ehe1(j)/escal_energia
            ehe2(j) = ehe2(j)/escal_energia
            ehe3(j) = ehe3(j)/escal_energia
            ehe4(j) = ehe4(j)/escal_energia
            ehe5(j) = ehe5(j)/escal_energia
         end if
      end do     

c     generate can vector
      call gen_can ( neb ,nar ,nint ,ncc ,nax, 		
     $     MAXNAR ,2 ,conca,
     $     xdimf ,2 ,xarxa,
     $     tiparc ,tiparx, 
     $     MAXNAR ,MAXINT ,minima, 
     $     can )

c     generate ivo, lvo, ivd and lvd vectors
      call gen_ivo_lvo_ivd_lvd ( neb ,nar ,nint ,ncc ,nax, 	
     $     MAXNAR ,2 ,conca,
     $     xdimf ,2 ,xarxa, 
     $     tiparc ,tiparx, 
     $     ipl ,ivo ,lvo ,ivd ,lvd )

c     generate jca vector
      call gen_jca ( neb ,nar ,nint ,ncc ,nax,
     $     MAXNAR ,2 ,conca,
     $     xdimf ,2 ,xarxa,
     $     tiparc ,tiparx ,jca )

c     generate cxv and iov vectors
      call gen_cxv_iov ( nar ,MAXNAR ,2 ,conca,
     $     tiparc ,vlx ,cxv ,iov )

c     generate ivv and mfv vectors
      call gen_ivv_mfv ( neb ,nar ,nint ,nvv ,nax, 	
     $     MAXNAR ,2 ,conca,
     $     xdimf ,2 ,xarxa,
     $     tiparc ,tiparx ,ipl ,ivv ,mfv )

c     generate von vector
      call gen_von(neb, nint, MAXNEB, MAXINT, minime, von)

c     write vectors elm,eqt,eqd,ecq,ecua for computing derivatives
c     during optimization stage
      do i=1,neb
         elm(i)=0.5d0*cvl(i)
         xua=cvq(i)/3.0d0
         eqt(i)=xua
         eqd(i)=2.0d0*xua
         ecq(i)=0.25d0*cvc(i)
         ecua(i)= .2d0*cvcu(i)
      end do

c     generate vector of pointer to pumping arcs and coefficients for
c     calculating jump to be surmounted (** modifies "snq" **)
      do i=1, ncc
	 if (jll(i).gt.0) then
            jabom(i) = 1
	 else
            jabom(i) = 0
         end if
      end do
      call gimme_jump_pump(salto, ncc, jabom, snq)

c     write data to check them (this must not be done, it is just
c     useful to compare with data generated by the old genhidro code)
c 110  format (10i6)
c 114  format (6a2)
c 115  format (5e13.7)
c 116  format (a80)
c      write(*,110) nint,nat
c      write(*,115) (probhd(i),i=2,nat)
c      write(*,110) (int(nhores(i)*escal_hores),i=1,nint)
c      write(*,115) pvi,0.0
c      write(*,110) neb,ncc,nvv,ntv
c      write(*,115) (cvl(i),i=1,neb)
c      write(*,115) (cvq(i),i=1,neb)
c      write(*,115) (cvc(i),i=1,neb)
c      write(*,115) (cvcu(i),i=1,neb)
c      write(*,115) (rin(i),i=1,neb*nat)
c      write(*,115) (rfl(i),i=1,neb*nat)
c      do i= 0,nint
c         write(*,115) (von(j),j=neb*i+1,neb*(i+1))
c      end do
c      write(*,110) (jcl(i),i=1,ncc)
c      write(*,110) (jeo(i),i=1,ncc)
c      write(*,110) (jed(i),i=1,ncc)
c      write(*,110) (jll(i),i=1,ncc)
c      write(*,115) (snq(i),i=1,ncc)
c      write(*,115) (0.,i=1,ncc)
c      write(*,110) (jca(i),i=1,ncc*nint)
c      write(*,110) (ivo(i),i=1,ncc*nint)
c      write(*,110) (lvo(i),i=1,ncc*nint)
c      write(*,110) (ivd(i),i=1,ncc*nint)
c      write(*,110) (lvd(i),i=1,ncc*nint)
c      write(*,115) (can(i),i=1,ncc*nint)
c      if (nvv.gt.0) then
c         write(*,115) (cxv(i),i=1,nvv)
c         write(*,110) (iov(i),i=1,nvv)
c         write(*,110) (ivv(i),i=1,nvv*nint)
c         write(*,110) (mfv(i),i=1,nvv*nint)
c      end if
c      do i=1, ncc
c         write(*,*) ehe4(i), ehe3(i), ehe2(i), ehe1(i), ehe0(i)
c      end do      

      end


c***********************************************************************
c     Generation of multiinterval or replicated hydro network. 
c     Arcs are sorted by destinations, and by origins for the
c     same destination. Sink is the last node.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_xarxa ( nnu ,nar ,nre ,nnx ,nax ,mdi, escal_mdi 
     $     ,cdimf ,cdimc ,conca
     $     ,xdimf ,xdimc ,xarxa 
     $     ,retard ,tiparc ,tiparx )
     $     
      implicit none
c     arguments
      integer*4	nnu ,nar ,nre ,nnx ,nax ,retard(*) 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc) 
     $     ,xdimf ,xdimc ,xarxa(xdimf,xdimc)
      real*8 mdi(*), escal_mdi
      character*1	tiparc(*) ,tiparx(*)
      
c     local variables
      integer*4 i ,j ,amdi ,ix ,npou ,trobat ,ij ,imin
     $     ,aux1,aux2
      character*1 auxc

c     body of routine
      npou = nnu * nre + 1
        
c     compute network arcs related to real hydro system arcs
      ix = 1
      do i=1,nar
         do j=1,nre
            xarxa(ix,1) = conca(i,1) + (j - 1) * nnu
            tiparx(ix) = tiparc(i)
c     compute origin node for the arc
            if (conca(i,2).eq.npou) then
               xarxa(ix, 2) = npou
            else
               amdi = mdi(j)*escal_mdi
               ij = j
               trobat = 0
               do while (trobat.eq.0)
                  if (amdi.ge.retard(i)) then
                     xarxa(ix,2) = conca(i,2) + (ij - 1) * nnu
                     trobat = 1
                  end if
                  if (trobat.eq.0) then
                     ij = ij + 1
                     if (ij.gt.nre) then
                        xarxa(ix,2) = npou
                        trobat = 1
                     end if
                  end if
                  if (trobat.eq.0) then
                     amdi = amdi + mdi(ij)*escal_mdi
                  end if
               end do
            end if
            ix = ix + 1
         end do
      end do
      
c     compute network arcs related to storage arcs (between intervals?)
      do i=1,nnu
         do j=1,nre-1
            xarxa(ix,1) = i + (j - 1) * nnu
            xarxa(ix,2) = i + j * nnu
            tiparx(ix) = 'm'
            ix = ix + 1
         end do
      end do
      
      nax = ix - 1
      nnx = npou
      
c     sorting arcs
      do i=1,nax
         imin = i
c     find smallest from i+1 to nax
         do j=i+1,nax
            if ( (  xarxa(j,2).lt.xarxa(imin,2)       ) .or.
     $           ( (xarxa(j,2).eq.xarxa(imin,2)) .and.
     $           (xarxa(j,1).lt.xarxa(imin,1))      ) .or.
     $           ( (xarxa(j,2).eq.xarxa(imin,2)) .and.
     $           (xarxa(j,1).eq.xarxa(imin,1)) .and.
     $           ( tiparx(j).lt.tiparx(imin) ) 	) ) imin = j
         end do
         
c     swaping with the i-th
         if (imin.ne.i) then
            aux1 = xarxa(i,1)
            aux2 = xarxa(i,2)
            auxc = tiparx(i)
            
            xarxa(i,1) = xarxa(imin,1)
            xarxa(i,2) = xarxa(imin,2)
            tiparx(i)  = tiparx(imin)
            
            xarxa(imin,1) = aux1
            xarxa(imin,2) = aux2
            tiparx(imin)  = auxc
         end if
      end do
      
      return
      end


c***********************************************************************
c     Generation of ipl and mnk vectors.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_ipl_mnk ( nnx ,nax ,xdimf ,xdimc ,xarxa
     $     ,ipl ,mnk )
      implicit none
      
c     arguments
      integer*4 nnx ,nax ,xdimf ,xdimc ,xarxa(xdimf,xdimc)
     $     ,ipl(*) ,mnk(*)
      
c     local variables
      integer*4 i ,px
      
c     body of routine
      px = 1
      do i=1,nnx
         do while (xarxa(px,2).lt.i)
            px = px + 1
         end do
         ipl(i) = px
      end do
      do i=1,nax
         mnk(i) = xarxa(i,1)
      end do
      return
      end 
      
c***********************************************************************
c     Generation of vector can of minimum flows of the discharge
c     and pumping arcs for all the intervals
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_can ( nnu ,nar ,nre ,ncc ,nax 		
     $     ,cdimf ,cdimc ,conca
     $     ,xdimf ,xdimc ,xarxa 
     $     ,tiparc ,tiparx 
     $     ,mdimf ,mdimc ,minima 
     $     ,can )
      implicit none

c     arguments
      integer*4 nnu ,nar ,nre ,ncc ,nax
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc) 
     $     ,xdimf ,xdimc ,xarxa(xdimf,xdimc) 
     $     ,mdimf ,mdimc 
      real*8		minima(mdimf,mdimc) ,can(*)
      character*1	tiparc(*) ,tiparx(*) 
      
c     local variables
      integer*4 i ,b ,k ,o ,n ,a ,aux_em1 ,aux_in1 
     $     ,aux_arc ,aux_ar2
c     body of routine
      do i=1,nax
         if ( (tiparx(i).eq.'b') .or. 
     $        (tiparx(i).eq.'c') .or. 
     $        (tiparx(i).eq.'w') .or. 
     $        (tiparx(i).eq.'x') .or. 
     $        (tiparx(i).eq.'s') .or. 
     $        (tiparx(i).eq.'t') .or. 
     $        (tiparx(i).eq.'u') .or. 
     $        (tiparx(i).eq.'r') .or. 
     $        (tiparx(i).eq.'d') .or. 
     $        (tiparx(i).eq.'e') .or. 
     $        (tiparx(i).eq.'f') .or.
     $        (tiparx(i).eq.'g') ) then
            
            o = aux_em1( nnu ,nre ,xarxa(i,1)) 
            n = aux_in1( nnu ,nre ,xarxa(i,1)) 
            a = aux_ar2( nar ,tiparc ,cdimf ,cdimc ,conca 
     $           ,o ,tiparx(i) )
            b = aux_arc( nar ,tiparc ,cdimf ,cdimc ,conca 
     $           ,o ,tiparx(i) )
            k = (n-1)*ncc + a
            can(k) = minima(b,n) 	
         end if
      end do
      return
      end
      

c***********************************************************************
c     generation of index vectors ivo(), lvo(), ivd(), lvd()
c     ivo(i),lvo(i): arcs of initial and final volums of the origin 
c     reservoir of the discharge or pumping arc i, in the multiinterval
c     network.
c     ivd(i),lvd(i): arcs of initial and final volumns of the 
c     destination reservoir of the discharge or pumping arc i, in 
c     the multiinterval network.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_ivo_lvo_ivd_lvd ( nnu ,nar ,nre ,ncc ,nax 
     $     ,cdimf ,cdimc ,conca
     $     ,xdimf ,xdimc ,xarxa 
     $     ,tiparc ,tiparx 
     $     ,ipl ,ivo ,lvo ,ivd ,lvd )
      implicit none

c     arguments
      integer*4 nnu ,nar ,nre ,nax ,ncc 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc) 
     $     ,xdimf ,xdimc ,xarxa(xdimf,xdimc) 
     $     ,ipl(*) 
     $     ,ivo(*) ,lvo(*) ,ivd(*) ,lvd(*) 
      character*1	tiparc(*) ,tiparx(*)
      
c     local variables
      integer*4 k ,i ,j ,in ,em ,a 
     $     ,aux_in1 ,aux_em1 ,aux_ar2
      
c     body of routine
      do i=1,nax
         if ( (tiparx(i).eq.'b') .or. 
     $        (tiparx(i).eq.'c') .or. 
     $        (tiparx(i).eq.'w') .or. 
     $        (tiparx(i).eq.'x') .or. 
     $        (tiparx(i).eq.'s') .or. 
     $        (tiparx(i).eq.'t') .or. 
     $        (tiparx(i).eq.'u') .or. 
     $        (tiparx(i).eq.'r') .or. 
     $        (tiparx(i).eq.'d') .or. 
     $        (tiparx(i).eq.'e') .or. 
     $        (tiparx(i).eq.'f') .or.
     $        (tiparx(i).eq.'g') ) then
            
            em = aux_em1( nnu ,nre ,xarxa(i,1)) 
            in = aux_in1( nnu ,nre ,xarxa(i,1)) 
            a  = aux_ar2( nar ,tiparc ,cdimf ,cdimc ,conca 
     $           ,em ,tiparx(i) )
            j = (in-1)*ncc + a
            
c     calcul de ivo(j)
            if (in.eq.1) then
               ivo(j) = - em
            else
               k = ipl(xarxa(i,1))
               do while (tiparx(k).ne.'m') 
                  k = k + 1
               end do
               ivo(j) = k
            end if
            
c     calcul de lvo(j)
            if (in.eq.nre) then
               lvo(j) = - em
            else
               k = ipl(xarxa(i,1)+nnu)
               do while (tiparx(k).ne.'m') 
                  k = k + 1
               end do
               lvo(j) = k
            end if
            
c     calcul de ivd(j)
            if ( (tiparx(i).eq.'d').or.
     $           (tiparx(i).eq.'e').or.
     $           (tiparx(i).eq.'f').or.
     $           (tiparx(i).eq.'g')     ) then
               ivd(j) = 0
            else
               in = aux_in1( nnu ,nre ,xarxa(i,2)) 
               em = aux_em1( nnu ,nre ,xarxa(i,2)) 
               if (in.eq.1) then
                  ivd(j) = - em
               else
                  k = ipl(xarxa(i,2))
                  do while (tiparx(k).ne.'m') 
                     k = k + 1
                  end do
                  ivd(j) = k
               end if
            end if
            
c     calcul de lvd(j)
            if ( (tiparx(i).eq.'d').or.
     $           (tiparx(i).eq.'e').or.
     $           (tiparx(i).eq.'f').or.
     $           (tiparx(i).eq.'g')     ) then
               lvd(j) = 0
            else
               if (in.eq.nre) then
                  lvd(j) = - em
               else
                  k = ipl(xarxa(i,2)+nnu)
                  do while (tiparx(k).ne.'m') 
                     k = k + 1
                  end do
                  lvd(j) = k
               end if
            end if
            
         end if
      end do
      
      return
      end
      
      
c***********************************************************************
c     generation of vectors rin() and rfl() of initial and final volums
c     accumulated by commodity, for all the reservoirs
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_rin_rfl ( nnu ,nks ,vdimf ,vdimc ,volini ,volfin 
     $     ,lin ,lfl)
      implicit none
      
c     arguments
      integer*4 nnu ,nks ,vdimf ,vdimc
      real*8		volini(vdimf,vdimc) ,volfin(vdimf,vdimc) 
     $     ,lin(*) ,lfl(*)
      
c     local variables
      integer*4 i ,k ,ix ,l
      
c     body of routine
      ix = 1
      do k=1,nks
         do i=1,nnu
            lin(ix) = 0.
            lfl(ix) = 0.
            ix = ix + 1
         end do
      end do
      do k=1,nks
         do i=1,nnu
            ix = (k-1)*nnu+i
            do l=1,nks-k+1
               lin(ix) = lin(ix) + volini(k,i)
               lfl(ix) = lfl(ix) + volfin(k,i)
               ix = ix + nnu
            end do  	      
         end do
      end do
      return
      end
      

c***********************************************************************
c     generation of the vectors jeo() jed() of origin and 
c     destinations reservoirs of the discharge and pumping arcs
c     of the hydro system.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_jeo_jed( nar ,cdimf ,cdimc ,conca ,tiparc 
     $     ,jeo ,jed )      
      implicit none
      
c     arguments
      integer*4 nar ,cdimf ,cdimc ,conca(cdimf,cdimc) 
     $     ,jeo(*) ,jed(*) 
      character*1	tiparc(*)
      
c     local variables
      integer*4 i ,j
      
c     body of routine
      j  = 0
      do i=1,nar
         if ( (tiparc(i).eq.'b') .or. 
     $        (tiparc(i).eq.'c') .or. 
     $        (tiparc(i).eq.'w') .or. 
     $        (tiparc(i).eq.'x') .or. 
     $        (tiparc(i).eq.'s') .or. 
     $        (tiparc(i).eq.'t') .or. 
     $        (tiparc(i).eq.'u') .or. 
     $        (tiparc(i).eq.'r') .or. 
     $        (tiparc(i).eq.'d') .or. 
     $        (tiparc(i).eq.'e') .or. 
     $        (tiparc(i).eq.'f') .or.
     $        (tiparc(i).eq.'g') ) then
            j = j + 1
            jeo(j) = conca(i,1)
            if ( (tiparc(i).eq.'d').or.
     $           (tiparc(i).eq.'e').or. 
     $           (tiparc(i).eq.'f').or. 
     $           (tiparc(i).eq.'g') ) then
               jed(j) = 0
            else
               jed(j) = conca(i,2)
            end if
         end if
      end do
      return
      end
      
      
c***********************************************************************
c     generation of the vector snq() of head differences between
c     discharges.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_snq( nar ,cdimf ,cdimc ,conca ,tiparc 
     $     ,cvb ,ctx ,sbx ,snq )
      implicit none
      
c     arguments
      integer*4 nar 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc)
      real*8		cvb(*) ,ctx(*) ,sbx(*) ,snq(*) 
      character*1	tiparc(*)
      
c     local variables
      integer*4 i ,k ,o ,d
      
c     body of routine
      k = 1
      do i=1,nar
         o = conca(i,1)
         d = conca(i,2)
         if ( (tiparc(i).eq.'d').or.
     $        (tiparc(i).eq.'e').or.
     $        (tiparc(i).eq.'f').or.
     $        (tiparc(i).eq.'g')    ) then 
            snq(k) = cvb(o) - ( ctx(o) - sbx(i) )
            k = k + 1
         else if ( (tiparc(i).eq.'s').or.
     $           (tiparc(i).eq.'t').or.
     $           (tiparc(i).eq.'u').or.
     $           (tiparc(i).eq.'r').or.
     $           (tiparc(i).eq.'b').or.
     $           (tiparc(i).eq.'c').or.
     $           (tiparc(i).eq.'w').or.
     $           (tiparc(i).eq.'x')    ) then 
            snq(k) = cvb(o) - cvb(d) 
            k = k + 1
         end if
      end do
      
      return
      end
      
      
c***********************************************************************
c     generation of vector jll() for identifying pumping arcs between
c     those of type b, s or d.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_jll (nar ,cdimf ,cdimc ,conca ,tiparc ,jll)
      implicit none
      
c     arguments
      integer*4 nar ,jll(*)
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc) 
      character*1	tiparc(*)
      
c     local variables
      integer*4 i ,k 
      
c     body of routine
      k = 0
      do i=1,nar
         if ( (tiparc(i).eq.'b').or.
     $        (tiparc(i).eq.'c').or.
     $        (tiparc(i).eq.'w').or.
     $        (tiparc(i).eq.'x')   ) then
            k = k + 1
            jll(k) = k
         else if ( (tiparc(i).eq.'d').or.
     $           (tiparc(i).eq.'e').or.
     $           (tiparc(i).eq.'f').or.
     $           (tiparc(i).eq.'g').or.
     $           (tiparc(i).eq.'s').or.
     $           (tiparc(i).eq.'t').or.
     $           (tiparc(i).eq.'u').or.
     $           (tiparc(i).eq.'r')    ) then
            k = k + 1
            jll(k) = -k
         end if
      end do
      return
      end
      

c***********************************************************************
c     generation of vector jca() with the numbers of the arcs in
c     the multiinterval network of the arcs b, s or d
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_jca ( nnu ,nar ,nre ,ncc ,nax 
     $     ,cdimf ,cdimc ,conca
     $     ,xdimf ,xdimc ,xarxa
     $     ,tiparc ,tiparx ,jca )
      implicit none
      
c     arguments
      integer*4 nnu ,nar ,nre ,nax ,ncc
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc)
     $     ,xdimf ,xdimc ,xarxa(xdimf,xdimc)
     $     ,jca(nar*nre)
      character*1	tiparc(*) ,tiparx(*)
      
c     local variables
      integer*4 i ,k ,n ,a ,o ,aux_ar2 ,aux_in1 ,aux_em1
      
c     body of routine
      do i=1,nax
         if ( (tiparx(i).eq.'b') .or.
     $        (tiparx(i).eq.'c') .or.
     $        (tiparx(i).eq.'w') .or.
     $        (tiparx(i).eq.'x') .or.
     $        (tiparx(i).eq.'d') .or.
     $        (tiparx(i).eq.'e') .or.
     $        (tiparx(i).eq.'f') .or.
     $        (tiparx(i).eq.'g') .or.
     $        (tiparx(i).eq.'s') .or.
     $        (tiparx(i).eq.'t') .or.
     $        (tiparx(i).eq.'u') .or.
     $        (tiparx(i).eq.'r') ) then
            o = aux_em1( nnu ,nre ,xarxa(i,1) )
            n = aux_in1( nnu ,nre ,xarxa(i,1) )
            a = aux_ar2( nar ,tiparc ,cdimf ,cdimc ,conca 
     $           ,o ,tiparx(i) )
            k = (n-1)*ncc + a
            jca(k) = i
         end if
      end do

      return
      end
      
c***********************************************************************
c     generation of vector ivv() of spillage arcs in the multiinterval
c     network, and vector mfv() of arcs of final volumes of the origin
c     reservoirs of the spillage arcs in the multiinvertal network
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_ivv_mfv ( nnu ,nar ,nre ,nvv ,nax 
     $     ,cdimf ,cdimc ,conca
     $     ,xdimf ,xdimc ,xarxa
     $     ,tiparc ,tiparx ,ipl ,ivv ,mfv )
      implicit none
      
c     arguments
      integer*4 nnu ,nar ,nre ,nvv ,nax
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc) 
     $     ,xdimf ,xdimc ,xarxa(xdimf,xdimc) 
     $     ,ipl(*) ,ivv(*) ,mfv(*)
      character*1	tiparc(*) ,tiparx(*) 
      
c     local variables
      integer*4 i ,j ,k ,o ,n ,a ,aux_em1 ,aux_in1 ,aux_ar3
c     body of routine
      do i=1,nax
         if ( (tiparx(i).eq.'v') .or.
     $        (tiparx(i).eq.'o') .or.
     $        (tiparx(i).eq.'p') .or.
     $        (tiparx(i).eq.'q') ) then
            o = aux_em1( nnu ,nre ,xarxa(i,1) )
            n = aux_in1( nnu ,nre ,xarxa(i,1) )
            a = aux_ar3( nar ,tiparc ,cdimf ,cdimc ,conca 
     $           ,o ,tiparx(i) )
            k = (n-1)*nvv + a
            ivv(k) = i
            if (n.eq.nre) then 
               mfv(k) = -o
            else
               j = ipl(xarxa(i,1)+nnu)
               do while (tiparx(j).ne.'m')
                  j = j + 1
               end do
               mfv(k) = j
            end if
         end if
      end do
      return
      end
      

c***********************************************************************
c     generation of the vector cxv() of maximum capacities of the 
c     origin reservoirs for the spillage arcs, and the vector iov()
c     of origin reservoirs for the spillage arcs.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_cxv_iov ( nar ,cdimf ,cdimc ,conca
     $     ,tiparc ,vlx ,cxv ,iov )
      implicit none
      
c     arguments
      integer*4 nar ,cdimf ,cdimc ,conca(cdimf,cdimc) 
     $     ,iov(*)
      real*8		vlx(*) ,cxv(*) 
      character*1	tiparc(*) 
      
c     local variables
      integer*4 i ,k 
c     body of routine
      k = 1
      do i=1,nar
         if ( (tiparc(i).eq.'v') .or.
     $        (tiparc(i).eq.'o') .or.
     $        (tiparc(i).eq.'p') .or.
     $        (tiparc(i).eq.'q') ) then
            cxv(k) = vlx( conca(i,1) ) + 0.01
            iov(k) = conca(i,1)
            k = k + 1
         end if
      end do
      return
      end
      

c***********************************************************************
c     generation of the vector jcl() of codes (1,2,3) of the discharge
c     and pumping arcs of the hydro system.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_jcl ( nar ,tiparc ,jcl)
      implicit none
      
c     arguments
      integer*4 nar ,jcl(*)
      character*1	tiparc(*) 
      
c     local variables
      integer*4 i ,k 
c     body of routine
      
c     body of routine
      k = 0
      do i=1,nar
         k = k + 1
         if      ( (tiparc(i).eq.'d').or.
     $        (tiparc(i).eq.'s').or.
     $        (tiparc(i).eq.'b')   ) then
            jcl(k) = 1
         else if ( (tiparc(i).eq.'e').or.
     $           (tiparc(i).eq.'t').or.
     $           (tiparc(i).eq.'c')    ) then
            jcl(k) = 2
         else if ( (tiparc(i).eq.'f').or.
     $           (tiparc(i).eq.'u').or.
     $           (tiparc(i).eq.'w')    ) then
            jcl(k) = 3
         else if ( (tiparc(i).eq.'g').or.
     $           (tiparc(i).eq.'r').or.
     $           (tiparc(i).eq.'x')    ) then
            jcl(k) = 4
         else 
            k = k - 1
         end if
      end do
      
      return
      end


c***********************************************************************
c     generation of the vector von() derived from minime() with
c     minimum values for reservoir and interval.
c     See the calling routine, routine lec_conca and file htcoor_f.h 
c     for an explanation of the parameters
c***********************************************************************
      subroutine gen_von (neb, nint, MAXNEB, MAXINT, minime, von)
      implicit none
c     arguments
      integer*4	neb, nint, MAXNEB, MAXINT
      real*8 minime(MAXNEB,MAXINT), von(neb*(nint+1))
      
c     local variables
      integer*4 i,j,ct 
      
c     body of routine
      do i= 1,neb
         von(i)= 0.0d0
      end do
      ct= neb+1
      do i= 1,nint-1
         do j= 1,neb
            von(ct)= minime(j,i)
            ct= ct+1
         end do
      end do
      do i= 1,neb
         von(ct)= 0.0d0
         ct=ct+1
      end do
      end
      

c***********************************************************************
c     auxiliary routine for computing the number of the reservoir
c     associated with the node 'nus' in the multiinterval network
c***********************************************************************
      integer*4 function aux_em1 ( nnu ,nre ,nus  )
      implicit none
c     arguments
      integer*4	nnu ,nre ,nus 
      
c     local variables
      integer*4 i
      
c     body of the routine
      if (nus.eq.((nnu*nre)+1)) then
         aux_em1 = nus
      else
         i = int( nus / nnu )
         if (i*nnu.eq.nus) i = i - 1
         aux_em1 = nus - ( i * nnu ) 
      end if
      return
      end

      
c***********************************************************************
c     auxiliary routine for computing the interval associated to 
c     node 'nus' in the multiinterval network
c***********************************************************************
      integer*4 function aux_in1 ( nnu ,nre ,nus )      
      implicit none
c     arguments
      integer*4	nnu ,nre ,nus 
      
c     local variables
      integer*4	i
      
c     body of the routine
      if (nus.eq.((nnu*nre)+1)) then
         aux_in1 = 0
      else
         i = int( nus / nnu )
         if (i*nnu.ne.nus) i = i + 1
         aux_in1 = i
      end if
      return
      end
      

c***********************************************************************
c     auxiliary routine to test if the arc of type 'tarc' leaving
c     from reservoir 'emd' exists, and returns the number of arc
c     in the hydro system
c***********************************************************************
      integer*4 function aux_arc( nar ,tiparc ,cdimf ,cdimc ,conca 
     $     ,emd ,tarc )      
      implicit none
      
c     arguments
      integer*4	nar ,emd 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc)
      character*1	tiparc(*) ,tarc
      
c     local variables
      integer*4	i
      
c     body of the routine
      aux_arc = 0
      i = 1
      do while (i.le.nar)
         if ( (tiparc(i).eq.tarc) .and.
     $        (conca(i,1).eq.emd)       ) then
            aux_arc = i
            i = nar
         end if
         i = i + 1
      end do
      return
      end
        

c***********************************************************************
c     auxiliary routine to test if the arc of type 'tarc' leaving
c     from reservoir 'emd' exists, and returns the number of the
c     arc in the hydro system, but considering only the arcs of type 
c     'd,e,f,g, s,t,u,r, b,c'
c***********************************************************************
      integer*4 function aux_ar2( nar ,tiparc ,cdimf ,cdimc ,conca 
     $     ,emd ,tarc )
      implicit none
      
c     arguments
      integer*4	nar ,emd 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc)
      character*1	tiparc(*) ,tarc
      
c     local variables
      integer*4	i ,j
      
c     body of the routine
      aux_ar2 = 0
      j = 0
      i = 1
      do while (i.le.nar)
         if ( (tiparc(i).eq.'d').or.
     $        (tiparc(i).eq.'e').or.
     $        (tiparc(i).eq.'f').or.
     $        (tiparc(i).eq.'g').or.
     $        (tiparc(i).eq.'b').or.
     $        (tiparc(i).eq.'c').or.
     $        (tiparc(i).eq.'w').or.
     $        (tiparc(i).eq.'x').or.
     $        (tiparc(i).eq.'s').or.
     $        (tiparc(i).eq.'t').or.
     $        (tiparc(i).eq.'u').or.
     $        (tiparc(i).eq.'r')    ) j = j + 1
         
         if ( (tiparc(i).eq.tarc) .and.
     $        (conca(i,1).eq.emd)       ) then
            aux_ar2 = j
            i = nar
         end if
         i = i + 1
      end do
      return
      end


c***********************************************************************
c     auxiliary routine to test if the arc of type 'tarc' leaving 
c     from 'emd' exists, and returns the number of the arc in the
c     hydro system, but considering only the arcs of type 'v'
c***********************************************************************
      integer*4 function aux_ar3( nar ,tiparc ,cdimf ,cdimc ,conca 
     $     ,emd ,tarc )
      implicit none
      
c     arguments
      integer*4	nar ,emd 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc)
      character*1	tiparc(*) ,tarc
      
c     local variables
      integer*4	i ,j
      
c     body of the routine
      aux_ar3 = 0
      j = 0
      i = 1
      do while (i.le.nar)
         if ( (tiparc(i).eq.'v') .or.
     $        (tiparc(i).eq.'o') .or.
     $        (tiparc(i).eq.'p') .or.
     $        (tiparc(i).eq.'q') ) j = j + 1
         if ( (tiparc(i).eq.tarc) .and.
     $        (conca(i,1).eq.emd)       ) then
            aux_ar3 = j
            i = nar
         end if
         i = i + 1
      end do
      return
      end
      
