c***********************************************************************
c     Routine for computing dimensions of the optimization problem
c     
c     INPUT parameters:
c     
c     OUTPUT parameters:
c     m: number of constraints
c     n: number of variables
c     nb: number of bounds (m+n)
c     nncon: number of nonlinear variables
c     nnobj: number of nonlinear objective variables
c     nnjac: number of nonlinear jacobian variables
c     ne: number of nonzero entries in jacobian (for linear and
c         nonlinear constraints)
c     iexit: if iexit= 0 routine exited normally
c***********************************************************************
      subroutine generate_problem_dimensions(m,n,nb,nncon,nnobj,
     $     nnjac,ne,iexit)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 m,n,nb,nncon,nnobj,nnjac,ne,iexit
c     local variables
      integer*4 iaux,i,j,neb_with_discharge
      logical trobat

      iexit= 0

c     1. compute m and nncon: number of constraints and nonlinear ones
c
c     First nonlinear constraints
c
      nncon= 0
c     
c     1.1 nonlinear constraints to store in HDkiii (interval i, comm. k)
c     hydro generations per commodity and interval
      nncon= nncon+nint*nat
c     1.2 nonlinear constraints to store in PGtttiii power generated by
c     each thermal unit at each interval
      nncon= nncon+nint*nterm
c     1.3 nonlinear constraints for coverage of the g.d.c (COiii) 
c     for each interval
      nncon= nncon+nint
c
c     Second linear constraints
c
      m=0
c     
c     1.4 constraints HINTiii related to hydro energy inserted between 
c     thermal units (one per interval)
      m= m+nint
c     1.5 for hydro system: a balance equation for each commodity at
c     the sink node (nat), plus a mutual capacity constraint for
c     each arc in the multiinterval network (nax), plus a balance
c     equation for each commodity at each reservoir for all the
c     intervals (nat*neb*nint) (constraints B and T).
      m= m+nat+nax+nat*neb*nint
c     1.6 balance equations of purchase of fuels XCc
      m= m+ncomb
c     1.7 balance equations for the thermal system Xctttiii, per fuel,
c     thermal unit, and interval
      m= m+nthrmarticles*nint
c     1.8 equations PGMiii to guarantee that thermal units, plus 
c     deterministic and stochastic hydro generation provide a minimum
c     power
      m= m+nint
c     1.9 equations PHDiii to guarantee that power hydro generation 
c     (stochastic+deterministic) is always below a maximum value
      m= m+nint
c
c     add nonlinear to linear to get total number of constraints
      m= m+nncon


      n= 0
c     2. compute n: number of variables
c     2.1 number of arcs in the multiinterval hydro network (nax) by 
c     the number of commodities
      n= n+nax*nat
c     2.2 variables of change of variable HDkiii (hydro-energy at 
c     interval i, commodity k)
      n= n+nint*nat
c     2.3 variables of hydro energy inserted between thermal units 
c     HItttiii (hydro-inserted-energy at interval i, over 
c     thermal unit j)
      n= n+nint*nterm
c     add variables related with thermal system
c     2.4 fuel purchased by each thermal unit at each interval for each
c       fuel Cctttiii = nint*nthrmarticles
c     2.5 amount not purchased for each fuel: Nc = ncomb
c     2.6 fuel not consumed at last interval by each thermal unit:
c         Rcttt = nthrmarticles
c     2.7 available thermal unit for next interval for each thermal 
c         unit, each interval and each fuel: 
c         Zctttiii = (nint-1)*nthrmarticles
c     2.8 generation variables at each interval for each thermal 
c         unit, with each fuel:  Gctttiii = nint*nthrmarticles
      n= n + 3*nint*nthrmarticles+ncomb
c     2.9 variables with power generated by each thermal unit at each
c       interval PGtttiii = nint*nterm
      n= n+nint*nterm
c     2.10 variables of power generated by stochastic hydro-energy
c       PHDiii = nint
      n= n+nint

c     all the variables are considered nonlinear in the objective
c     function (as well as in the jacobian)
      nnobj= n
      nnjac= n

c     number of variables plus slacks
      nb= n+m

      ne= 0
c     3. number of nonzero entries in jacobian (considering linear and 
c     nonlinear constraints)
c     3.1 arcs of the multiinterval hydro network V,D,M, etc affect to:
c     - linear constraints: each each arc has 2 entries, one for each
c     balance equations at origin and destination nodes, and 1 entry at
c     mutual capacity constraints.
      ne= ne+nax*nat*3
c     - nonlinear constraints: each generation arc (discharge+pumping)
c     variables Dkeeeiii affect to constraints HDliii for l= k to nat 
c     (nat-k+1 elements for each D/Bkeeeiii) 
      ne= ne+nint*ncc*nat*(nat+1)/2
c     variables Mkeeeiii (iii= 1 to nint-1) affect to HDliii and 
c     HDliii+1 for l= k to nat, if there exists a generation arc
c     with origin at eee or if there is some pumping arc with origin
c     or destination at eee (nat-k+1 elements for each suitable D/B keeeiii)
c     we must count the number of reservoirs with leaving discharges or 
c     leaving/arriving pumpings
c     (no "solapament" arcs are considered)
      neb_with_discharge= 0
      do j= 1,neb
         i=1
         trobat= .false.
         do while ((i.le.nar).and.(.not.trobat))
            trobat= ((conca(i,1).eq.j) .and. 
     $           (index('defgsturbcwx',tiparc(i)).ne.0))
     $           .or.
     $           ((conca(i,2).eq.j).and.(index('bcwx',tiparc(i)).ne.0))
            i= i+1
         end do
         if (trobat) neb_with_discharge= neb_with_discharge+1
      end do
      ne= ne+(nint-1)*neb_with_discharge*nat*(nat+1)
c      write(*,*) '3.1', ne
c     3.2 variables HDkiii affect to:
c     if k=1: constraints HDkiii (1 element), PGtttiii (nterm elements),
c     COiii (1 element), PGMiii (1 element), PHD (1 element)
c     if k>1: constraints HDkiii (1 element), HINTiii (1 element)
      ne= ne+nint*(4+nterm)+(nint*(nat-1))*2
c      write(*,*) '3.2', ne
c     3.3 variables HItttiii affect to: COiii (1 element), 
c     HINTiii (1 element), PGjjjiii for jjj= ttt+1 to nterm 
c     (nt*(nt-1)/2 elements)
      ne= ne+nint*(nterm*2+nterm*(nterm-1)/2)
c      write(*,*) '3.3', ne
c     3.4 variables Cctttiii: since they are arcs of the thermal
c     system, only have 2 entries (constraints XC and X)
      ne= ne+2*nthrmarticles*nint
c      write(*,*) '3.4', ne
c     3.5 variables Nc have 1 entry at constraints XC
      ne= ne+ncomb
c      write(*,*) '3.5', ne
c     3.6 variables Rcttt have 1 entry at constraints Xctttiii where
c     iii is the last interval
      ne= ne+nthrmarticles
c      write(*,*) '3.6', ne
c     3.7 variables Zctttiii, for iii=1 to nint-1 (we exclude last
c     interval) have two entries at constraints Xctttiii
c     (since they are arcs of the multiinterval thermal network)
      ne= ne+2*nthrmarticles*(nint-1)
c      write(*,*) '3.7', ne
c     3.8 variables Gctttiii affect to PGjjjiii for jjj= ttt to nterm
c     (computed through iaux), COiii (1 element), Xctttiii (1 element)
      iaux= 0
      do j=1,nterm
         iaux= iaux+nartterm(j)*(nterm-j+1)
      end do
      ne= ne+nint*(nthrmarticles*2+iaux)
c      write(*,*) '3.8', ne
c     3.9 variables PGtttiii affect to PGjjjiii for jjj= 1 to nterm
c     (nterm elements), COiii (1 element), PGMiii (1 element)
      ne= ne+nint*nterm*(nterm+2)
c      write(*,*) '3.9', ne
c     3.10 variables PHDiii affect to PGtttiii for ttt= 1 to nterm
c     (nterm elements), COiii (1 element), PGMiii (1 element),
c     PHDiii (1 element)
      ne= ne+nint*(nterm+3)
c      write(*,*) '3.10', ne

      end 


c***********************************************************************
c     Routine to fill minos vectors with model information
c     
c     INPUT parameters:
c     m: number of constraints
c     n: number of variables
c     nb: number of bounds (m+n)
c     nncon: number of nonlinear variables
c     nnobj: number of nonlinear objective variables
c     nnjac: number of nonlinear jacobian variables
c     ne: number of nonzero entries in jacobian (for linear and
c         nonlinear constraints)
c     
c     OUTPUT parameters:
c     a(ne): constraint matrix stored columnwise
c     ha(ne): row indices for each nonzero in a() 
c     ka(n+1): pointers to the beginning of each column in a() and ha()
c     bl(nb): lower bounds for variables and slacks
c     bu(nb): upper bounds for variables and slacks
c     hs(nb): estimate and workspace for initial state for variables 
c     pi(m): initial estimate and workspace for multipliers 
c     xn(nb): initial estimates of variables and slacks
c     name1(nb): first 4 char positions of the names of vars and const
c     name2(nb): last 4 fchar positions of the names of vars and const
c     ne_njac: number of nonlinear entries of the jacobian
c     iexit: if iexit= 0 routine exited normally
c***********************************************************************
      subroutine generate_minos_model(m,n,nb,nncon,nnobj,nnjac,ne,
     $     ne_njac,a,ha,ka,bl,bu,hs,pi,xn,name1,name2,iexit)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 m,n,nb,nncon,nnobj,nnjac,ne,ne_njac,
     $     ha(ne),ka(n+1),hs(nb),iexit 
      real*8 a(ne),bl(nb),bu(nb),xn(nb),pi(m)
      character*4 name1(nb),name2(nb)


      iexit= 0

c     generation of names of variables and constraints
      call generate_constraints_names(n,nb,name1,name2)
      call generate_variables_names(n,nb,name1,name2)

c     generation of constraints structure (a,ha,ka)
      call generate_constraints_structure(m,n,ne,ne_njac,a,ha,ka)
      
c     generation of lower and upper bounds for variables and slacks
      call generate_bounds(m,n,nb,bl,bu)

c     generation of initial point
      call generate_init_point(n,nb,m,nncon,hs,xn,pi,bl,bu)

      end


c***********************************************************************
c     generation of constraints names
c
c     INPUT parameters:
c     n,nb
c     
c     OUTPUT parameters:
c     name1,name2
c***********************************************************************
      subroutine generate_constraints_names(n,nb,name1,name2)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 n,nb
      character*4 name1(nb),name2(nb)
c     functions
      character*1 codi_comb
      character*1 artic
      character*3 inter,aux_emb,aux_int
c     local variables
      integer*4 i,j,k,ct
      character*8 str

      ct= n
c     1.1 constraints HDkiii
      do k= 1,nat
         do i= 1,nint
            str= 'hd'//artic(k)//inter(i)
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
         end do
      end do
c     1.2 constraints PGtttiii
      do i= 1,nint
         do j= 1,nterm
            str= 'pg'//idterm(j)//inter(i)
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
         end do
      end do
c     1.3 constraints COiii
       do i= 1,nint
         str= 'co'//inter(i)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
c     1.4 constraints HINTiii
      do i= 1,nint
         str= 'hint'//inter(i)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
c     1.5 balance and mutual capacity equations of hydro system
      do k= 1,nat
         do i= 1,nint
c     balances at reservoirs
            do j= 1,neb
               str= 'b'//artic(k)//embas(j) //inter(i)
               ct= ct+1
               name1(ct)= str(1:4)
               name2(ct)= str(5:8)
            end do
         end do
c     balance at sink
         str= 'b'//artic(k)//embas(neb+1)//'000'
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
c     mutual capacities of hydro arcs
      do i= 1,nax
         str= 't'//tiparx(i)//aux_emb(neb,nint,xarxa(i,1),embas)//
     $        aux_int(neb,nint,xarxa(i,1))
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
      
c     1.6 constraints XC
      do i= 1,ncomb
         str= 'xc'//codi_comb(i)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
c     1.7 constraints Xctttiii
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               str= 'x'//codi_comb(artterm(j,k))//idterm(j)//inter(i)
               ct= ct+1
               name1(ct)= str(1:4)
               name2(ct)= str(5:8)
            end do
         end do
      end do
c     1.8 constraints PGMiii
      do i= 1,nint
         str= 'pgm'//inter(i)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
c     1.9 constraints PHDiii
      do i= 1,nint
         str= 'phd'//inter(i)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do

      end


c***********************************************************************
c     generation of variables names
c
c     INPUT parameters:
c     n,nb
c     
c     OUTPUT parameters:
c     name1,name2
c***********************************************************************
      subroutine generate_variables_names(n,nb,name1,name2)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 n,nb
      character*4 name1(nb),name2(nb)
c     functions
      character*1 codi_comb
      character*1 artic
      character*3 inter,term,aux_emb,aux_int
c     local variables
      integer*4 i,j,k,ct
      character*8 str
      
      ct= 0
c     2.1 hydro network variables
      do k= 1,nat
         do i= 1,nax
            str= tiparx(i)//artic(k)//
     $           aux_emb(neb,nint,xarxa(i,1),embas)//
     $           aux_int(neb,nint,xarxa(i,1))
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
        end do
      end do
c     2.2 variables HDkiii
      do i= 1,nint
         do k= 1,nat
            str= 'hd'//artic(k)//inter(i) 
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
         end do
      end do
c     2.3 variables HItttiii
      do i= 1,nint
         do j= 1,nterm
            str= 'hi'//term(j)//inter(i)
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
         end do
      end do
c     2.4 variables Cctttiii
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               str= 'c'//codi_comb(artterm(j,k))//idterm(j)//inter(i) 
               ct= ct+1
               name1(ct)= str(1:4)
               name2(ct)= str(5:8)
            end do
         end do
      end do
c     2.5 variables Nc
      do k= 1,ncomb
         str= 'n'//codi_comb(k)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
c     2.6 variables Rcttt
      do j= 1,nterm
         do k= 1,nartterm(j)
            str= 'r'//codi_comb(artterm(j,k))//idterm(j)
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
         end do
      end do
c     2.7 variables Zctttiii
      do i= 1,nint-1
         do j= 1,nterm
            do k= 1,nartterm(j)
               str= 'z'//codi_comb(artterm(j,k))//idterm(j)//inter(i)
               ct= ct+1
               name1(ct)= str(1:4)
               name2(ct)= str(5:8)
            end do
         end do
      end do
c     2.8 variables Gctttiii
       do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               str= 'g'//codi_comb(artterm(j,k))//idterm(j)//inter(i)
               ct= ct+1
               name1(ct)= str(1:4)
               name2(ct)= str(5:8)
            end do
         end do
      end do
c     2.9 variables PGtttiii
       do i= 1,nint
         do j= 1,nterm
            str= 'pg'//idterm(j)//inter(i)
            ct= ct+1
            name1(ct)= str(1:4)
            name2(ct)= str(5:8)
         end do
      end do
c     2.10 variables PHDiii
      do i= 1,nint
         str= 'phd'//inter(i)
         ct= ct+1
         name1(ct)= str(1:4)
         name2(ct)= str(5:8)
      end do
            
      end


c***********************************************************************
c     auxiliary routine to get the code (an ascii character) for
c     a given type of fuel
c***********************************************************************
      character*(*) function codi_comb(nc)
      implicit none
      integer*4 nc
      
      if ((nc.ge.1).and.(nc.le.10)) then
         codi_comb= char(48+nc-1)
      else if ((nc.ge.11).and.(nc.le.36)) then
         codi_comb= char(65+nc-11)
      else if ((nc.ge.37).and.(nc.le.62)) then
         codi_comb= char(97+nc-37)
      else if ((nc.ge.63).and.(nc.le.77)) then
         codi_comb= char(33+nc-63)
      else if ((nc.ge.78).and.(nc.le.84)) then
         codi_comb= char(58+nc-78)
      else if ((nc.ge.85).and.(nc.le.90)) then
         codi_comb= char(91+nc-85)
      else if ((nc.ge.91).and.(nc.le.94)) then
         codi_comb= char(123+nc-91)
      else
c     if nc>91 we use the character ' ', thus it is possible
c     to have repeated codes
         codi_comb= ' '
      end if
      end


c***********************************************************************
c     auxiliary routine to get the code (an ascii character) for
c     a hydro commodity
c***********************************************************************
      character*1 function artic(k)
      implicit none
      integer*4 k
      character*1 str
      
      call itostr(k,str,1)
      artic= str
      end


c***********************************************************************
c     auxiliary routine to get the code (three ascii character) for
c     an interval
c***********************************************************************
      character*3 function inter(i)
      implicit none
      integer*4 i
      character*3 str
      
      call itostr(i,str,3)
      inter= str
      end


c***********************************************************************
c     auxiliary routine to get the code (three ascii character) for
c     a thermal unit
c***********************************************************************
      character*3 function term(j)
      implicit none
      integer*4 j
      character*3 str
      
      call itostr(j,str,3)
      term= str
      end


c***********************************************************************
c     auxiliary routine to get the identifier (three ascii character) 
c     of the reservoir associated to node 'nus' in the multiinterval
c     hydro network
c***********************************************************************
      character*3 function aux_emb(neb,nint,nus,embas)
      implicit none
      integer*4 neb, nint, nus
      character*3 embas(*)
      integer*4 i

      if (nus.eq.((neb*nint)+1)) then
         aux_emb= embas(neb+1)
      else
         i= int( nus / neb )
         if (i*neb.eq.nus) i = i - 1
         aux_emb= embas(nus-(i*neb))
      end if
      end


c***********************************************************************
c     auxiliary routine to get the code (three ascii character) 
c     of the interval associated to node 'nus' in the multiinterval
c     hydro network
c***********************************************************************
      character*3 function aux_int(neb,nint,nus)
      implicit none
      integer*4 neb, nint, nus
      character*3 inter
      integer*4 i
      
      if (nus.eq.((neb*nint)+1)) then
         aux_int = '000'
      else
         i = int( nus / neb )
         if (i*neb.ne.nus) i = i + 1
         aux_int = inter(i)
      end if
      end
      
      
c***********************************************************************
c     generation of constraints names
c
c     INPUT parameters:
c     m,n,ne
c     
c     OUTPUT parameters:
c     a,ha,ka,ne_njac
c     ne_njac gives the number of entries in the nonlinear terms
c             of the jacobian
c***********************************************************************
      subroutine generate_constraints_structure(m,n,ne,ne_njac,a,ha,ka)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 m,n,ne,ne_njac,ha(ne),ka(n+1)
      real*8 a(ne)
c     functions
      integer*4 aux_in1, aux_em1
c     parameters
      real*8 NLIN_ENTRY
      parameter (NLIN_ENTRY= 0.0d0)
c     local variables
      integer*4 i,ii,j,jj,k,l,ct,ct_ne,ct_n,emb
      integer*4 ini_hd_ki, ini_pg_ti, ini_co_i, ini_hint_i,
     $     ini_b_kei, ini_t_a, ini_xc_c, ini_x_cti, ini_pgm_i, ini_phd_i
      logical trobat

c     First create row indices to the beginning of each group of
c     constraints. The order of the constraints must be the same
c     than that of routines generate_bounds and generate_constraints_
c     names
      ct= 0
      ini_hd_ki= ct
      ct= ct+nint*nat
      ini_pg_ti= ct
      ct= ct+nint*nterm 
      ini_co_i= ct
      ct= ct+nint
      ini_hint_i= ct
      ct= ct+nint
      ini_b_kei= ct
      ct= ct+nat*neb*nint+nat
      ini_t_a= ct
      ct= ct+nax
      ini_xc_c= ct
      ct= ct+ncomb
      ini_x_cti= ct
      ct= ct+nthrmarticles*nint
      ini_pgm_i= ct
      ct= ct+nint
      ini_phd_i= ct

c     The order of the variables must be the same than that 
c     of routines generate_bounds and generate_variables_names
c     
      ct_ne= 0
      ct_n= 0
      ne_njac= 0

c     3.1 entries of arcs of the multiinterval hydro network
      do k= 1,nat
         do i= 1,nax
c     new variable ka() pointer
            ct_n= ct_n+1
            ka(ct_n)= ct_ne+1
c     3.1.1 constraints HDliii (l= k to nat) for generation arcs
            if (index('defgsturbcwx',tiparx(i)).ne.0) then
               do l= k,nat
                  ne_njac= ne_njac+1
                  ct_ne= ct_ne+1
                  a(ct_ne)= NLIN_ENTRY
                  ha(ct_ne)= ini_hd_ki+(l-1)*nint+
     $                 aux_in1(neb,nint,xarxa(i,1))
               end do
c     3.1.2 constraints HDliii and HDliii+1 (l= k to nat) for
c     volume arcs Mkeeeiii (iii= 1 to nint-1; we guarantee that only
c     arcs until interval iii-1 are considered by using xarxa(i,2) when 
c     computing the reservoir 'emb' below), if there exists a 
c     generation arc with origin at eee or if there is some pumping 
c     arc with origin or destination at eee 
c     (no "solapament" arcs are considered)
            else if (tiparx(i).eq.'m') then
               emb= aux_em1(neb,nint,xarxa(i,2))
               ii= 1
               trobat= .false.
               do while ((ii.le.nar).and.(.not.trobat))
                  trobat= ((conca(ii,1).eq.emb) .and. 
     $                 (index('defgsturbcwx',tiparc(ii)).ne.0))
     $                 .or.
     $                 ((conca(ii,2).eq.emb).and.
     $                 (index('bcwx',tiparc(ii)).ne.0))
                  ii= ii+1
               end do
               if (trobat) then      
                  do l= k,nat
c     3.1.2.1 constraint HDliii
                     ne_njac= ne_njac+1
                     ct_ne= ct_ne+1
                     a(ct_ne)= NLIN_ENTRY
                     ha(ct_ne)= ini_hd_ki+(l-1)*nint+
     $                    aux_in1(neb,nint,xarxa(i,1))
c     3.1.2.2 constraint HDliii+1
                     ne_njac= ne_njac+1
                     ct_ne= ct_ne+1
                     a(ct_ne)= NLIN_ENTRY
                     ha(ct_ne)= ini_hd_ki+(l-1)*nint+
     $                    aux_in1(neb,nint,xarxa(i,2))
                  end do
               end if
            end if
c     3.1.3 constraint Bkeeeiii (where eee= origin reservoir)
            ct_ne= ct_ne+1
            a(ct_ne)= 1.0d0
            ha(ct_ne)= ini_b_kei+(k-1)*(nint*neb+1)+
     $           (aux_in1(neb,nint,xarxa(i,1))-1)*neb+
     $           aux_em1(neb,nint,xarxa(i,1))
c     3.1.4 constraint Bkdddiii (where ddd= destination reservoir)
c     or Bkeeeiii+1 (if this arc is a volume arc Mkeeeiii)
c     it contemplates also the case that the destination node was
c     the sink 
            if (xarxa(i,2).eq.(nint*neb+1) ) then
c     destination is sink node
               ct_ne= ct_ne+1
               a(ct_ne)= -1.0d0
               ha(ct_ne)= ini_b_kei+k*(nint*neb+1)
            else
c     destination is not sink node
               ct_ne= ct_ne+1
               a(ct_ne)= -1.0d0
               ha(ct_ne)= ini_b_kei+(k-1)*(nint*neb+1)+
     $              (aux_in1(neb,nint,xarxa(i,2))-1)*neb+
     $              aux_em1(neb,nint,xarxa(i,2))
            end if
c     3.1.5 constraint Tteeeiii
            ct_ne= ct_ne+1
            a(ct_ne)= 1.0d0
            ha(ct_ne)= ini_t_a+i
         end do
      end do
c      write(*,*) '3.1', ct_ne
      
c     3.2 entries of variables HDkiii
      do i= 1,nint
c     3.2.1 commodity 1. 
c     new variable ka() pointer
         ct_n= ct_n+1
         ka(ct_n)= ct_ne+1
c     3.2.1.1 constraint HDkiii
         ne_njac= ne_njac+1
         ct_ne= ct_ne+1
         a(ct_ne)= -1.0d0 !**Warning** it is a nonlinear constraint
         ha(ct_ne)= ini_hd_ki+i
c     3.2.1.2 constraints PGtttiii
         do j= 1,nterm
            ne_njac= ne_njac+1
            ct_ne= ct_ne+1
            a(ct_ne)= NLIN_ENTRY
            ha(ct_ne)= ini_pg_ti+(i-1)*nterm+j
         end do
c     3.2.1.3 constraint COiii
         ne_njac= ne_njac+1
         ct_ne= ct_ne+1
         a(ct_ne)= NLIN_ENTRY
         ha(ct_ne)= ini_co_i+i
c     3.2.1.4 constraint PGMiii
         ct_ne= ct_ne+1
         a(ct_ne)= 1.0d0/tbase(i)
         ha(ct_ne)= ini_pgm_i+i
c     3.2.1.5 constraint PHDiii
         ct_ne= ct_ne+1
         a(ct_ne)= 1.0d0/tbase(i)
         ha(ct_ne)= ini_phd_i+i
         
c     3.2.2 commodities from 2 to nat
         do k= 2,nat
c     new variable ka() pointer
            ct_n= ct_n+1
            ka(ct_n)= ct_ne+1
c     3.2.2.1 constraint HDkiii
            ne_njac= ne_njac+1
            ct_ne= ct_ne+1
            a(ct_ne)= -1.0d0 !**Warning** it is a nonlinear constraint
            ha(ct_ne)= ini_hd_ki+(k-1)*nint+i
c     3.2.2.2 constraint HINTiii
            ct_ne= ct_ne+1
            a(ct_ne)= probhd(k)
            ha(ct_ne)= ini_hint_i+i
         end do
      end do
c      write(*,*) '3.2', ct_ne

c     3.3 entries of variables HItttiii
      do i= 1,nint
         do j= 1,nterm
c     new variable ka() pointer
            ct_n= ct_n+1
            ka(ct_n)= ct_ne+1
c     3.3.1 constraints PGjjjiii (jjj= j+1,nterm)
            do jj= j+1,nterm
               ne_njac= ne_njac+1
               ct_ne= ct_ne+1
               a(ct_ne)= NLIN_ENTRY
               ha(ct_ne)= ini_pg_ti+(i-1)*nterm+jj
            end do
c     3.3.2 constraint COiii
            ne_njac= ne_njac+1
            ct_ne= ct_ne+1
            a(ct_ne)= NLIN_ENTRY
            ha(ct_ne)= ini_co_i+i
c     3.3.3 constraint HINTiii
            ct_ne= ct_ne+1
            a(ct_ne)= -1.0d0
            ha(ct_ne)= ini_hint_i+i
         end do
      end do
c      write(*,*) '3.3', ct_ne
         
c     3.4 entries of variables Cctttiii
      l= 0
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
c     new variable ka() pointer
               ct_n= ct_n+1
               ka(ct_n)= ct_ne+1
c     3.4.1 constraint XCc
               ct_ne= ct_ne+1
               a(ct_ne)= 1.0d0
               ha(ct_ne)= ini_xc_c+artterm(j,k)
c     3.4.2 constraint Xctttiii
               ct_ne= ct_ne+1
               a(ct_ne)= -1.0d0
               l= l+1
               ha(ct_ne)= ini_x_cti+l
            end do
         end do
      end do
c      write(*,*) '3.4', ct_ne
      
c     3.5 entries of variables Nc
      do k= 1,ncomb
c     new variable ka() pointer
         ct_n= ct_n+1
         ka(ct_n)= ct_ne+1
c     3.5.1 constraint XCc
         ct_ne= ct_ne+1
         a(ct_ne)= 1.0d0
         ha(ct_ne)= ini_xc_c+k
      end do
c      write(*,*) '3.5', ct_ne

c     3.6 entries of variables Rcttt
      l= 0
      do j= 1,nterm
         do k= 1,nartterm(j)
c     new variable ka() pointer
            ct_n= ct_n+1
            ka(ct_n)= ct_ne+1
c     3.6.1 constraint Xctttiii (where iii= nint)
            ct_ne= ct_ne+1
            a(ct_ne)= 1.0d0
            l= l+1
            ha(ct_ne)= ini_x_cti+(nint-1)*nthrmarticles+l
         end do
      end do
c      write(*,*) '3.6', ct_ne

c     3.7 entries of variables Zctttiii (iii= 1,nint-1)
      l= 0
      do i=1,nint-1
         do j= 1,nterm
            do k= 1,nartterm(j)
c     new variable ka() pointer
               ct_n= ct_n+1
               ka(ct_n)= ct_ne+1
c     3.7.1 constraint Xctttiii 
               ct_ne= ct_ne+1
               a(ct_ne)= 1.0d0
               l= l+1
               ha(ct_ne)= ini_x_cti+l
c     3.7.2 constraint Xcttt_iii+1
               ct_ne= ct_ne+1
               a(ct_ne)= -1.0d0
               ha(ct_ne)= ini_x_cti+nthrmarticles+l
            end do
         end do
      end do
c      write(*,*) '3.7', ct_ne

c     3.8 variables Gctttiii
      l= 0
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
c     new variable ka() pointer
               ct_n= ct_n+1
               ka(ct_n)= ct_ne+1
c     3.8.1 constraints PGjjjiii (jjj= j to nterm)
               do jj= j,nterm
                  ne_njac= ne_njac+1
                  ct_ne= ct_ne+1
                  a(ct_ne)= NLIN_ENTRY
                  ha(ct_ne)= ini_pg_ti+(i-1)*nterm+jj
               end do
c     3.8.2 constraint COiii
               ne_njac= ne_njac+1
               ct_ne= ct_ne+1
               a(ct_ne)= NLIN_ENTRY
               ha(ct_ne)= ini_co_i+i
c     3.8.3 constraint Xctttiii
               ct_ne= ct_ne+1
               a(ct_ne)= 100.0d0/rendiment_g(j,k)
               l= l+1
               ha(ct_ne)= ini_x_cti+l
            end do
         end do
      end do
c      write(*,*) '3.8', ct_ne
      
c     3.9 variables PGtttiii
      do i= 1,nint
         do j= 1,nterm
c     new variable ka() pointer
            ct_n= ct_n+1
            ka(ct_n)= ct_ne+1
c     3.9.1 constraints PGjjjiii (jjj= 1 to nterm)
            do jj= 1,nterm
               ne_njac= ne_njac+1
               ct_ne= ct_ne+1
               a(ct_ne)= NLIN_ENTRY
               ha(ct_ne)= ini_pg_ti+(i-1)*nterm+jj
            end do
c     3.9.2 constraint COiii 
            ne_njac= ne_njac+1
            ct_ne= ct_ne+1
            a(ct_ne)= NLIN_ENTRY
            ha(ct_ne)= ini_co_i+i
c     3.9.3 constraints PGMiii
            ct_ne= ct_ne+1
            a(ct_ne)= 1.0d0
            ha(ct_ne)= ini_pgm_i+i
         end do
      end do
c      write(*,*) '3.9', ct_ne
      
c     3.10 variables PHDiii
      do i= 1,nint
c     new variable ka() pointer
         ct_n= ct_n+1
         ka(ct_n)= ct_ne+1
c     3.10.1 constraints PGtttiii (ttt= 1 to nterm)
         do jj= 1,nterm
            ne_njac= ne_njac+1
            ct_ne= ct_ne+1
            a(ct_ne)= NLIN_ENTRY
            ha(ct_ne)= ini_pg_ti+(i-1)*nterm+jj
         end do
c     3.10.2 constraint COiii
         ne_njac= ne_njac+1
         ct_ne= ct_ne+1
         a(ct_ne)= NLIN_ENTRY
         ha(ct_ne)= ini_co_i+i
c     3.10.3 constraint PGMiii 
         ct_ne= ct_ne+1
         a(ct_ne)= 1.0d0
         ha(ct_ne)= ini_pgm_i+i
c     3.10.4 constraint PHDiii 
         ct_ne= ct_ne+1
         a(ct_ne)= 1.0d0
         ha(ct_ne)= ini_phd_i+i
      end do
c     update las pointer
      ka(ct_n+1)= ct_ne+1

c      write(*,*) '3.10', ct_ne

c      write(*,*) 'ne_njac=', ne_njac

      end


c***********************************************************************
c     generation of lower and upper bounds for variables and slacks
c
c     INPUT parameters:
c     n,nb
c     
c     OUTPUT parameters:
c     bl(nb),bu(nb)
c***********************************************************************
      subroutine generate_bounds(m,n,nb,bl,bu)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 m,n,nb
      real*8 bl(nb),bu(nb)
c     functions
      integer*4 aux_em1, aux_in1, aux_arc_mps
      real*8 aux_cp
c     parameters
      real*8 ZERO
      parameter (ZERO= 0.0d0)
c     local variables
      integer*4 i,j,k,l,ct,o,d
      real*8 acumpterm,cap1,cap2,racum
      character*1 tarc

c     bounds for variables
      
      ct= 0
c     2.1 hydro network variables
      do k= 1,nat
         do i= 1,nax
            cap1= aux_cp(neb,nar,nint,xarxa(i,1),xarxa(i,2),
     $           tiparx(i),tiparc,MAXNAR,2,conca, 
     $           MAXNAR,MAXINT,capmut, 
     $           MAXNEB,MAXINT,capemb )
            if (tiparx(i).ne.'m') then
               o = aux_em1( neb ,nint ,xarxa(i,1)) 
               d = aux_em1( neb ,nint ,xarxa(i,2)) 
               tarc = tiparx(i)
               j= aux_arc_mps(nar,tiparc,MAXNAR,2,conca,o,
     $              d,tarc)
               l= aux_in1(neb ,nint ,xarxa(i,1))
               cap2 = limarc(k,j,l)
            else
               j= aux_em1( neb ,nint,xarxa(i,1)) 
               l= aux_in1( neb ,nint,xarxa(i,1))
               cap2 = limemb(k,j,l)
            end if
            ct= ct+1
            bu(ct)= min(cap1,cap2) 
            bl(ct)= 0.0d0
         end do
      end do
c     2.2 variables HDkiii
      do i= 1,nint
c     commodity 1 (deterministic hydro generation)
         ct= ct+1
         bl(ct)= ZERO
         bu(ct)= max_ho(i)
c     rest of commodities
         do k= 2,nat
            ct= ct+1
            bl(ct)= ZERO
            bu(ct)= phdmax*nhores(i)
         end do
      end do
c     2.3 variables HItttiii
      do i= 1,nint
         do j= 1,nterm
            ct= ct+1
            bl(ct)= ZERO
            bu(ct)= phdmax*nhores(i)
         end do
      end do
c     2.4 variables Cctttiii
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               if (es_contracte(j).and.(ini_contract(j).ne.i)) then
c     energy provided by contract at intervals different from its
c     beginning interval, is zero
                  bl(ct)= ZERO
                  bu(ct)= ZERO
               else
                  bl(ct)= ZERO
                  bu(ct)= potcomprar(j,k)
               end if
            end do
         end do
      end do
c     2.5 variables Nc
      do k= 1,ncomb
         ct= ct+1
         bl(ct)= ZERO
         bu(ct)= totalcomb(k)
      end do
c     2.6 variables Rcttt
      do j= 1,nterm
         do k= 1,nartterm(j)
            ct= ct+1
            if (es_contracte(j)) then
c     no residual energy is allowed (either or not this contract is 
c     available at last interval)
               bl(ct)= ZERO
               bu(ct)= ZERO
            else
               bl(ct)= ZERO
               bu(ct)= potguardar(j,k)
            end if
         end do
      end do
c     2.7 variables Zctttiii
      do i= 1,nint-1
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               if (es_contracte(j) .and. ((ini_contract(j).gt.i).or.
     $              (fi_contract(j).le.i))) then
c     no energy between intervals out of the period of availability
c     of the contract
                  bl(ct)= ZERO
                  bu(ct)= ZERO
               else
                  bl(ct)= ZERO
                  bu(ct)= potguardar(j,k)
               end if
            end do
         end do
      end do
c     2.8 variables Gctttiii
       do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               if (es_contracte(j) .and. ((ini_contract(j).gt.i).or.
     $              (fi_contract(j).lt.i))) then
c     no generation at intervals out of the period of availability
c     of the contract
                  bl(ct)= ZERO
                  bu(ct)= ZERO
               else
                  bl(ct)= ZERO
                  bu(ct)= pmaxterm(j)*nhores(i)
               end if
            end do
         end do
      end do
c     2.9 variables PGtttiii
      inixpgt= ct
      do i= 1,nint
         do j= 1,nterm
            ct= ct+1
c     if there are no fuel limits, then we impose power generated must
c     be equal to maximum power for thermal units (following that 
c     suggested by the convolution techniques we use)
            if((.not.es_contracte(j)).and.(.not.te_limits_comb(j))) then
               bl(ct)= pmaxterm(j)
c**               bl(ct)= pminterm(j)
c**               bl(ct)= 0.0d0
            else
               bl(ct)= ZERO
            end if
            bu(ct)= pmaxterm(j)
         end do
      end do
c     2.10 variables PHDiii
      do i= 1,nint
         ct= ct+1
         bl(ct)= ZERO
         bu(ct)= phdmax
      end do
      
c     bounds for slacks
c     minos5.5 represents equations l< g(x) <u as
c     g(x)+s= 0, l< -s <u, and therefore: bl=-u< s <bu=-l.
      
c     1.1 constraints HDkiii
      do k= 1,nat
         do i= 1,nint
            ct= ct+1
            bl(ct)= ZERO
            bu(ct)= ZERO
         end do
      end do
c     1.2 constraints PGtttiii
      do i= 1,nint
         do j= 1,nterm
            ct= ct+1
            bl(ct)= ZERO
            bu(ct)= ZERO
         end do
      end do
c     1.3 constraints COiii
      do i= 1,nint
         ct= ct+1
         bl(ct)= ZERO
         bu(ct)= ZERO
      end do
c     1.4 constraints HINTiii
      do i= 1,nint
         ct= ct+1
         bl(ct)= ZERO
         bu(ct)= ZERO
      end do
c     1.5 balance and mutual capacity equations of hydro system
      do k= 1,nat
         do i= 1,nint
c     balances at reservoirs 
            do j= 1,neb
               ct= ct+1
               bl(ct)= -aportacio(k,j,i)
               bu(ct)= -aportacio(k,j,i)
            end do
         end do
c     balance at sink
         racum= 0.0d0
         do j= 1,neb
            do i= 1,nint
               racum= racum+aportacio(k,j,i)
            end do
         end do
         ct= ct+1
         bl(ct)= racum
         bu(ct)= racum
      end do
c     mutual capacities of hydro arcs
      do i= 1,nax
         racum= aux_cp(neb,nar,nint,xarxa(i,1),xarxa(i,2),
     $        tiparx(i), tiparc,MAXNAR,2,conca,
     $        MAXNAR,MAXINT,capmut, 
     $        MAXNEB,MAXINT,capemb )
         ct= ct+1
         bl(ct)= -(racum+0.01) 
         bu(ct)= ZERO
      end do
c     1.6 constraints XC
      do i= 1,ncomb
         ct= ct+1
         bl(ct)= -totalcomb(i)
         bu(ct)= -totalcomb(i)
      end do
c     1.7 constraints Xctttiii
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               bl(ct)= ZERO
               bu(ct)= ZERO
            end do
         end do
      end do
c     1.8 constraints PGMiii
      do i= 1,nint
         ct= ct+1
         acumpterm= 0.0d0
         do j= 1,nterm
            if (es_contracte(j)) then
               if ((i.ge.ini_contract(j)).and.(i.le.fi_contract(j))) 
     $              then
                  acumpterm= acumpterm+pmaxterm(j)
               end if
            else
               acumpterm= acumpterm+pmaxterm(j) 
            end if
         end do
         bl(ct)= -(acumpterm+ppunta(i))
         bu(ct)= -(acumpterm+phdmax-min_gt(i))
      end do
c     1.9 constraints PHDiii
      do i= 1,nint
         ct= ct+1
         bl(ct)= -phdmax
         bu(ct)= ZERO
      end do

      end


c***********************************************************************
c     generation of initial point
c
c     INPUT parameters:
c     n,nb
c     
c     OUTPUT parameters:
c     hs(nb),xn(nb)
c***********************************************************************
      subroutine generate_init_point(n,nb,m,nncon,hs,xn,pi,bl,bu)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 n,nb,m,nncon,hs(nb)
      real*8 xn(nb),pi(m),bl(nb),bu(nb)
c     parameters
      integer*4 BS0,BS1,BS3,SBS,LB,UB
      parameter (BS0= 0, BS1= 1, BS3= 3, SBS= 2, LB= 4, UB= 5)      
c     local variables
      integer*4 i,j,k,ct

c     simple initialization of the variables
      do i= 1,n
         hs(i)= 0
         xn(i)= 0.0d0
      end do

c     initialize Lagrange multipliers vector for the nonlinear constr.
      do i= 1,nncon
         pi(i)= 0.0d0
      end do
      return

c     a more ellaborate initialization of the variables
c     remember that: 
c     hs(i)= 0,1,3: variable x(i) is eligible for the basis
c                   (with preference to variables with hs(i)=3)
c     hs(i)= 2: variable x(i) is made superbasic
c     hs(i)= 4: variable x(i) is made nonbasic at lower bound
c     hs(i)= 5: variable x(i) is made nonbasic at upper bound
c     the parameters BS0,BS1,BS3,SBS,LB and UB store these values
      ct= 0
c     2.1 hydro network variables
      do k= 1,nat
         do i= 1,nax
            ct= ct+1
            if (tiparx(i).eq.'m') then
c     volume arc
               xn(ct)= bu(ct)
               hs(ct)= BS1
            else 
c     other arcs
               hs(ct)= BS1
            end if
         end do
      end do
c     2.2 variables HDkiii
      do i= 1,nint
c     commodity 1 (deterministic hydro generation)
         ct= ct+1
         hs(ct)= BS1
c     rest of commodities
         do k= 2,nat
            ct= ct+1
            xn(ct)= bu(ct)
            hs(ct)= BS1
         end do
      end do
c     2.3 variables HItttiii
      do i= 1,nint
         do j= 1,nterm
            ct= ct+1
            xn(ct)= bu(ct)
            hs(ct)= BS1
         end do
      end do
c     2.4 variables Cctttiii
      do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               if (es_contracte(j).and.(ini_contract(j).ne.i)) then
c     energy provided by contract at intervals different from its
c     beginning interval, is zero
                  xn(ct)= bu(ct)
                  hs(ct)= UB
               else
                  hs(ct)= BS1
               end if
            end do
         end do
      end do
c     2.5 variables Nc
      do k= 1,ncomb
         ct= ct+1
         hs(ct)= BS1
      end do
c     2.6 variables Rcttt
      do j= 1,nterm
         do k= 1,nartterm(j)
            ct= ct+1
            if (es_contracte(j)) then
c     no residual energy is allowed (either or not this contract is 
c     available at last interval)
               xn(ct)= bu(ct)
               hs(ct)= UB
            else
               hs(ct)= BS1
            end if
         end do
      end do
c     2.7 variables Zctttiii
      do i= 1,nint-1
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               if (es_contracte(j) .and. ((ini_contract(j).gt.i).or.
     $              (fi_contract(j).le.i))) then
c     no energy between intervals out of the period of availability
c     of the contract
                  xn(ct)= bu(ct)
                  hs(ct)= UB
               else
                  hs(ct)= BS1
               end if
            end do
         end do
      end do
c     2.8 variables Gctttiii
       do i= 1,nint
         do j= 1,nterm
            do k= 1,nartterm(j)
               ct= ct+1
               if (es_contracte(j) .and. ((ini_contract(j).gt.i).or.
     $              (fi_contract(j).lt.i))) then
c     no generation at intervals out of the period of availability
c     of the contract
                  xn(ct)= bu(ct)
                  hs(ct)= UB
               else
                  xn(ct)= bu(ct)
                  hs(ct)= UB
               end if
            end do
         end do
      end do
c     2.9 variables PGtttiii
      do i= 1,nint
         do j= 1,nterm
            ct= ct+1
            xn(ct)= bu(ct)
            hs(ct)= UB
         end do
      end do
c     2.10 variables PHDiii
      do i= 1,nint
         ct= ct+1
         xn(ct)= bu(ct)
         hs(ct)= BS1
      end do
      
      end


c***********************************************************************
c     auxiliary routine that computes the mutual capacity of the
c     arc (origen, desti) of the multiinterval hydro network
c     of type 'tarc'
c***********************************************************************
      real*8 function aux_cp ( neb ,nar ,nint ,origen ,desti ,tarc
     $     ,tiparc ,cdimf ,cdimc ,conca 
     $     ,pdimf ,pdimc ,capmut 
     $     ,bdimf ,bdimc ,capemb )
      implicit none
c     arguments
      integer*4 neb ,nar ,nint ,origen ,desti ,cdimf ,cdimc
     $     ,conca(cdimf,cdimc) 
     $     ,bdimf ,bdimc
     $     ,pdimf ,pdimc
      real*8		capemb(bdimf,bdimc) ,capmut(pdimf,pdimc) 
      character*1	tarc ,tiparc(*)
c     functions
      integer*4 aux_in1 ,aux_em1 ,aux_arc_mps
c     local variables
      integer*4	o ,d ,i ,a 
      
      i = aux_in1 (neb ,nint ,origen)
      o = aux_em1 (neb ,nint ,origen)
      d = aux_em1 (neb ,nint ,desti)
      if (tarc.eq.'m') then
         aux_cp = capemb(o,i) 
      else
         a = aux_arc_mps(nar ,tiparc ,cdimf ,cdimc ,conca ,o ,d ,tarc)
         aux_cp = capmut(a,i) 
      end if

      end	


c***********************************************************************
c     auxiliary routine that check if arc emd-->emo of type 'tarc' 
c     does exist in the hydro system and returns its number
c***********************************************************************
      integer*4 function aux_arc_mps( nar ,tiparc ,cdimf ,cdimc ,conca 
     $     ,emd ,emo ,tarc )
      implicit none
c     arguments
      integer*4	nar ,emd ,emo 
     $     ,cdimf ,cdimc ,conca(cdimf,cdimc)
      character*1	tiparc(*) ,tarc
c     local variables
      integer*4 	i
      
      aux_arc_mps = 0
      i = 1
      do while (i.le.nar)
         if ( (tiparc(i).eq.tarc) .and. (conca(i,1).eq.emd) ) then 
            aux_arc_mps = i
            i = nar
         end if
         i = i + 1
      end do

      end


c***********************************************************************
c     Routine to optimize problem
c     
c     INPUT parameters:
c     m: number of constraints
c     n: number of variables
c     nb: number of bounds (m+n)
c     nncon: number of nonlinear constraints
c     nnobj: number of nonlinear objective variables
c     nnjac: number of nonlinear jacobian variables
c     ne: number of nonzero entries in jacobian (for linear and
c         nonlinear constraints)
c     a(ne): constraint matrix stored columnwise
c     ha(ne): row indices for each nonzero in a() 
c     ka(n+1): pointers to the beginning of each column in a() and ha()
c     bl(nb): lower bounds for variables and slacks
c     bu(nb): upper bounds for variables and slacks
c     hs(nb): estimate and workspace for initial state for variables 
c     pi(m): initial estimate and workspace for multipliers 
c     xn(nb): initial estimates of variables and slacks
c     name1(nb): first 4 char positions of the names of vars and const
c     name2(nb): last 4 fchar positions of the names of vars and const
c     nwcore: number of eight-byte words for workspace
c     z(nwcore): workspace
c     
c     OUTPUT parameters:
c     hs(nb): state of variables at the optimal solution
c     xn(nb): optimal values
c     pi(m): lagrange multipliers of the constraints at optimal solution
c     rc(nb): reduced costs at optimal solution
c     obj: optimal objective function
c     mincor: estimation of workspace required to solve the problem
c     inform: optimization status returned by minoss (see Minos manual
c             page 132)
c     iexit: if iexit= 0 routine exited normally
c***********************************************************************
      subroutine optimize_problem(m,n,nb,nncon,nnobj,nnjac,ne,a,ha,ka,
     $     bl,bu,hs,pi,xn,rc,name1,name2,obj,mincor,z,nwcore,inform,
     $     iexit)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 m,n,nb,nncon,nnobj,nnjac,ne,mincor,nwcore,inform,iexit
      integer*4 ha(ne),ka(n+1),hs(nb)
      real*8 a(ne),bl(nb),bu(nb),xn(nb),pi(m),rc(nb),obj,z(nwcore)
      integer*4 name1(nb),name2(nb)
c     local variables
      character*8 names(5), starttype
      real*8 sinf,objadd
      integer*4 ispecs,iprint,isumm,iobj,ns,ninf,nname,majit,minit
      integer*4 i,j,ct,nopt,incr_nopt

      iexit= 0
      
c     Specify some of the MINOS files.
c     ispecs  is the Specs file   (0 if none).
c     iprint  is the Print file   (0 if none).
c     isumm   is the Summary file (0 if none).
 
      ispecs = 50
      if (ispecs.gt.0) rewind(ispecs)
      iprint = 6
      isumm  = 0
      
c     Set options to default values.
      
      call mispec( ispecs, iprint, isumm, nwcore, inform )
      
c     Specify options (like Specs file)
      majit= 100000
      minit= 15

      call miopt ('Minimize', 0, 0, inform)
      call miopt ('Jacobian Sparse', iprint, isumm, inform)
      call miopti ('Derivative Level', 3, iprint, isumm, inform)
c      call miopti ('Verify Level', 3, iprint, isumm, inform)
c     problem number is 1003 (1001 and 1002 are already occupied 
c     by routine ajuste_conlim)
      call miopti ('Problem number', 1003, iprint, isumm, inform)
      call miopti('Superbasics Limit',int(n*0.2), iprint, isumm, inform)
      call miopti('Iterations', majit*minit, iprint, isumm, inform)
      call miopti('Major Iterations', majit, iprint, isumm, inform)
      call miopti('Minor Iterations', minit, iprint, isumm, inform)
      call miopti('Penalty Parameter',50,iprint, isumm, inform)
c      call mioptr('Linesearch Tolerance', 2.d-1, iprint, isumm, inform)
c      call mioptr('Subspace Tolerance', 1.d-1, iprint, isumm, inform)
c      call mioptr('Feasibility Tolerance',1.d-6,iprint, isumm, inform)
c      call mioptr('Optimality Tolerance',1.d-5,iprint, isumm, inform)
c     solve through minoss
      iobj= 0
      objadd= 0.0d0
      names(1) = 'htcoor  '
      names(2) = 'linobj  '
      names(3) = 'rhs     '
      names(4) = 'ranges  '
      names(5) = 'bounds  '
      nname= nb

      call minoss( 'Cold', m, n, nb, ne, nname,
     $     nncon, nnobj, nnjac,
     $     iobj, objadd, names,
     $     a, ha, ka, bl, bu, name1, name2,
     $     hs, xn, pi, rc, 
     $     inform, mincor, ns, ninf, sinf, obj,
     $     z, nwcore )
      if ((inform.eq.0).or.(inform.eq.9)) return
      return

c     reoptimize adjusting lower bounds for power generation
c     variables PGtttiii
      inform= 0
      nopt= 0
      incr_nopt= 2
      do while ((nopt.le.10).and.((inform.eq.0).or.(inform.eq.9)))
         ct= inixpgt
         do i= 1,nint
            do j= 1,nterm
               ct= ct+1
               if(.not.(es_contracte(j).or.te_limits_comb(j))) then
                  bl(ct)= pmaxterm(j)/10.0*nopt
               end if
            end do
         end do         
         if (nopt.eq.0) then
            starttype= 'Cold'
         else
            starttype= 'Hot'
         endif
         call minoss( starttype, m, n, nb, ne, nname,
     $        nncon, nnobj, nnjac,
     $        iobj, objadd, names,
     $        a, ha, ka, bl, bu, name1, name2,
     $        hs, xn, pi, rc, 
     $        inform, mincor, ns, ninf, sinf, obj,
     $        z, nwcore )
         nopt= nopt+incr_nopt
      end do
      end

c***********************************************************************
c     creation of pointers to the part of jacobian related with 
c     the hydro system
c
c     INPUT parameters:
c     inijachd(ne_njac): auxiliary vector
c     ne_njac: dimension of inijachd() and ijachd()
c     
c     OUTPUT parameters:
c     ijachd(ne_njac): pointers to jacobian entries related to hydro
c            generation. 
c     iexit: if iexit= 0 routine exited normally
c            if iexit= -15 dimensions of ijachd and inijachd
c                    insufficients for this model. This should
c                    never happen, since ne_njac overestimates
c                    this value.
c***********************************************************************
      subroutine crea_indexs_hd(inijachd,ijachd,ne_njac,iexit)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 iexit,ne_njac,inijachd(ne_njac),ijachd(ne_njac)
c     variables locals
      integer*4 i,j,k,l,koc,ij,iaux2,jac,ivi,ivf,iwf,iwi,
     $     despblochd(MAXNAT)

      p_ijachd4= %loc(ijachd)
      four_bytes_addresses= .true.
      if (.not.(p_ijachd4 .eq. %loc(ijachd))) then
         p_ijachd8= %loc(ijachd)
         four_bytes_addresses= .false.
      end if
      maxjac= ne_njac
c
c     INDEXS DE JACOBIA PER PART HIDRAULICA
c
c     inicialitzem part de jacobia pel que fa a variables hidrauliques
c     i restriccions de canvi de variable. guardarem la posicio de 
c     l'entrada al jacobia de cada variable hidraulica.
c     nomes posem els indexs per un bloc (dels nat*nat blocs que ens 
c     apareixen). Els indexs de la resta de blocs s'obtenen per simetria
c     sumant desplacaments. El valor d'aquest desplacaments es troba 
c     a la variable elemblochd (nombre d'elements a cada bloc on hi ha
c     elements del jacobia).
c
      iexit= 0
c     netegem inijachd
      if (nax.gt.maxjac) then
         iexit= -15
         return
      endif
      do i= 1,nax
         inijachd(i)= 0
      end do
c     posem quants termes afecten a un bloc en concret.
      koc= 0
      do i= 1,nint
         do j= 1,ncc
            koc= koc+1
            inijachd(jca(koc))= 1
            ivi=ivo(koc)
            ivf=lvo(koc)
            iwi=ivd(koc)
            iwf=lvd(koc)
            if (ivi.gt.0) inijachd(ivi)= 2
            if (ivf.gt.0) inijachd(ivf)= 2
            if (iwi.gt.0) inijachd(iwi)= 2
            if (iwf.gt.0) inijachd(iwf)= 2
         end do    
      end do
c     trobem el punter a primer per cada variable sumant acumuladament.
      elemblochd= 1
      do i= 1,nax
         k= inijachd(i)
         if (inijachd(i).gt.0) inijachd(i)= elemblochd
         elemblochd= elemblochd + k
      end do
      elemblochd= elemblochd-1
c     trobem despblochd(k): desplacament per cada article als blocs
      j= 0
      do k= 1,nat
         despblochd(k)= j
         j= j+nat+1-k
      end do
      do k= 1,nat
         despblochd(k)= elemblochd*despblochd(k)
      end do
c     un cop tenim aixo creem un vector amb els indexs d'entrada al 
c     jacobia. per fer aquest vector usem els dos anteriors despblochd
c     i inijachd.
      ij= 0
      koc=0
      do i= 1,nint
         do j= 1,ncc
            koc=koc+1
            jac=jca(koc)
            ivi=ivo(koc)
            ivf=lvo(koc)
            iwi=ivd(koc)
            iwf=lvd(koc)
            do k= 1,nat
c     derivada respecte la Descarrega.
               do l= 1,k
                  iaux2= despblochd(l)+(inijachd(jac)-1)*(nat+1-l)+1
     $                 +k-l
                  ij= ij+1
                  if (ij.gt.maxjac) then
                     iexit= -15
                     return
                  end if
                  ijachd(ij)= iaux2
               end do
c     derivada respecte Volum-Inicial-Origen
               if (ivi.gt.0) then
                  do l= 1,k
                     iaux2= despblochd(l)+(inijachd(ivi)-1)*(nat+1-l)+1
     $                    +2*(k-l)+1
                     ij= ij+1
                     if (ij.gt.maxjac) then
                        iexit= -15
                        return
                     end if
                     ijachd(ij)= iaux2
                  end do
               end if
c     derivada respecte Volum-Inicial-Desti
               if (iwi.gt.0) then
                  do l= 1,k
                     iaux2= despblochd(l)+(inijachd(iwi)-1)*(nat+1-l)+1
     $                    +2*(k-l)+1
                     ij= ij+1
                     if (ij.gt.maxjac) then
                        iexit= -15
                        return
                     end if
                     ijachd(ij)= iaux2
                  end do
               end if
c     derivada respecte Volum-Final-Origen
               if (ivf.gt.0) then
                  do l= 1,k
                     iaux2= despblochd(l)+(inijachd(ivf)-1)*(nat+1-l)+1
     $                    +2*(k-l)
                     ij= ij+1
                     if (ij.gt.maxjac) then
                        iexit= -15
                        return
                     end if
                     ijachd(ij)= iaux2
                  end do
               end if
c     derivada respecte Volum-Final-Desti
               if (iwf.gt.0) then
                  do l= 1,k
                     iaux2= despblochd(l)+(inijachd(iwf)-1)*(nat+1-l)+1
     $                    +2*(k-l)
                     ij= ij+1
                     if (ij.gt.maxjac) then
                        iexit= -15
                        return
                        end if
                     ijachd(ij)= iaux2
                  end do
               end if
            end do
         end do
      end do
      end 


c***********************************************************************
c     creation of pointers to the part of jacobian related with 
c     the thermal system
c
c     INPUT parameters:
c     ne_njac: dimension of inijaterm()
c     
c     OUTPUT parameters:
c     ijacterm: pointers to jacobian entries related to thermal
c            generation. 
c     iexit: if iexit= 0 routine exited normally
c            if iexit= -16 dimension of ijacterm insufficients 
c                      for this model.This should
c                      never happen, since ne_njac overestimates
c                      this value.
c***********************************************************************
      subroutine crea_indexs_term(ijacterm,ne_njac,iexit)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 iexit,ne_njac,ijacterm(ne_njac)
c     local variables
      integer*4 i,j,k,l,iax,ij,ijt1,ijt2,ijt3,ijt4,ijt5,
     $     inijacterm1(MAXINT),inijacterm2(MAXINT),inijacterm3(MAXINT),
     $     inijacterm4(MAXINT),inijacterm5(MAXINT)

      if (four_bytes_addresses) then
         p_ijacterm4= %loc(ijacterm)
      else
c         p_ijacterm8= %loc(ijacterm)
      end if

      iexit= 0
c     compute total power provided by the thermal system at
c     each interval, considering contracts availability 
      do i= 1,nint
         sumpmaxterm(i)= 0.0d0
         do j= 1,nterm
            if (.not.(es_contracte(j) .and. ((ini_contract(j).gt.i).or.
     $           (fi_contract(j).lt.i)) )) then
               sumpmaxterm(i)= sumpmaxterm(i)+pmaxterm(j)
            end if
         end do
      end do
c     
      i= 0
      do j= 1,nterm
         do k= 1,nartterm(j)
            i= i+1
            if (.not.es_contracte(j)) then
               tipuscomb(i)=artterm(j,k)
            else
c     si es un contracte marquem el tipus de combustible com 
c     la termica amb signe canviat
               tipuscomb(i)= -j
            end if
         end do
      end do
c     creem vectors d'index per hn:index de gen. hidraulica obtingudes  
c     a restriccions no lineals a principi de cada interval.
      k= nax*nat
      do i= 1,nint
         hn(i)= k+1+(i-1)*nat
      end do
c     creem vectors d'index per hl:index de gen. hidraulica obtingudes  
c     a restriccions lineals a principi de cada interval.
      k= k+nint*nat
      do i= 1,nint
         hl(i)= k+1+(i-1)*nterm
      end do
c     creem vectors d'index per c:index de compra combustible
      k= k+nint*nterm
      do i= 1,nint
         c(i)= k+1+(i-1)*nthrmarticles
      end do
c     creem vectors d'index per r:index de remanents de combustible.
      k= k+nint*nthrmarticles+ncomb
      do i= 1,nthrmarticles 
         r(i)= i+k
      end do
c     creem vectors d'index per t:index de generacio termica a principi
c     de cada interval 
      k= k+nint*nthrmarticles
      do i= 1,nint
         t(i)= k+1+(i-1)*nthrmarticles
      end do
c     creems vectors d'index per pg: index de potencia generada per
c     cada interval
      k= k+nint*nthrmarticles
      do i= 1,nint
         pg(i)= k+1+(i-1)*nterm
      end do
c     creems vectors d'index per phd: index de potencia hidraulica 
c     d'aigues no sempre probables per cada interval.
      k= k+nint*nterm
      do i= 1,nint
         iphd(i)= k+i
      end do
c     
      k= k+nint
c
c     INDEXS DE JACOBIA PER PART TERMICA
c     indexs a inici de jacobia termica segons 5 tipus de variables:
c     1: hd que interve en la forma de corba de Bezier.
c     2: hd que interve en la generacio intercalada d'energia.
c     3: generacio termica.
c     4: potencia generada.
c     5: potencia hidraulica generada.

      inijacterm1(1)= elemblochd*nat*(nat+1)/2+1
      iax= (nterm+1)+nat
      do i= 2,nint
         inijacterm1(i)= inijacterm1(i-1)+iax
      end do
      
      inijacterm2(1)= inijacterm1(1)+nint*iax
      iax= nterm*(nterm+1)/2 
      do i= 2,nint
         inijacterm2(i)= inijacterm2(i-1)+iax
      end do

      inijacterm3(1)= inijacterm2(1)+nint*iax
      iax= 0
      do j= 1,nterm
         iax= iax+(nterm+2-j)*(nartterm(j))
      end do
      do i= 2,nint
         inijacterm3(i)= inijacterm3(i-1)+iax
      end do
      
      inijacterm4(1)= inijacterm3(1)+nint*iax
      iax= (nterm+1)*nterm
      do i= 2,nint
         inijacterm4(i)= inijacterm4(i-1)+iax
      end do

      inijacterm5(1)= inijacterm4(1)+nint*iax
      iax= nterm+1
      do i= 2,nint
         inijacterm5(i)= inijacterm5(i-1)+iax
      end do

      ij= 0
      do i= 1,nint
c     restriccions de potencia generada per cada termica
         do j= 1,nterm
            ijt1= inijacterm1(i)-1
            ijt1= ijt1+j+1
            ij= ij+1
            if (ij.gt.maxjac) then
               iexit= -16
               return
            end if
            ijacterm(ij)= ijt1
            ijt2= inijacterm2(i)-1
            do l= 1,j-1
               ijt2= ijt2+j-l
               ij= ij+1
               if (ij.gt.maxjac) then
                  iexit= -16
                  return
               end if
               ijacterm(ij)= ijt2
               ijt2= ijt2+nterm+1-j
            end do
            ijt3= inijacterm3(i)-1
            do l= 1,j
               do k= 1,nartterm(l)
                  ijt3= ijt3+j+1-l
                  ij= ij+1
                  if (ij.gt.maxjac) then
                     iexit= -16
                     return
                  end if
                  ijacterm(ij)= ijt3
                  ijt3= ijt3+nterm+1-j
               end do
            end do
            ijt4= inijacterm4(i)-1
            do l= 1,nterm
               ijt4= ijt4+j
               ij= ij+1
               if (ij.gt.maxjac) then
                  iexit= -16
                  return
               end if
               ijacterm(ij)= ijt4
               ijt4= ijt4+nterm+1-j
            end do
            ijt5= inijacterm5(i)-1
            ijt5= ijt5+j
            ij= ij+1
            if (ij.gt.maxjac) then
               iexit= -16
               return
            end if
            ijacterm(ij)= ijt5
        end do
c     restriccio de cobriment (interval i). posem darrera hidraulica
c     g(.) -valor jacobia-
         ijt1= inijacterm1(i)-1
         ijt1= ijt1+nterm+2
         ij= ij+1
         if (ij.gt.maxjac) then
            iexit= -16
            return
         end if
         ijacterm(ij)= ijt1
         ijt2= inijacterm2(i)-1 
         do j= 1,nterm
            ijt2= ijt2+nterm+1-j
            ij= ij+1
            if (ij.gt.maxjac) then
               iexit= -16
               return
            end if
            ijacterm(ij)= ijt2
         end do
         ijt3= inijacterm3(i)-1
         do j= 1,nterm
            do k= 1,nartterm(j)
               ijt3= ijt3+nterm+2-j
               ij= ij+1
               if (ij.gt.maxjac) then
                  iexit= -16
                  return
               end if
               ijacterm(ij)= ijt3
            end do
         end do
         ijt4= inijacterm4(i)-1
         do j= 1,nterm
            ijt4= ijt4+nterm+1
            ij= ij+1
            if (ij.gt.maxjac) then
               iexit= -16
               return
            end if
            ijacterm(ij)= ijt4
         end do
         ijt5= inijacterm5(i)-1
         ijt5= ijt5+nterm+1
         ij= ij+1
         if (ij.gt.maxjac) then
            iexit= -16
            return
         end if
         ijacterm(ij)= ijt5
      end do
      end

c***********************************************************************
c     Routine to check correctness of a(),ha(),ka()
c     Generate MPS columns section of the model
c     
c     INPUT parameters:
c     m: number of constraints
c     n: number of variables
c     nb: number of bounds (nb= m+n)
c     ne:  number of nonzero entries in jacobian (linear and nonlinear)
c     a(ne): constraint matrix stored columnwise
c     ha(ne): row indices for each nonzero in a() 
c     ka(n+1): pointers to the beginning of each column in a() and ha()
c     name1(nb): first 4 char positions of the names of vars and const
c     name2(nb): last 4 fchar positions of the names of vars and const
c     bl(nb): lower bounds for variables and slacks
c     bu(nb): upper bounds for variables and slacks
c
c     OUTPUT parameters:
c     columns, bounds and rhs MPS sections through standard output 
c***********************************************************************
      subroutine check_mps(m,n,nb,ne,a,ha,ka,name1,name2,bl,bu)
      implicit none
c     arguments
      integer*4 m,n,nb,ne,ha(ne),ka(n+1)
      real*8 a(ne),bl(nb),bu(nb)
      character*4 name1(nb),name2(nb)
c     local variables
      integer*4 i,j

      write(*,*) 'columns section'
      do i= 1,n
         do j= ka(i),ka(i+1)-1 
            write(*,1003) name1(i),name2(i),name1(n+ha(j)),
     $           name2(n+ha(j)),a(j)
         end do
      end do
      
c     check information. this will be removed, it is just for 
c     checking the model
      write(*,*) 'bounds and rhs sections'
      write(*,*) 'variables'
      write(*,*)  '   i        bl(i)     name(i)         bu(i)'  
      do i= 1,n
         write(*,1000) i,bl(i),name1(i),name2(i),bu(i)
      end do
      write(*,*) 'constraints'      
      write(*,*)  '   i        bl(i)     name(i)         bu(i)'  
      do i= n+1,n+m
         write(*,1000) i,-bu(i),name1(i),name2(i),-bl(i)
      end do
      
 1000 format(i5,3x,f12.5,3x,2a4,3x,f12.5)
 1003 format('    ',2a4,'  ',2a4,'  ',f12.7)      
      end

c***********************************************************************
c     routine for writing some information of the solution obtained
c     with minos
c     
c     INPUT parameters:
c     n: number of variables
c     m: number of constraints
c     nncon: number of nonlinear constraints
c     x(n): solution provided by Minos
c     f: objective function provided by Minos
c***********************************************************************
      subroutine write_info_optim(n,m,nncon,ne,x,f)
      implicit none
      include 'htcoor_f.h'
c     arguments
      integer*4 n,m,nncon,ne
      real*8 x(n),f
c     local variables
      integer*4 i,j,k,l,igj,nterm_amb_limit,iprint
      real*8 vaux(MAXINT),raux,sumterm,hd1,raux0,raux1,raux2,raux3,
     $     phdcobri,cost_term,acum_cost,xaux, acum_thermal_energy

      iprint= 6

      write(iprint,*) '**********************************************'
      write(iprint,*) '         SUMMARY OF SOLUTION OBTAINED         '
      write(iprint,*) '**********************************************'
      write(iprint,*) 'OBJECTIVE FUNCTION COST:',
     $     f*ordre_preu*escal_energia,' currency units'
      write(iprint,*)

      write(iprint,*) 'PROBLEM DIMENSIONS'
      write(iprint,*) '       nint       nterm ncontractes     ',
     $     'nembass       ncomm'
      write(iprint,'(5i12)') nint,nterm-ncontr,ncontr,neb,nat
      write(iprint,*)
      nterm_amb_limit= 0
      do j= 1,nterm
         if (te_limits_comb(j)) nterm_amb_limit= nterm_amb_limit+1
      end do
      write(iprint,*) '#thermal units with fuel limits=', 
     $     nterm_amb_limit
      write(iprint,*)
      write(iprint,*) '     #variables   #linear const. ',
     $     '#nonlinear const.     #nnz constraints'
      write(iprint,'(3i16)') n,m-nncon,nncon,ne
      write(iprint,*)
      write(iprint,*) 'GENERATION THERMAL UNITS/CONTRACTS in MWh'
      igj= 0
      acum_cost= 0.0
      acum_thermal_energy= 0.0
      do j= 1,nterm
         write(iprint,*)
         if (es_contracte(j)) then
            write(iprint,*) ' ---  Generation CONTRACT ',j,
     $           '  (',nomterm(j),')',
     $           ' for all the intervals'
         else
            write(iprint,*) ' ---  Generation THERMAL UNIT ',j, 
     $           '  (',nomterm(j),')',
     $           ' for all the intervals'
         end if
         raux= 0.0
         cost_term= 0.0
         do i= 1,nint
            vaux(i)= 0.0
            do k= 0,nartterm(j)-1
               vaux(i)= vaux(i)+x(t(i)+igj+k)
               if (.not.es_contracte(j)) then !es termica
                  l= artterm(j,k+1)
                  cost_term=cost_term+x(c(i)+igj+k)*preu(l)
                  if (i.eq.nint) then
                     cost_term= cost_term-
     $                    apreciacio(l)*x(r(igj+k+1))*preu(l)
                  end if
               else
                  xaux= x(c(i)+igj+k)
                  cost_term=cost_term+xaux*(cost_lin(j)+
     $                 cost_quad(j)*xaux)
               end if
            end do
            raux= raux+vaux(i)
         end do
         igj= igj+nartterm(j)
         acum_cost= acum_cost+cost_term

         write(iprint,1000) (vaux(i)*escal_energia,i=1,nint)
         write(iprint,*) 'Total generation thermal unit/contract:',
     $        raux*escal_energia
         acum_thermal_energy= acum_thermal_energy+raux*escal_energia
         write(iprint,*) 'Total cost thermal unit/contract:',
     $        cost_term*ordre_preu*escal_energia,' currency units'
      end do
      write(iprint,*)
      write(iprint,*) '**Total energy of thermal+contract system:',
     $     acum_thermal_energy
      write(iprint,*) '**Total cost of thermal+contract system:',
     $     acum_cost*ordre_preu*escal_energia,' currency units'
         

      write(iprint,*)
      write(iprint,*) 'HYDRAULIC GENERATION in MWh'
      write(iprint,*) 'Deterministic hydro-generation',
     $     ' (for all the intervals)'
      raux= 0.0
      do i=1,nint
         vaux(i)= x(hn(i))
         raux= raux+vaux(i)
      end do
      write(iprint,1000) (vaux(i)*escal_energia,i=1,nint)
      write(iprint,*) 'Total Deterministic hydro-generation:',
     $     raux*escal_energia
      write(iprint,*)
      write(iprint,*)'Stochastic hydro-generation',
     $     ' (for all the intervals)'
      raux= 0.0
      do i=1,nint
         vaux(i)= 0.0
         do j= 0,nterm-1
            vaux(i)= vaux(i)+x(hl(i)+j)
         end do
         raux= raux+vaux(i)
      end do
      write(iprint,1000) (vaux(i)*escal_energia,i=1,nint)
      write(iprint,*) 'Total stochastic hydro -generation:',
     $     raux*escal_energia
      
      write(iprint,*)
      write(iprint,*) 'EXTERNAL ENERGY in MWh (for all the intervals)'
      raux= 0.0
      do i= 1,nint
         hd1= x(hn(i))
         sumterm= 0.0d0
         do j= 0,nterm-1
            sumterm= sumterm+x(pg(i)+j)
         end do
         phdcobri= x(iphd(i))
         raux0= hd1-e0b(i)
         raux1= sumpmaxterm(i)-sumterm 
         raux2= phdmax-hd1/tbase(i)-phdcobri 
         raux3= raux1+raux2
         vaux(i)= (e0a(i)*raux3*raux3)/raux0 + e0c(i)
         raux= raux+vaux(i)
      end do
      write(iprint,1000) (vaux(i)*escal_energia,i=1,nint)
      write(iprint,*) 'Total External energy=',raux*escal_energia
      write(iprint,*) 
      write(iprint,*) '**Cost of external energy:',
     $     raux*preuext*ordre_preu*escal_energia,' currency units'
      
      write(iprint,*) 
      write(iprint,*) '**Total cost of thermal system+contracts+',
     $     'en.externa:',
     +     (acum_cost+raux*preuext)*ordre_preu*escal_energia,
     +     ' currency units'
      write(iprint,*) '**Difference total cost computed--objective f.',
     +     (acum_cost+raux*preuext-f)*ordre_preu*escal_energia

 1000 format(4f15.2)
      end

c***********************************************************************
c     auxiliary routine for writing constraints structure in matlab
c     sparse matrix format. 
c     
c     INPUT parameters:
c     m: number of variables
c     n: number of constraints
c     ne: number of nonzero elements
c     ha(ne): row indices
c     ka(n+1): pointers to the beginning of each column
c     un: number of a fortran free unit
c***********************************************************************
      subroutine write_matlab_sparse_matrix(m,n,nnz,ha,ka,nu)
      implicit none
c     arguments
      integer*4 m,n,nnz,ha(nnz),ka(n+1),nu
c     local variables
      integer*4 i,j
      
      open(unit=nu,file='jacobian.dat',status='unknown')

      write(*,*) 'm= ',m,'n= ',n,'nnz= ',nnz
      do j= 1,n
         do i= ka(j),ka(j+1)-1
            write(nu,*) ha(i),',',j,',',1
         end do
      end do
      end
      
