c**********.P.*.A.*.N.*.E.*.S.*__*.R.*.E.*.A.*.L.**************************
c
c     rutina per transformar una forma de monotona en una altra donades
c     unes panes de les termiques, limitacions de combustible i 
c     potencies maximes de generacio.
c     Els parametres son:
c     - ppunta: potencia punta monotona
c     - nhores: nombre d'hores de monotona
c     - nqt: nombre de quadrets de temps de les monotones
c     - inm(nqt): monotona de temps d'entrada
c     - outm(nqpfinal): monotona de potencia de sortida
c     - nterm: nombre de termiques 
c     - p(nterm): probabilitat de pana de cada termica
c     - limits(nterm): limits de combustilbe de cada termica
c     - potmax(nterm): potencies maximes de cada termica
c***********************************************************************
      subroutine panes_real(ppunta,nhores,z,r,nterm,p,
     +     limits,potmax,esc)
      implicit none
c     includes
      include 'mono.h'
c     constants auxiliars
      integer*4 MAXQP
      parameter(MAXQP=1000)
c     parametres
      integer*4 nterm
      real*8 ppunta,nhores,p(nterm),limits(nterm),potmax(nterm)
      record /mon_carga/ z
      record /mon_carga_r/ r
      record /incres/ esc
c     variables locals
      integer*4 i,j,k
      integer*4 n
      real*8 treball(MAXQP)
      real*8 enacum,qaux1

c     fem inicialitzacions

      do i=1,z.nqp
         treball(i)= dble(z.filas(i))
      end do
      do i=z.nqp+1,MAXQP
         treball(i)= 0.0d0
      end do
      r.nqp= 0

c     fem convolucions i deconvolucions exactes
      do j=1,nterm
         n= nint(potmax(j)/esc.inc_p)
c     considerem la limitacio de combustible
         enacum= 0.0d0
         i=0
         do while ((i.lt.n).and.(enacum.le.limits(j)))
            i= i+1
            enacum= enacum+(1.0d0-p(j))*treball(i)*esc.inc_e
         end do
         if (enacum.gt.limits(j)) then
c23456789012345678901234567890123456789012345678901234567890123456789012
            enacum= enacum-(1.0d0-p(j))*treball(i)*esc.inc_e*0.5d0
            if (limits(j).gt.enacum) then
               n= i
            else
               n= i-1
            end if
         end if
c     trobem la nova forma
         do k=1,n
            r.filas(k+r.nqp)= (1.0d0-p(j))*treball(k)
         end do
         r.nqp= r.nqp+n
c     convolucionem 
         do k=1,z.nqp
            treball(k)= p(j)*treball(k)+(1.0d0-p(j))*treball(k+n)
         end do
      end do
c     afegim cua de rata finalment
      do k=1,z.nqp
         r.filas(k+r.nqp)= treball(k)
      end do
      r.sigmat= r.nqp
      r.nqp= r.nqp+z.nqp

c     post-proces final per tractar la punta d'energia
c     arreglem part previa a nqptall per si hi ha discontinuitats a la
c     nova forma de la monotona a la part de potencia minima.

      qaux1= 0.0d0
      r.pmin= z.pmin
      do i=1,z.pmin
         qaux1= qaux1+r.filas(i)
      end do
      qaux1= dble(qaux1)/dble(z.pmin)
      do i=1,z.pmin
         r.filas(i)= qaux1
      end do
      end                       !fi subrutina


