C	:::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C	creates shapes of the LDC of each subperiod, from the
C	load estimates. An LDC is an integer array with size "nqt",
C	each value is proportional to the load; the peak load is
C	fixed to a value of reference "nqp".
C	Parameters:
C	Input.  nh(): hours of each subperiod		  integer
C		nper: number of subperiods		  integer
C		nqt : discretization of duration	  integer
C		nqp : discretization of load		  integer
C		load(): array of hourly loads		  real
C		maxqt: first dimension of matrix "monot"  integer
C	Output. monot(,): array of LDC			  integer
C..................................................................
	subroutine getLDCshapes(nh, nper, n_qt, n_qp, monot, maxqt,
     >				 load, iexit)
	implicit none
	include 'HTC_parameter.h'

	integer	nper, n_qt, n_qp, maxqt, monot(maxqt,*), iexit
	real*8	nh(*)
	real	load(*)

	integer	i, j, k, l, m
	real	crg(MAXLOADS), a, s, x, y, z, conv

	l = 1
	do i=1, nper
	    m = nh(i)
	    if (nh(i).gt.MAXLOADS) then
C		write(*,*) '[getLDCshapes] Error: not enough space'
		!! This error can be avoided by increasing MAXLOADS
		iexit=-27
		return
	    end if
	    do j=1, m		  !!! Take loads from i-th subperiod
		crg(j) = load(l)
		l=l+1
	    end do		  	  !!! and sort them
	    call ordena_vector(crg, m)
	    conv = dble(n_qp)/crg(1)
					!!! Discretize according nqp & nqt
	    if (nh(i).ge.n_qt) then
		z = nh(i)/n_qt		!!!     nh(i) > nqt
		a = 0.d0
		j = 1
		k = 1
		do while (j .le. n_qt)
		    x = 0.0
		    if (k.gt.1) x = a*crg(k-1)
		    y = z - a
		    do while (y .ge. 1.0)
			x = x+crg(k)
			k = k+1
			y = y-1.0
		    end do
		    x = (x+y*crg(k))/z
		    monot(j, i) = nint(conv*x)
		    k = k+1
		    a = 1.0-y
		    j = j+1
		end do
					!!!     nh(i) < nqt
	    else
		s = dble(n_qt)/nh(i)
		a = 0.d0
		j = 1
		k = 1
		do while (j .le. n_qt)
		    y = s - a
		    do while (y .ge. 1.0 .and. j .le. n_qt)
			monot(j, i) = nint(conv*crg(k))
			j = j+1
			y = y-1.0
		    end do
		    a = 1.0-y
		    if (j.le.n_qt) 
     >			monot(j,i) = nint(conv*(y*crg(k)+a*crg(k+1)))
		    k = k+1
		    j = j+1
		end do
	    end if
	end do

	end
CCCCCCCC
	subroutine ordena_vector(c, n)	       !! sorts a real vector c()
	implicit none

	integer	n, i, j
	real	c(*), mx, aux

	do i=1, n-1	
	    mx = c(i)
	    do j=i+1, n
		 if (c(j) .gt. mx) then
			aux = c(j)
			c(j) = mx
			mx = aux
		 end if
	    end do
	    c(i) = mx
	end do
	end
C	:::::::::::::::::::::::::::::::::::::::::::::::::::::::::
C	Extracts the maximum load of each subperiod, from the load
C	estimates.
C	
C	Parameters:
C	Input.  nh(): hours of each subperiod		  integer
C		nper: number of subperiods		  integer
C		load(): array of hourly loads		  real
C	Output. pp(): peak loads per subperiod		  real*8
C..................................................................
	subroutine getLDCpeaks(nh, nper, load, pp)
	implicit none

	integer	nper
	real	load(*)
	real*8	nh(*), pp(*), x

	integer	i, j, l, m

	l = 1
	do i=1, nper
	    m = nh(i)
	    x = -1.0
	    do j=1, m		  !!! Take loads from i-th subperiod
		if (x.lt.load(l)) x = load(l)
		l=l+1
	    end do
	    pp(i) = x
	end do
	end 
CCCCCCCC  -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
	subroutine test_of_readLDCshapes
	implicit none

	real cargas(2000), ran, mx, e
	integer	monot(100, 5), nqt, nqp, nh(5), nper, seed, i, j, k, f
	data	nper/ 5/ nqt/ 100/ nqp/ 200/ nh/36,317,262,331,487/

	read(*,*) seed

	do i=1, 2000
		cargas(i) = 0.4+ran(seed)**(0.5)
	end do
	call getLDCshapes(nh, nper, nqt, nqp, monot, 100, cargas)
!		EP!, nh should be real*8

	k = 1
	do i=1, nper
		e = 0.0
		mx = 0.0
		do j=1, nh(i)
			if (cargas(k).gt.mx) mx = cargas(k)
			e = e+cargas(k)
			k=k+1
		end do
		f = 0
		do j=1, nqt
			f = f+monot(j,i)
		end do
		write(*,*) e, f*mx/nqp*real(nh(i))/nqt
	end do
	end
CXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
CXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	subroutine skip_rems(unit, line, final)
	implicit none

	integer		unit
	character*(*)	line
	logical		final, stop

	stop = .false.
	do while (.not. stop)
		read(unit,10, end=100) line
		if (index(line,'%') .eq. 0) stop = .true.
	end do
	return
10	format((a))
100	final = .true.
	end
c	.....................................................
	integer	function search_for(x, set, n)
	implicit none

	character*(*)	x, set(*)
	integer		n, i

	i = 1
	do while (i .le. n .and. index(set(i),x).eq.0)
		i = i+1
	end do
	if (i .gt. n) i = 0
	search_for = i
	return
	end
c	.....................................................
	subroutine identif(x, m)		!! Calcula un identificador
	implicit none				!! de tres letras dado un
						!! nombre. No garantiza que
	character*(*)	x			!! el identificador sea
	character*3	m			!! unico, pero hay 17576
	integer		a, len, i		!! posibilidades
	real		z			!! (bastante) equiprobables

	z = 0
	do i=1, len(x)
		z = 1.25736*z+ichar(x(i:i))
	end do
	a = mod(int(300*z),17576)
	m(3:3) = char(ichar('A') + mod(a,26))
	a = a/26
	m(2:2) = char(ichar('A') + mod(a,26))
	a = a/26
	m(1:1) = char(ichar('A') + mod(a,26))
	end
c	.....................................................
c	:::::::::::::::::::::::::::::::::::::::::::::::::::::
C	This function obtains some data related to the thermal park
C	and writes the old-style .trm and .cmb files for future reading.
C	 - the .trm file includes maintenances
C	 - the .cmb file includes a lower bound for the use of fuel
C
C	'ftrm' and 'fcmb' are the names to be used for the files. 
C.........................................................
	subroutine creat_trm_cmb(model, ftrm, fcmb, nint, iexit)
	implicit none
	character*(*)	model, ftrm, fcmb
	integer		nint, iexit
	include 'HTC_parameter.h'
	include 'store_trm.h'

	real	Infty
	parameter (Infty=9.E18)

	character	tipo
	character*3	iden
	character*12	comb(MAXCOMB),    nom_thu(MAXTERM),
     >			thu_man(MANTEN)
	real		maxpow(MAXTERM),  reliab(MAXTERM), 
     >			PEprice(MAXTERM), opcst(MAXTERM),
     >			HCons_0(MAXTERM), HCons_1(MAXTERM),
     >			HCons_2(MAXTERM), 
     >			efic(MAXTERM),    Cst_gen(MAXTERM),
     >			bl_fuel(MAXCOMB), bu_fuel(MAXCOMB),
     >			preci_f(MAXCOMB), pmx_man(MANTEN)
	integer		com_thu(MAXTERM), orden(MAXTERM),
     >			from(MANTEN),      until(MANTEN)
	logical		is_contr(MAXTERM)
	integer		N_comb, N_comb_not_used, N_unit, N_man, i, j, k
	real		PRICE_Ex, prcmedio, mn, MPW

	call get_thermal_park(N_unit, N_comb, N_man, is_contr, maxpow,  
     >		minpow, HCons_2, HCons_1, HCons_0, PEprice, opcst,    
     >		com_thu, bl_fuel, bu_fuel, nom_thu, reliab, comb,   
     >		thu_man, from, until, pmx_man, PRICE_Ex, model, iexit)
	if (iexit.ne.0) return

	do i=1, N_unit
		MPW = maxpow(i)
		efic(i) = 3600*MPW/
     >		      ((HCons_2(i)*MPW + HCons_1(i))*MPW + HCons_0(i))
		Cst_gen(i) = 3.6*PEprice(i)/efic(i) + opcst(i)
	end do
C
C	O r d e n   d e   m e r i t o
C
	do i=1, N_unit
		orden(i) = i
	end do

	do i=1, N_unit-1
	    mn = Cst_gen(orden(i))
	    do j = i+1, N_unit
		if (Cst_gen(orden(j)) .lt. mn) then
		    k = orden(j)
		    mn = Cst_gen(k)
		    orden(j) = orden(i)
		    orden(i) = k
		end if
	    end do
	end do

	N_comb_not_used= 0
	do i=1, N_comb
	    k=0
	    prcmedio = 0.0
	    do j= 1, N_unit
		if (com_thu(j) .eq. i) then
		    k = k+1
		    prcmedio = prcmedio + Cst_gen(j)*efic(j)
		end if
	    end do
	    if (k.ne.0) then
	       preci_f(i) = prcmedio/k
	    else
	       preci_f(i)= -1.0
	       N_comb_not_used= N_comb_not_used+1
	    end if
	end do

!	call concat(fcmb, model, '.cmb')
!	call concat(ftrm, model, '.trm')
	open(unit=10, file=fcmb, status='unknown')
	open(unit=11, file=ftrm, status='unknown')
	write(10, *) N_comb-N_comb_not_used
	do i=1, N_comb
	   if (preci_f(i).ge.0.0d0) then
	      write(10,96) preci_f(i), 1000/3.6*bu_fuel(i), 1.0, 
     >	                   comb(i), 1000/3.6*bl_fuel(i)
	   end if
	end do
	write(10,*) PRICE_Ex

	write(11, *) N_unit, N_Man
	do i= 1, N_unit
	    k = orden(i)
	    tipo = 'T'
	    if (is_contr(k)) tipo = 'C'
	    call identif(nom_thu(k), iden)
	    write(11,97) tipo, iden, maxpow(k),  
     >			reliab(k), nom_thu(k), 1
	    if (.not. is_contr(k)) then
		write(11,98) comb(com_thu(k)), preci_f(
     >			com_thu(k))/Cst_gen(k), Infty, Infty
	    else
		write(11,*) 0.0, Cst_gen(k), 1, nint
	    end if
	end do
	do i= 1, N_Man
	    write(11,99) thu_man(i), pmx_man(i), from(i), until(i)
	end do
	write(11,*) 'Maintenances:'
	write(11,*) ' * Name of unit; * maximum power; * initial interv;'
     >		    , ' * final interv.'
	close(10)
	close(11)

96	format(x, f11.5, x, e13.7, x, f5.3, x, a12, x, e13.7)
97	format(x, a1, x, a3, x, f7.1, x, f7.3, x, a12, i1)
98	format(x, a12, x, f9.7, x, e13.7, x, e13.7)
99	format(x, a12, x, f7.1, x, i4, x, i4)

	end
C::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
	subroutine get_thermal_park(N_unit, N_comb, N_man, is_contr,  
     >	     maxpow, minpow, HCons_2, HCons_1, HCons_0, PEprice, opcst,    
     >	     com_thu, bl_fuel, bu_fuel, nom_thu, reliab, comb,  
     >	     thu_man, from, until, pmx_man, pr_ext, model, iexit)

	implicit none

	integer		N_unit, N_comb, N_man, com_thu(*), from(*), 
     >			until(*), iexit
	real		maxpow(*), reliab(*), PEprice(*), HCons_0(*), 
     >			HCons_1(*), HCons_2(*), opcst(*), bl_fuel(*), 
     >			bu_fuel(*), pmx_man(*), minpow(*), pr_ext
	logical		is_contr(*)
	character*(*)	comb(*), nom_thu(*), thu_man(*), model

c	local variables

	real		PRICE_Ex
	parameter	(PRICE_Ex=1000.0)
	character*12	fuel, rien
	character*40	fich
	character*80	line
	integer		i, input, search_for
	logical		esc, fin
	real		z

C	..... Asume que los datos estan en un fichero '<model>.thun' .....

	iexit = 0
	pr_ext = PRICE_Ex
	call concat(fich, model, '.thun')
	input = 10
	open(unit=input, file=fich, status='old', err=888)
	fin = .false.
	call skip_rems(input, line, fin)
	call skip_rems(input, line, fin)
	call skip_rems(input, line, fin)
	call skip_rems(input, line, fin)
C
C	L e c t u r a   d e   c o m b u s t i b l e s
C
	esc = .false.
	i = 0
	do while (.not. esc)
		i = i+1
		read(line, *) comb(i), bl_fuel(i), bu_fuel(i)
		read(input, 10) line
		if (index(line,'%') .ne. 0) esc = .true.
	end do
	N_comb = i
	call skip_rems(input, line, fin)
C
C	L e c t u r a   d e   u n i d a d e s
C
	esc = .false.
	i = 0
	do while (.not. esc)
		i = i+1
		read(line, *) nom_thu(i), minpow(i), maxpow(i), 
     >			      reliab(i), fuel
		com_thu(i) = search_for(fuel, comb, N_comb)
		read(input, 10) line
		if (index(line,'%') .ne. 0) esc = .true.
		is_contr(i)= .false.
	end do
	N_unit = i
	call skip_rems(input, line, fin)
C
C	L e c t u r a   d e   m a n t e n i m i e n t o s
C
	esc = .false.
	i = 0
	do while (.not. esc)
		i = i+1
		read(line, *) thu_man(i), z, pmx_man(i), 
     >				from(i), until(i)
		read(input, 10) line
		if (index(line,'%') .ne. 0) esc = .true.
	end do
	N_man = i
	call skip_rems(input, line, fin)
C
C	L e c t u r a   d e   p r e c i o s   e n e r g i a
C
	esc = .false.
	i = 0
	do while (.not. esc)
		i = i+1
		read(line, *) rien, PEprice(i)
		read(input, 10) line
		if (index(line,'%') .ne. 0) esc = .true.
	end do
	call skip_rems(input, line, fin)
C
C	L e c t u r a   d e   c o s t e s   d e   o p e r a c i o n
C
	esc = .false.
	i = 0
	do while (.not. esc)
		i = i+1
		read(line, *) rien, z,z,z, opcst(i)
		read(input, 10) line
		if (index(line,'%') .ne. 0) esc = .true.
	end do
	call skip_rems(input, line, fin)
C
C	L e c t u r a   d e   H e a t   C o n s u m p t i o n
C
	esc = .false.
	i = 0
	do while (.not. esc)
		i = i+1
		read(line, *) rien, HCons_2(i), HCons_1(i), HCons_0(i)
		read(input, 10) line
		if (index(line,'%') .ne. 0) esc = .true.
	end do
	call skip_rems(input, line, fin)

	close(input)
10	format((a))
	return
888	continue !!!!write(*,*) '[get_thermal_park] File not found:', fich
        iexit=-28
	return
	end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
	subroutine bring_minpower(p, n)
	implicit none

	real*8	p(*)
	integer	n, i

	include 'HTC_parameter.h'
	include 'store_trm.h'

	do i=1, n
		p(i) = minpow(i)
	end do

	end
