#define Hm3Day_m3Sec (double)11.574074
#define MAXZ 15000
#define MAXRES 200
#define PATHDades "./"
#define FEBRERO 1
#define DICIEMBRE 11
#define Next_Month(m) ((m+1)%12)
#define Next_Year_maybe(a,m) (m==DICIEMBRE?a+1:a)
#define MIN(x,y) ((x>y)?y:x)
#define XOR(a,b) ((a && !b) || (!a && b))
#define BISIESTO(y) (((y) % 4==0 && (y) % 100) || (y) % 400==0)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "structs.h"

/*...............................................................
 *
 *	Interface with the functions dealing with inflows management
 *	We need:
 *		- name of the case (basin)
 *		- date of first day of the study
 *		- array with number of days of each interval
 *		- list of dates for first day of inflows for the
 *		  history of each reservoir
 *		- number of commodities and probabilities
 *	In addition, there must exist a file containing historical inflows
 *	for each reservoir (in the future, this will be replaced with calls
 *	to a BD).
 *	The output is an array with quantiles for each reservoir, interval
 *	and commodity (in this order), but is not returned via parameter;
 *	instead, the result is kept in a private variable activated by
 *	"gimme_pointer" for future use.
 */

get_cdf_inflows(cca, date1, days, Ni, date2, nat, pr)
char		*cca;
Fecha		*date1, date2[];
int		days[], *Ni, *nat;
double		pr[];
{
	Sgle_int	*IN;
	Fecha		*e, *f, *inf_beg;
	int		i, j, cpm, D, Nres, Nz, tam, iexit;
	double		*x, *ptr_X;
	float	 	z[MAXZ];
	char		dest[80], aux[40];
	embalse	 	reser[MAXRES];

	IN = (Sgle_int *)malloc( sizeof(Sgle_int) * (1+(*Ni)) );
	IN[0].f.d = date1->d;
	IN[0].f.m = date1->m;
	IN[0].f.a = date1->a;
	e = &(IN[0].f);
	for (i=0; i<*Ni; i++) {		/*--- Builds list of dates ---*/
	    IN[i].dr = cpm = days[i];
	    f = &(IN[i+1].f);
	    memcpy((char*)f, (char*)e, sizeof(Fecha));
	    D=dias_del_mes(*e);			/* Find out beginning */
	    while (e->d+cpm > D) {		/* of next interval */
		cpm -= D - e->d + 1;
		f->a = Next_Year_maybe(e->a, e->m);
		f->m = Next_Month(e->m);
		f->d = 1;
		D=dias_del_mes(*f);
		e = f;
	    }
	    f->d += cpm;
	    e->d = f->d;
	}

	get_fconcaname(cca, dest);
	if (iexit=lect_conca_1(dest, reser, &Nres, *nat)) return iexit;
	if (Nres > MAXRES) {
	    /* fprintf(stderr, "[get_cdf_inflows] Too many reservoirs\n"); */
	    return -20 ;
	}
	tam = Nres*(*nat)*(*Ni);
	gimme_pointer(&x, &tam);
	ptr_X = x;
	for (i=0; i<Nres; i++) {
	    inf_beg = date2+i;

	    sprintf(aux, "%s.dat", reser[i].nombre);
	    if (iexit=read_inflows(aux, z, &Nz, MAXZ)) return iexit;
	    reparte_inflows(z, Nz, IN, *Ni, *inf_beg);
	    test_for_negativity(IN, *Ni, reser[i].mnemo);

	    for (j=0; j<*Ni; j++)
	        convert(IN[j].muestra, IN[j].n, Hm3Day_m3Sec/IN[j].dr);

	    quantile_solver(IN, *Ni, ptr_X, *nat, pr, i);
/***	    if (write_quant(ptr_X, *nat, *Ni, reser[i].mnemo, mstr, (int)IN[0].f.m))
		exit(-1);
***/
	    for (j=0; j<*Ni; j++) free(IN[j].muestra);
	    ptr_X += (*nat)*(*Ni);
	}
	free(IN);
	return 0;	/* normal exit */
}

/** Interface with Fortran **/
get_cdf_inflows_(cca, date1, days, Ni, date2, nat, pr)
char		*cca;
Fecha		*date1, date2[];
int		days[], *Ni, *nat;
double		pr[];
{
	return get_cdf_inflows(cca, date1, days, Ni, date2, nat, pr);
}

get_fconcaname(cca, dest)
char	*cca, *dest;
{
	char	*z2;

	for (z2 = cca; isalpha(*z2); z2++);
	strcpy(dest, PATHDades);
/*	strncat(dest, cca, (int)(z2-cca));   
	strcat(dest,"/");
*/	strcat(dest, cca);
	strcat(dest, ".hydro");
}

/**** Quantiles in array X are organized according to:
      1st interval ->  x(0), x(01), x(02), ..., x(0K-1)
	...
      Ni-th interv ->  x(0), x(01), x(02), ..., x(0K-1)
***/

/*...............................................................
 *   Input: the interval decomposition				(w)
 *	    # of intervals					(Ni)
 *	    # of commodities					(nat)
 *	    probabilities per block				(pr)
 *          reservoir: 0 ... Nreservoir-1			(r)
 *   Output: array of quantiles (see above)			(x)
 *
 *   This function finds a block distribution for the samples of each 
 *   interval according to the parameters nat and pr.
 */

quantile_solver(w, Ni, x, nat, pr, r)
Sgle_int	w[];
int		Ni, nat, r;
double		x[], pr[];
{
	int	i, n, k=0, *nroom, *ka, *ha, *hs, nwcore, done;
	double	delta, aux, *rroom, *a, *bl, *bu, *pi, *rc, *z;

	nroom = (int *)malloc(7*nat*sizeof(int));
	nwcore = (2000*nat);
	rroom = (double *)malloc((11*nat+4+nwcore)*sizeof(double));
	ka = nroom;
	ha = ka+nat+1;		/* allocates several arrays in a */
	hs = ha+4*nat-2;	/* single vector (room) */
	a = rroom;
	bl = a+4*nat-2;
	bu = bl+2*nat+1;
	pi = bu+2*nat+1;
	rc = pi+nat+1;
	z = rc+2*nat+1;
	for (i=0; i<Ni; i++) {
	    n = w[i].n;
	    qwantilae_ordena(w[i].muestra, n, x+k, pr, nat);
	    delta = (double)(nat-1)/(n+1);
	    aux = (1.0+delta)*x[k] - delta*x[k+1];
	    x[k+nat-1]  = (1.0+delta)*x[k+nat-1] - delta*x[k+nat-2];
	    x[k] = aux;

	    quant_estimator_(w[i].muestra, &n, &nat, pr, x+k, ka, 
			a, ha, bl, bu, hs, pi, rc, z, &nwcore, &done);

	    /* if (!done)
		fprintf(stderr, "(Weak estimation performed for (R,I)=(%d,%d))\n", r, i); */
	    k += nat;
	}
	free(nroom);
	free(rroom);
}

test_for_negativity(interv, Ni, mnemo)
int		Ni;
Sgle_int	interv[];
char		*mnemo;
{
	int	i, m, g;
	float	x;

/*-----	printf("%s ", mnemo);	-----*/
	for (i=0; i<Ni; i++) {
	    x = 0.0, g = 0;
	    for (m=0; m<interv[i].n; m++)
		if (interv[i].muestra[m] < 0.0) {
			x -= interv[i].muestra[m], g++;
			interv[i].muestra[m] = 0.0;
		}
/*-----	    if (g>0) printf("%4.2f ", x/g);
	    else printf(".... ");		-----*/
	}
/*-----	putchar('\n');	-----*/
}

convert(x, n, Cte)
float	x[];
double 	Cte;
int	n;
{
	int	i;
	for (i=0; i<n; i++) x[i] *= Cte;
}

write_quant(x, nat, Ni, mnemo, mstr, mes)
double	x[];
int	nat, Ni, mstr, mes;
char	*mnemo;
{
	char	nam[15];
	FILE	*f;
	int	i, j, l, up;

	sprintf(nam, "%s%d%02d%02d.apo", mnemo, nat, mstr, mes+1);
	f = fopen(nam, "w");
	if (!f) {
		fprintf(stderr, "Can't save file %s\n", nam);
		return -1;
	}
	for (i=l=0; i<Ni; i++) {
	    up = 0;
	    for (j=0; j<nat; ) {
		up += 6;
		for ( ; j < up && j < nat; j++, l++)
			fprintf(f, "%12.4lf", j? x[l]-x[l-1]: x[l]);
		putc('\n',f);
	    }
	}
	fclose(f);
	return 0;
}

netoyer(linea)		/* removes CR, initial blanks and so on */
char	*linea;
{
	char	*p;

	p = linea+strlen(linea)-1;
	while (isspace(*p)) {
		*p = '\0';
		p--;
	}
	p = linea + strspn(linea, " \t");
	if (p!=linea) memmove(linea, p, strlen(p)+1);
}

/*...............................................................
 *   Input: name of the file that contains the info		(arc)
 *          number of commodities				(nart)
 *   Output: array of reservoirs'info				(emb)
 *           number of reservoirs				(n_emb)
 *...............................................................
 */ 

lect_conca_1(arc, emb, n_emb, nart)	/* returns names and mnemotecnics */
char	*arc;				/*	of reservoirs         */
embalse  emb[];
int  	*n_emb, nart;
{
	int i, n;
	char linea[132], *ptr;
	FILE 	*fi;

	fi = fopen(arc, "r");
	if (!fi) {
		/* fprintf(stderr, "Can't read file %s\n", arc); */
		return(-19);
	}
	fscanf(fi,"%d\n", n_emb);
	for (i=0; i<*n_emb; i++) {
	     fgets(linea, 132, fi);
	       netoyer(linea);
	       emb[i].nombre = malloc( strlen(linea)+1 );
	       strcpy(emb[i].nombre, linea);
	     fgets(linea, 132, fi);
	       ptr = linea;
               while (isspace(*ptr)) ptr++;
	       strncpy(emb[i].mnemo, ptr, TAM_MNMO);
	       emb[i].mnemo[TAM_MNMO] = '\0';
	     for(n=0; n<nart; n+=7) fgets(linea, 132, fi);
	     fgets(linea, 132, fi);
	     fgets(linea, 132, fi);
	}
	fclose(fi);
	return(0);
}

swap(p1, p2, n)
void 	*p1, *p2;
int	n;
{
	char	buf[512], *k;

	if (n<512) {
		memcpy(buf, (char*)p2, n);
		memcpy((char*)p2, (char*)p1, n);
		memcpy((char*)p1, buf, n);
	} else {
		k = malloc(n);
		memcpy(k, (char*)p2, n);
		memcpy((char*)p2, (char*)p1, n);
		memcpy((char*)p1, k, n);
		free(k);
	}
}

limpia(linea)		/* removes chars != ' ' or [digits] */
char	*linea;
{
	char	*p, *q;

	p = q = linea;
	while (*p) {
		if (*p==' ' || isdigit(*p))
			*q++ = *p;
		else if (*p=='/' || *p=='-') 
			*q++ = ' ';
		p++;
	}
	*q = '\0';
}

static int lmes[12] = {31,28,31,30,31,30,31,31,30,31,30,31};

dias_del_mes(f)
Fecha	f;
{
	if (f.m!=FEBRERO) return lmes[f.m];
	else if ( BISIESTO(f.a) ) return 29;
	     else return 28;
}

fin_de_mes(f)			/* days to end of the month */
Fecha	f;
{
	int	d_mes;

	d_mes = dias_del_mes(f);
	return d_mes - f.d + 1;
}

/*...............................................................
 *   Generates description of intervals with 'date first day' and
 *   'duration in days', taking into account the date of the beginning 
 *   and the length of the period. Just one month appears per interval;
 *   the weekday is the same of the weekday of the beginning, unless 
 *   the interval begins at the first of the month.
 *
 *   Input: date of the beginning of the period under study  (inicio)
 *          length in days of the period                        (tot)
 *   Output: array of intervals                          (intervalos)
 *           number of suitable intervals found                  (Ni)
 *...............................................................
 */ 

init_terv_periodo(intervalos, inicio, tot, Ni)
Sgle_int	intervalos[];
Fecha		*inicio;	
int		tot, *Ni;	/* tot: total dias periodo; Ni: # intervalos */
{
	int	i=0, fin=0, faltan;
	Sgle_int	*pi=intervalos;
	Fecha	g;

	pi->f.d = inicio->d;
	pi->f.m = inicio->m;
	pi->f.a = inicio->a;
	while (!fin) {
	    faltan = fin_de_mes(pi->f);
	    if (faltan < 7 && faltan < tot) {
		pi->dr = faltan;
		tot -= pi->dr;
		pi++;
		i++;
		pi->f.d = 1;
		pi->f.m = Next_Month((pi-1)->f.m);
		pi->f.a = Next_Year_maybe((pi-1)->f.a,(pi-1)->f.m);
		pi->dr = MIN(tot,7-faltan);
		tot -= pi->dr;
		pi++;
		i++;
		if (tot) {
			pi->f.d = (pi-1)->f.d + (pi-1)->dr;
			pi->f.m = (pi-1)->f.m;
			pi->f.a = (pi-1)->f.a;
		}
	    } else if (faltan < 7 && faltan >= tot) {
		pi->dr = tot;
		tot = 0;
		i++;
	    } else {			/* faltan >= 7 */ 
		pi->dr = MIN(7,tot);
		tot -= pi->dr;
		pi++;
		i++;
		if (tot && faltan > 7) {
			pi->f.d = (pi-1)->f.d+7;
			pi->f.m = (pi-1)->f.m;
			pi->f.a = (pi-1)->f.a;
		} else {
			pi->f.d = 1;
			pi->f.m = Next_Month((pi-1)->f.m);
			pi->f.a = Next_Year_maybe((pi-1)->f.a,(pi-1)->f.m);
		}
	    }
	    fin = tot==0;
	}
	*Ni = i;
}

get_fdatename(f, nm)		/* 'intvalsMMDDYY.lis' */
Fecha	f;
char	nm[];
{
	sprintf(nm, "intvals%02hd%02hd%02hd.lis", f.m+1, f.d, f.a%100);
}

/*...............................................................
 *   creates a file containing the interval decomposition of the period.
 *   Input: array of intervals 					(x)
 *          # of intervals 					(Ni)
 *	    name of the file to be created			(name)
 *   no output parameters; writes in the file.
 *...............................................................
 */

save_array_in(name, x, Ni)
char		name[];
Sgle_int	x[];
int		Ni;
{
	FILE	*f;
	int	i;

	f = fopen(name, "w");
	if (f==0) {
		/* fprintf(stderr, "Can't save file %s\n", name); */
		return -22;
	}
	for (i=0; i<Ni; i++)
		fprintf(f,"%2hd/%2hd/%4hd %hd\n", 
			x[i].f.m+1, x[i].f.d, x[i].f.a, x[i].dr);
	fclose(f);
	return 0;
}

/*...............................................................
 *   reads from a file containing an interval decomposition of 
 *   the period (possibly, a user-defined decomp.).
 *   Each line must contain one date such as MM/DD/YYYY and the
 *   duration in days (some symbols may be removed). The date has
 *   to be likely, and the whole period has to be continuous.
 *
 *   Input: name of the input file				(d)
 *   Output: array of intervals 				(x)
 *          # of intervals 					(Ni)
 *...............................................................
 */

read_array_in(d, x, Ni)
char		*d;
Sgle_int	x[];
int		*Ni;
{
	FILE	*f;
	int	i=0, z, error=0;
	char	linea[80];

	f = fopen(d, "r");
	if (f==0) {
	/*	fprintf(stderr, "Can't read file %s\n", d); */
		return -23;
	}
	fgets(linea, 80, f);
	while (!feof(f)) {
		limpia(linea);
		sscanf(linea,"%hd%hd%hd%hd",
			&x[i].f.m, &x[i].f.d, &x[i].f.a, &x[i].dr);
		x[i].f.m--;
		if (x[i].f.m < 0 || x[i].f.m > 11 || x[i].f.d < 1 ||
			x[i].f.d > dias_del_mes(x[i].f) || x[i].dr <= 0)
			printf("Wrong line: # %d\n", i+1), error=-1;
		if (i>0) {
		    z = seek_complete(x[i-1].f, x[i].f);
		    if (z != x[i-1].dr) 
			/*fprintf(stderr, "Warning: dates don't match with length (line %d)\n", i), */ error=-24;
		}
		fgets(linea, 80, f);
		i++;
	}
	fclose(f);
	*Ni = i;
	return error;
}

/*...............................................................
 *   load data from a file (one datum per line). The inflows are
 *   intended to be *** daily Hm^3 ***
 *
 *   Input: name of the input file				(d)
 *   Output: array of inflows 					(z)
 *          # of days loaded 					(Nz)
 *          size of z { Nz <= }					(maxz)
 *...............................................................
 */

read_inflows(d, z, Nz, maxz)
char	*d;
float	z[];
int	*Nz, maxz;
{
	FILE	*f;
	int	i=0, error=0;
	char	linea[80];

	f = fopen(d, "r");
	if (f==0) {
	/*	fprintf(stderr, "Can't read file %s\n", d); */
		return -12;
	}
	fgets(linea, 80, f);
	while (!feof(f) && i<maxz) {
		sscanf(linea, "%f", z+i);
		fgets(linea, 80, f);
		i++;
	}
	if (!feof(f)) {
	/*	fprintf(stderr, "Not enough space reading inflows from file %s\n", d); */
		error = -21;
	}
	*Nz = i;
	fclose(f);
	return error;
}

date_less_equal(d1, d2)
Fecha	d1, d2;
{
	if (d1.m < d2.m) return 1;
	if (d1.m == d2.m && d1.d <= d2.d) return 1;
	return 0;
}

seek(d1, d2)				/* how many days between d1 and d2 */
Fecha	d1, d2;				/* (doesn't mind of years) */
{
	int	p;

	p = -d1.d;
	if (date_less_equal(d1, d2)) {
		while (d1.m < d2.m) {
			p += dias_del_mes(d1);
			d1.m++;
		}
		p += d2.d;
	} else {
		while (d1.m <= DICIEMBRE) {
			p += dias_del_mes(d1);
			d1.m++;
		}
		d1.m = 0;
		d1.a++;
		while (d1.m < d2.m) {
			p += dias_del_mes(d1);
			d1.m++;
		}
		p += d2.d;
	}
	return p;
}

seek_complete(d1, d2)			/* how many days between d1 and d2 */
Fecha	d1, d2;				/* (considers years) */
{
	int	p, sgn=1, y, prev_feb;

	if ((d1.a > d2.a) || (d1.a==d2.a && !date_less_equal(d1, d2)))
		swap((void*)&d1, (void*)&d2, sizeof(Fecha)), sgn = -1;

	p = seek(d1, d2);
	if (date_less_equal(d1, d2)) y = d1.a;
	else y = d1.a+1;
	prev_feb = d2.d <=28 && d2.m <= FEBRERO;
	while (y < d2.a) {
		if (BISIESTO(y) && !prev_feb) p += 366;
		else if (BISIESTO(y-1) && prev_feb) p += 366;
		else p += 365;
		y++;
	}
	return sgn*p;
}

float accumulate(z, Nz, p, f, m, ok)	/* returns the amount of m values */
float	z[];				/* z[p], ..., z[p+m-1], and */
int	Nz, *p, m, *ok;			/* updates pointer p and date f; */
Fecha	*f;				/* ok = true if p+m <= Nz */
{
	float	x=0.0;
	int	cpm=m, D;

	for (; *p<Nz && m>0; (*p)++, m--) x += z[*p];
	*ok = m==0;
	if (*ok==0) 
		cpm -= m;

	D=dias_del_mes(*f);
	while (f->d+cpm > D) {
		cpm -= D - f->d + 1;
		f->a = Next_Year_maybe(f->a, f->m);
		f->m = Next_Month(f->m);
		f->d = 1;
		D=dias_del_mes(*f);
	}
	f->d += cpm;
	return x;
}

float skip_29_feb(z, Nz, p, f, m, ok)	/* accumulate skipping a Feb-29 */
float	z[];				/* present in z[] but not useful */  
int	Nz, *p, m, *ok;			/* for the present interval */ 
Fecha	*f;				
{
	Fecha	next;
	int	k;
	float	x;

	next.d = 29;
	next.m = FEBRERO;
	if (BISIESTO(f->a) && f->m<=FEBRERO) next.a = f->a;
	else next.a = 4*(f->a/4+1);
	if (next.a%100==0 && next.a%400) next.a +=4;
	k = seek_complete(*f, next);
	x = accumulate(z, Nz, p, f, k, ok);
	if (*ok) {
	    (*p)++;
	    f->d = 1;
	    f->m ++;
	    x += accumulate(z, Nz, p, f, m-k, ok);
	}
	return x;
}

float add_29_feb(z, Nz, p, f, m, ok)	/* accumulate "including" a value */
float	z[];				/* for Feb-29 made with values of */
int	Nz, *p, m, *ok;			/* Feb-28 and Mar-1 */
Fecha	*f;				
{
	Fecha	mar1;
	int	k, p0=*p;
	float	x;

	mar1.d = 1;
	mar1.m = FEBRERO+1;
	mar1.a = f->m <= FEBRERO ? f->a : f->a+1;
	k = seek_complete(*f, mar1);
	x = accumulate(z, Nz, p, f, m-1, ok);
	if (ok)
	    x += 0.5*(z[p0+k-1]+z[p0+k]);
	return x;
}

int got_29_feb(f, n)		/* Do have we 29 Feb from this date (f) */
Fecha	f;			/* (n) days on? */
int	n;
{
	Fecha	next;

	next.d = 29;
	next.m = FEBRERO;
	if (BISIESTO(f.a) && f.m<=FEBRERO) next.a = f.a;
	else next.a = 4*(f.a/4+1);
	if (next.a%100==0 && next.a%400) next.a +=4;

	return (n > seek_complete(f, next));
}

/*...............................................................
 *   Distributes the data contained in z[] into the intervals; each one
 *   receives the sum of (dr) inflows from different years. Considers
 *   special cases related to Feb/29.
 *   Input: array with (daily) historical records		(z)
 *          # of records					(Nz)
 *          intervals (just beginning and duration)		(interv)
 *          # of intervals					(Ni)
 *          date of first inflow in z[]				(inf_b)
 *   Output: samples for each interval			(interv[].muestra)
 *           and size					(interv[].n)
 *...............................................................
 */

reparte_inflows(z, Nz, interv, Ni, inf_b)
float		z[];
int		Nz, Ni;
Sgle_int	interv[];
Fecha		inf_b;
{
	int	i, j, ny, p, q, ok, in_29, f_29;
	Fecha	f, F;
	float	tot;

	ny = (Nz/365+2)*sizeof(float);
	for (i=0; i<Ni; i++) {
		interv[i].muestra = (float *) malloc(ny);
		interv[i].n = 0;
	}

	for (i=0; i<Ni; i++) {
	    F.d = interv[i].f.d;
	    F.m = interv[i].f.m;
	    F.a = inf_b.a;
	    in_29 = got_29_feb(interv[i].f, interv[i].dr);
		/*** in_29 =
			present interval includes the damn day.
			(examp. Feb-15-1996 ... Mar-15-1996)
		***/
	    if (F.m < inf_b.m || (F.m==inf_b.m && F.d < inf_b.d)) F.a++;
	    q = seek_complete(inf_b, F);
	    j = 0;
	    ok = 1;
	    while (ok) {
		memcpy((char*)&f, (char*)&F, sizeof(Fecha));
		f_29 = got_29_feb(f, interv[i].dr);
		/*** f_29 =
			scanned data at this point includes the damn day.
			(examp. Feb-15-1980 ... Mar-15-1980)
		***/
		if (!XOR(in_29, f_29)) {
		    p = q;
		    tot = accumulate(z, Nz, &p, &f, interv[i].dr, &ok);
		} else if (f_29) {
		    p = q;
		    tot = skip_29_feb(z, Nz, &p, &f, interv[i].dr, &ok);
		} else if (in_29) {
		    p = q;
		    tot = add_29_feb(z, Nz, &p, &f, interv[i].dr, &ok);
		}
		if (ok) {
		    interv[i].muestra[j++] = tot;
		    F.a++;
		    if ( (BISIESTO(F.a) && F.m>FEBRERO) ||
		         (BISIESTO(F.a-1) && F.m<=FEBRERO) ) 
		             q += 366;
		    else
		             q += 365;
		}
	    }
	    interv[i].n = j;
	}
}
