#define TRUE 1
#define FALSE 0
#define ABC ('z' - 'a' +1)

#define MU 2.722222

#define IBDPOU  0
#define NOUPOU  -1
#define POU     "ZZZ"

#define SPILL    0
#define DESC     1
#define BOMBEIG  2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Funcio cota-volum */
typedef struct{
          double a0,a1,a2,a3,a4;
	  int    tiene;
}CV;

/* Funcio d'energia */

typedef struct{
          double a0,a1,a2,a3,a4,a5;
}EH;

/* Funcio de rendiment */
/*   Coeficients :   b  -> basic;
                     nd -> creuat cabal-alcada 
                     n  -> lineal d'alcada
                     d  -> lineal de cabal
                     nn -> quadratic d'alcada
		     dd -> quadratic de cabal
*/
typedef struct{
          double b,nd,n,d,nn,dd;
}Rendiment;


/* H(s) pels arcs de bombeig */
typedef struct{
          double a0,a1,a2,a3,a4,a5,a6;
}HS;


/*****************************************************************************/
/*****************************************************************************/

/* Retorna la posicio dins el vector d'embassaments 'codires' de l'embassament 
   'codres' */

int busca_res(codires,codres,nres)
int *codires,codres,nres;
{
  int i;

  i=0;
  while(i<nres && codires[i]!=codres)
    i++;

  return i;
}

/* Passa a minuscules str2 retornant-la per str1 */

void minuscules(str1,str2)
char *str1,*str2;
{
  int i;
  
  for(i=0;i<=strlen(str2);i++){
    if(str2[i]>='A' && str2[i]<='Z')
      str1[i]=str2[i]+'a'-'A';
    else str1[i]=str2[i];
  }
  str1[i]='\0';
}


/* Escriu el numeros negatius en el format de FORTRAN E13.7: -1.988705E-05 ==>
		"-.1988705E-04"			        pos:  012345678901 */

void format(output,linia)
FILE *output;
char *linia;
{
  int j;
  char lin2[3];

  fprintf(output,"-.%c",linia[0]);
  for(j=2;j<10;j++)
    fprintf(output,"%c",linia[j]);
  lin2[0]=linia[10];
  lin2[1]=linia[11];
  lin2[2]='\0';
  j= ( linia[9]=='-' ? atoi(lin2)-1 : atoi(lin2)+1 );
  fprintf(output,"%.2d",j);
}

acortar(lin)		/* elimina los espacios que hay por la cola */
char	*lin;
{
	int len, i;
	len = strlen(lin);
	for (i=len-1; i>=0 && isspace(lin[i]); i--);
	lin[i+1] = '\0';
}

generar_id(idres, i, lin)	/* asignar un identificador de 3 letras */
char	*idres[], *lin;		/* al nodo "i"; los "i-1" restantes ya */
int	i;			/* estan asignados */
{
	int	k, repetir, h, j;
	char	*p, *x;
	k = 0;
	for (p=lin; *p; p++) {
	   k += (*p) & 31;
	   k = k << 5;
	}
	x = idres[i];
	x[3] = '\0';
	x[2] = 'a' + k%ABC;
	k /= ABC;
	x[1] = 'a' + k%ABC;
	k /= ABC;
	x[0] = 'a' + k%ABC;
	h = 0;
	repetir = TRUE;		/* Eliminacion de repeticiones */
	while (repetir) {
	    for (j=0; j<i && !strstr(idres[j],x); j++);
	    if (j<i) 
		x[2] = '0' + h++;
	    else
		repetir = FALSE;
	}
}

int contar_repes(tipus, codini, i)
int	tipus[], codini[], i;
{
	int	j, m=0;

	for (j=0; j<i; j++)
		if (tipus[j]==tipus[i] && codini[j]==codini[i]) m++;
	return m;
}

/*===================================================================*/
/*  here we have some global variables from this point on. The 
    variables are initialized by "deal_with_basin" and can be read
    from the outer only through special functions. These variables
    are:
	+ Initial storage for each reservoir
	+ Final storage for each reservoir
	+ 4th degree head-volume coefficient per reservoir
	+ polynomials of efficiency for each arc of discharge or pumping
	+ functions of jump to be surmounted by pumping
	+ Distributions of natural inflows (comm/reser/interval)
	  (this one is initialized by "get_cdf_inflows")
    Functions:
	+  gimme_inivol()
	+  gimme_finvol()
	+  gimme_4th_cv()
	+  gimme_eff_polyn()
	+  gimme_jump_pump()
	+  gimme_inflows()
	+  release_bckp()   ---> frees space of all arrays
....................................................................*/

double	*volini, *volfinal0, *cv4g, *rnd_gb, *apors;
HS 	*arcbomb;

gimme_inivol(x, MaxNat, Nat, Neb)
double	*x;
int	*MaxNat, *Nat, *Neb;
{
	int	j, k, l;

	l = 0;
	for (j=0; j<*Neb; j++) {
	    x[l] = volini[j];
	    for (k=1; k<*Nat; k++)
		x[l+k] = 0.0;
	    l += *MaxNat;
	}
}

gimme_finvol(x, MaxNat, Nat, Neb)
double	*x;
int	*MaxNat, *Nat, *Neb;
{
	int	j, k, l;

	l = 0;
	for (j=0; j<*Neb; j++) {
	    x[l] = volfinal0[j];
	    for (k=1; k<*Nat; k++)
		x[l+k] = 0.0;
	    l += *MaxNat;
	}
}

gimme_4th_cv(x, n)
double	*x;
int	*n;
{
	int	i;
	for (i=0; i<*n; i++)
		x[i] = cv4g[i];
}

gimme_eff_polyn(x5, x4, x3, x2, x1, x0, n)
double	*x5, *x4, *x3, *x2, *x1, *x0;
int	*n;
{
	int	i, j;
	j = 0;
	for (i=0; i<*n; i++) {
		x5[i] = rnd_gb[j++];
		x4[i] = rnd_gb[j++];
		x3[i] = rnd_gb[j++];
		x2[i] = rnd_gb[j++];
		x1[i] = rnd_gb[j++];
		x0[i] = rnd_gb[j++];
	}
}

gimme_jump_pump(x, n, bomb, snq)
double	*x, *snq;
int	*n, bomb[];
{
	int	i, j, k;

	for (i=k=0, j=0; i<*n; i++) 
	    if (bomb[i]) {	/* bomb[i] != 0  <=>  pumping channel */
		bomb[i] = j+1;
		x[j++] = arcbomb[k].a1;	  /* linear term upstream */
		x[j++] = arcbomb[k].a2;	  /* quadr. term upstream */
		x[j++] = arcbomb[k].a3;	  /* cubic  term upstream */
		x[j++] = -arcbomb[k].a4;  /* linear term downstream */
		x[j++] = -arcbomb[k].a5;  /* quadr. term downstream */
		x[j++] = -arcbomb[k].a6;  /* cubic  term downstream */
		snq[i] = arcbomb[k++].a0;	/* independent term */
	    }
	/* At exit, bomb[] points to the first coeff. of the pumping arc */
}

gimme_pointer(ptr, siz)
double	**ptr;
int	*siz;
{
	*ptr = apors = (double *)malloc(sizeof(double)*(*siz));
}

gimme_inflows(x, MaxNat, MaxEmb, Nat, Neb, Nint)
double	*x;
int	*MaxNat, *MaxEmb, *Nat, *Neb, *Nint;
{
	int	i, j, k, l, m;

	for (i=0; i<*Nint; i++)
	    for (j=0; j<*Neb; j++)
		for (k=0; k<*Nat; k++) {
		    l = *MaxNat*(*MaxEmb*i+j) + k;
		    m = *Nat*(*Nint*j+i) + k;
		    if (k)
			x[l] = apors[m] - apors[m-1];
		    else
			x[l] = apors[m];
		}
}

release_bckp()
{
	free(apors);
	free(volini);
	free(cv4g);
	free(rnd_gb);
	free(volfinal0);
	free(arcbomb);
}
/*****************************************************************/


deal_with_basin(fibd, nart, nres, narcs)
char	*fibd;
int	nart, nres, narcs;
{
  FILE 		*input, *output;
  char 		filemb[80], sort[80], linia[100], *sitio, *finids, **nomres, 
		**idres, *descar="DEFG", *bombeo="BCXW", *vertid="VOPQ";
  char 		tipo;
  CV 		*cotavol, *cvres;
  EH 		*energia;
  int 		i, j, font, desti, dummy;
  int 		*codires, *codini, *codesti, *tipus;
  double 	*sbx, *maxflow, *minflow, *maxpw, *minpw, *volmin, *volmax, 
		*cap_efect, *cotamax, *volfinal1;
  double 	x1, h, eh, Maxvol, phd, alfa1=0.0;
  Rendiment 	rend;

  sitio=(char *)malloc(500*nres);
  nomres=(char **)malloc(sizeof(char *)*nres);
  idres=(char **)malloc(sizeof(char *)*nres);
  codires=(int *)malloc(sizeof(int)*nres);
  volmin=(double *)malloc(sizeof(double)*nres);
  volmax=(double *)malloc(sizeof(double)*nres);
  cap_efect=(double *)malloc(sizeof(double)*nres);
  cotamax=(double *)malloc(sizeof(double)*nres);
  volini=(double *)malloc(sizeof(double)*nres);
  volfinal0=(double *)malloc(sizeof(double)*nres);
  volfinal1=(double *)malloc(sizeof(double)*nres);
  cvres=(CV *)malloc(sizeof(CV)*nres);
  codini=(int *)malloc(sizeof(int)*narcs);
  codesti=(int *)malloc(sizeof(int)*narcs);
  tipus=(int *)malloc(sizeof(int)*narcs);
  maxflow=(double *)malloc(sizeof(double)*narcs);
  minflow=(double *)malloc(sizeof(double)*narcs);
  maxpw=(double *)malloc(sizeof(double)*narcs);
  minpw=(double *)malloc(sizeof(double)*narcs);
  cotavol=(CV *)malloc(sizeof(CV)*narcs);
  energia=(EH *)malloc(sizeof(EH)*narcs);
  arcbomb=(HS *)malloc(sizeof(HS)*narcs);
  sbx=(double *)malloc(sizeof(double)*narcs);
  cv4g=(double *)malloc(sizeof(double)*nres);
  rnd_gb=(double *)malloc(sizeof(double)*narcs*6);
  for (i=0, finids=sitio; i<nres; i++, finids+=4)
	idres[i] = finids;

/*******   G E T   D A T A   F R O M   ... *******/

  sprintf(filemb, "%s.net", fibd);
  if ((input=fopen(filemb,"r"))==NULL){
    /** printf("(***) File %s doesn't exist. \n", filemb); **/
    return -25;
  }
  
  fgets(linia,100,input);
  while(strchr(linia,'%')!=NULL)	/* skip comments */
    fgets(linia,100,input);

  sscanf(linia,"%lf",&phd);
/*  fscanf(input,"%d",&nres); */
  fscanf(input,"%d",&dummy);
/*  fscanf(input,"%d \n",&narcs); */
  fscanf(input,"%d\n",&dummy);

  for (i = 0; i < nres; i++) {
	cvres[i].tiene = 0;
	fgets(linia, 100, input);
	while (strchr(linia, '%') != NULL)
		fgets(linia, 100, input);

	acortar(linia);
	nomres[i] = finids;
	finids += strlen(linia) + 1;
	strcpy(nomres[i], linia);
	generar_id(idres, i, linia);

	fscanf(input, "%d", &codires[i]);
	fscanf(input, "%lf%lf%lf", &volmin[i], &volmax[i], &cap_efect[i]);
	fscanf(input, "%lf", &cotamax[i]);
	fscanf(input, "%lf%lf%lf\n", &volini[i], &volfinal0[i], &volfinal1[i]);
	fgets(linia, 100, input);	/* Marg. cost */
  }

  for (i = j = 0; i < narcs; i++) {
	fgets(linia, 100, input);
	while (strchr(linia, '%') != NULL)
		fgets(linia, 100, input);
	fgets(linia, 100, input);

	fscanf(input, "%d%d\n", &codini[i], &codesti[i]);
	fscanf(input, "%d\n", &tipus[i]);
	if (tipus[i] == BOMBEIG) {
		dummy = codini[i];
		codini[i] = codesti[i];
		codesti[i] = dummy;
	}
	if (codesti[i]==IBDPOU) codesti[i]=NOUPOU;

	fscanf(input, "%lf%lf\n", &maxflow[i], &minflow[i]);
	if (tipus[i] == DESC || tipus[i] == BOMBEIG) {
		if (tipus[i] == DESC)
		       fscanf(input, "%lf", &energia[i].a5);
		fscanf(input, "%lf", &energia[i].a4);
		fscanf(input, "%lf", &energia[i].a3);
		fscanf(input, "%lf", &energia[i].a2);
		fscanf(input, "%lf", &energia[i].a1);
		fscanf(input, "%lf", &energia[i].a0);
	}
	if (tipus[i] == DESC) {
		fscanf(input, "%lf", &cotavol[i].a4);
		fscanf(input, "%lf", &cotavol[i].a3);
		fscanf(input, "%lf", &cotavol[i].a2);
		fscanf(input, "%lf", &cotavol[i].a1);
		fscanf(input, "%lf", &cotavol[i].a0);
	}
	if (tipus[i] == BOMBEIG) {
		fscanf(input, "%lf", &arcbomb[j].a6);
		fscanf(input, "%lf", &arcbomb[j].a5);
		fscanf(input, "%lf", &arcbomb[j].a4);
		fscanf(input, "%lf", &arcbomb[j].a3);
		fscanf(input, "%lf", &arcbomb[j].a2);
		fscanf(input, "%lf", &arcbomb[j].a1);
		fscanf(input, "%lf", &arcbomb[j++].a0);
	}
	if (tipus[i] == DESC || tipus[i] == BOMBEIG)
		fscanf(input, "%lf%lf\n", &maxpw[i], &minpw[i]);
  }
  fclose(input);

/*******   E N D   O F   R E T R I E V I N G   O F   D A T A   *******/

/***************************************************************************/
		/* THE FILE <BASIN>.hydro IS CREATED */

  sprintf(sort,"%s.hydro",fibd);
  if((output=fopen(sort,"w"))==NULL){
    /** printf("(***) Couldn't open file %s\n", sort); **/
    return (-26);
  }
  
  for (i=0;i<narcs;i++){
    codini[i]=busca_res(codires,codini[i],nres);
    if(codesti[i]!=NOUPOU)
        codesti[i]=busca_res(codires,codesti[i],nres);
    if (codini[i] >= nres || codesti[i] >= nres) 
        return (-29); 

    if(tipus[i]==DESC){
      Maxvol = volmax[codini[i]];
      sbx[i] = cotavol[i].a0 + Maxvol*(cotavol[i].a1 + Maxvol*
	   (cotavol[i].a2 + Maxvol*(cotavol[i].a3 + Maxvol*cotavol[i].a4)));
/*      cvres[codini[i]].a0=cotamax[codini[i]]-sbx[i];	*/
      cvres[codini[i]].a0= cotavol[i].a0 + cotamax[codini[i]]-sbx[i];
      cvres[codini[i]].a1= cotavol[i].a1;
      cvres[codini[i]].a2= cotavol[i].a2;
      cvres[codini[i]].a3= cotavol[i].a3;
      cvres[codini[i]].a4= cotavol[i].a4;
      cvres[codini[i]].tiene ++;
    }
    else sbx[i]=0.0;
  }

  fprintf(output,"%d\n",nres);
  for(i=0;i<nres;i++){
    fprintf(output," %s \n",nomres[i]);
    fprintf(output," %.3s %10.4f%10.4f %.2d %.2d \n",idres[i],volmax[i],
	      cap_efect[i],0,0);

    fprintf(output," %10.4f%10.4f%10.4f%10.4f%10.4f%10.4f%10.4f \n",volmax[i],
	    volmax[i],volmax[i],volmax[i],volmax[i],volmax[i],volmax[i]);

    if (cvres[i].tiene == 0) {
	/* printf("(---) %s doesn't have any arc of discharge\n", nomres[i]); */
	cvres[i].a0 = cotamax[i];
    }
    if (cvres[i].tiene > 1) {
	/* printf("(---) %s has more than 1 (%d) arc of discharge\n", nomres[i], cvres[i].tiene); */
    }
    if(cvres[i].a0>=0)
      fprintf(output," %8.7E",cvres[i].a0);
    else{
      sprintf(linia,"%7.6E",-cvres[i].a0);
      fprintf(output," ");
      format(output,linia);
    }
    if(cvres[i].a1>=0)
      fprintf(output," %8.7E",cvres[i].a1);
    else{
      sprintf(linia,"%7.6E",-cvres[i].a1);
      fprintf(output," ");
      format(output,linia);
    }
    if(cvres[i].a2>=0)
      fprintf(output," %8.7E",cvres[i].a2);
    else{
      sprintf(linia,"%7.6E",-cvres[i].a2);
      fprintf(output," ");
      format(output,linia);
    }
    if(cvres[i].a3>=0)
      fprintf(output," %8.7E",cvres[i].a3);
    else{
      sprintf(linia,"%7.6E",-cvres[i].a3);
      fprintf(output," ");
      format(output,linia);
    }
    fprintf(output,"\n");	/** Solo escribimos hasta 3er grado **/
/***    if(cvres[i].a4>=0)
      fprintf(output," %8.7E\n",cvres[i].a4);
    else{
      sprintf(linia,"%7.6E",-cvres[i].a4);
      fprintf(output," ");
      format(output,linia);
      fprintf(output,"\n");
    }
**/
    fprintf(output," %10.4f \n",cotamax[i]);
  }

  for (i = 0; i < nres; i++)
 	cv4g[i] = cvres[i].a4;
  j = 0;
  for (i = 0; i < narcs; i++) {
 	if (tipus[i] == DESC) {
 		rnd_gb[j++] = energia[i].a5;
		rnd_gb[j++] = energia[i].a4;
		rnd_gb[j++] = energia[i].a3;
		rnd_gb[j++] = energia[i].a2;
		rnd_gb[j++] = energia[i].a1;
		rnd_gb[j++] = energia[i].a0;
	}
 	if (tipus[i] == BOMBEIG) {
		rnd_gb[j++] = 0.0;
 		rnd_gb[j++] = energia[i].a4;
		rnd_gb[j++] = energia[i].a3;
		rnd_gb[j++] = energia[i].a2;
		rnd_gb[j++] = energia[i].a1;
		rnd_gb[j++] = energia[i].a0;
	}
  }

  fprintf(output,"%d \n",narcs);

  for(i=0;i<narcs;i++){
    j = contar_repes(tipus, codini, i);
    if(tipus[i]==DESC)    tipo = descar[j];
    if(tipus[i]==BOMBEIG) tipo = bombeo[j]; 
    if(tipus[i]==SPILL)   tipo = vertid[j];
    if((maxflow[i]=maxflow[i]*1.0E6/3600)>100000) maxflow[i]=99999;
    minflow[i]=minflow[i]*1.0E6/3600;
    if(codesti[i]!=NOUPOU)
      fprintf(output," %c %.3s %.3s %.5d %10.4f%10.4f %.2d %.2d\n",tipo,
	      idres[codini[i]],idres[codesti[i]],0,maxflow[i],minflow[i],0,0);
    else
      fprintf(output," %c %.3s %.3s %.5d %10.4f%10.4f %.2d %.2d \n",tipo,
	      idres[codini[i]],POU,0,maxflow[i],minflow[i],0,0);
    
    fprintf(output," %10.4f%10.4f%10.4f%10.4f%10.4f%10.4f%10.4f\n",maxflow[i],
	    maxflow[i],maxflow[i],maxflow[i],maxflow[i],maxflow[i],maxflow[i]);

    if (tipus[i]==DESC){
	rend.b=0.9;    /* This value has no future use */
	fprintf(output," %10.4f%10.4f \n",rend.b,sbx[i]);
    }

    if(tipus[i]==BOMBEIG)
	fprintf(output," %10.4f \n",-1.5);
  }
  fprintf(output,"%d \n",0);

  fclose(output);

  free(volmin);
  free(volmax);
  free(cap_efect);
  free(cotamax);
  free(volfinal1);
  free(codires);
  free(codini);
  free(codesti);
  free(tipus);
  free(maxflow);
  free(minflow);
  free(maxpw);
  free(minpw);
  free(cotavol);
  free(cvres);
  free(energia);
  free(sbx);
  free(idres);
  free(nomres);
  free(sitio);
  return 0;		/** Normal exit **/
}

/*****************************************************************
	esta funcion es provisional y no tiene mas objeto que 
	conseguir parametros iniciales de forma poco ortodoxa
*****************************************************************/
#include "structs.h"

do_magic(neb, nar, nat, init_date, inf_dtes)
int	*neb, *nar, *nat;
Fecha	*init_date, *inf_dtes[];
{
	int	i;
	char	input[80];
	Fecha	*x;

	printf("Nmero de embalses? ");
	gets(input);
	sscanf(input,"%d", neb);
	printf("Nmero de arcos? ");
	gets(input);
	sscanf(input,"%d", nar);
	printf("Nmero de artculos? ");
	gets(input);
	sscanf(input,"%d", nat);
	printf("Fecha inicial (d/m/a):");
	gets(input);
	sscanf(input,"%hd%hd%hd", &(init_date->d), &(init_date->m), &(init_date->a));
	init_date->m--;
	*inf_dtes = (Fecha *)malloc(sizeof(Fecha)*(*neb));
	printf("Fecha inicial aportaciones [0 si no son comunes] (d/m/a):");
	x = *inf_dtes;
	for (i=0; i<*neb; i++) {
	    gets(input);
	    sscanf(input,"%hd%hd%hd", &(x->d), &(x->m), &(x->a));
	    x->m--;
	    if (i==0 && x->d) {
		x++;
		x->d = NULL;		/* fechas comunes */
		break;
	    }
	    x++;
	}
	if (i==0 && !x->d) 	/* se ponen todas las fechas como la primera */
	x = *inf_dtes;
	for (i=1; i<*neb; i++) {
	    x++;
	    x->d = (*inf_dtes)->d;
	    x->m = (*inf_dtes)->m;
	    x->a = (*inf_dtes)->a;
	}
}

/*           Interface with Fortran */

gimme_inflows_(x, MaxEmb, MaxInt, Nat, Neb, Nint)
double	*x;
int	*MaxEmb, *MaxInt, *Nat, *Neb, *Nint;
{
	gimme_inflows(x, MaxEmb, MaxInt, Nat, Neb, Nint);
}

gimme_inivol_(x, MaxNat, Nat, Neb)
double	*x;
int	*MaxNat, *Nat, *Neb;
{
	gimme_inivol(x, MaxNat, Nat, Neb);
}

gimme_finvol_(x, MaxNat, Nat, Neb)
double	*x;
int	*MaxNat, *Nat, *Neb;
{
	gimme_finvol(x, MaxNat, Nat, Neb);
}

gimme_4th_cv_(x, n)
double	*x;
int	*n;
{
	gimme_4th_cv(x, n);
}

gimme_eff_polyn_(x5, x4, x3, x2, x1, x0, n)
double	*x5, *x4, *x3, *x2, *x1, *x0;
int	*n;
{
	gimme_eff_polyn(x5, x4, x3, x2, x1, x0, n);
}
gimme_jump_pump_(x, n, bomb, snq)
double	*x, *snq;
int	*n, bomb[];
{
  gimme_jump_pump(x, n, bomb, snq);
}
