c***********************************************************************
c     	Rutina ''ANALISIS_MONOTONAS''
c
c	Halla las escalas. Para un intervalo dado, halla la energa de 
c	la montona y las funciones de energa externa y tiempo mximo 
c	de utilizacin de la energa externa, en funcin de Ho.
!
C	Computes scales (power, time and energy). Finds the energy demand
C	for each subperiod, and the functions of external energy and
C	maximum duration of use of external energy, depending on Ho 
C	(deterministic hydrogeneration), besides of the capacity limits.
c***********************************************************************
	SUBROUTINE ANALISIS_MONOTONAS(e_potencia, e_horas, monot, NQT)

        implicit none
	include 'TadBezier.h'
	include 'varsbez.h'
	include 'mono.h'

	real *8 e_potencia, e_horas
	record /mon_carga/ monotona
	record /incres/ esc

        integer NQT, monot(NQT,*)
        integer*4 i, j, k

        integer*4 prob, v1, v2, forma, tipo
	logical	sirve(MAXPARQUE), fuera
        real *8 ppunta, nhores, epun, sumT
     >	     , z(6), w(6), ept, etm
     >	     , max_Ho, min_GT, tiempo_base

	common /sitio/ hd, trm, y, tmp, prob
	real *8 hd(MAXOBS), trm(MAXOBS), y(MAXOBS,MAXOBS), 
     >		tmp(MAXOBS,MAXOBS)

	v1=0
	v2=0
        DO i = 1, Nint
	   call NORMALIZA_ESCALAS
     >          ( xxx(i).Nhores, xxx(i).Ppunta, ept, etm )
           v1 = v1 + int(dlog10(ept))
           v2 = v2 + int(dlog10(etm))
        END DO
        e_potencia = dble(10**int(v1/Nint))
        e_horas = dble(10**int(v2/Nint))
c
c	Corregir las dimensiones de las cosas guardadas pa las termicas
c
	do j=1, planta.ntermicas
	   planta.potmax(j) = planta.potmax(j)/e_potencia
	   planta.ptlim(j) = planta.ptlim(j)/e_potencia
	   planta.limcomb(j) = planta.limcomb(j)/e_potencia/e_horas
	end do
	         
	sumT = 0.d0			! Total potencia del parque termico
	do k=1, planta.ntermicas
		sumT = sumT + planta.potmax(k)
	end do
C====
C       Esta rutina ya no sera necesaria, al crearse las formas a partir de
C		las cargas.
C	call leerformasmonotonas(fichero, monot, nqt, MAXQT)
C....
        DO i = 1, Nint
c
c	calcular unas escalas/cuadrito para esta monotona particular
c
	   ppunta = xxx(i).ppunta/e_potencia
	   nhores = xxx(i).nhores/e_horas
	   xxx(i).ppunta = ppunta
	   xxx(i).nhores = nhores
	   xxx(i).phdmax = xxx(i).phdmax/e_potencia
	   forma = xxx(i).nforma
	   esc.inc_p = ppunta / monot(1,forma)
	   esc.inc_t = nhores / nqt
	   esc.inc_e = esc.inc_p * esc.inc_t

!	   if (sumT .lt. ppunta) then
!     		write(*,*) '###### ',sumT,' < ',ppunta,' ######'
!		write(*,*) 'Intervalo ',i
!	   end if

	   j=0
	   do k=1, nqt
		j= j+monot(k,forma)
	   end do
	   epun = dble(j)*esc.inc_e
	   xxx(i).energia = epun

	   monotona.nqt = nqt
	   monotona.nqp = monot(1,forma)
	   monotona.np = int(xxx(i).Phdmax/esc.inc_p + 0.5)    
	   call calculo_filas(monotona,monot(1,forma))
	   xxx(i).pmin = monotona.pmin*esc.inc_p

	   max_Ho = xxx(i).phdmax*nhores*0.9
	   xxx(i).tbase = tiempo_base(monotona, esc, nhores, max_Ho
     >			, planta, xxx(i).phdmax)
	   max_Ho = xxx(i).phdmax*xxx(i).tbase
	   tipo = 0
	   min_GT = xxx(i).phdmax
	   xxx(i).min_GT = xxx(i).phdmax
	   do k=1, planta.ntermicas
		if (planta.es_ctr(k) .and. (planta.in_ctr(k).le.i
     >		                    .and. i.le.planta.fi_ctr(k)))
     >		    min_GT = min_GT + planta.potmax(k)
			! i no en [A,B] <-->  A > i || B < i
		fuera = planta.in_ctr(k).gt.i.or.planta.fi_ctr(k).lt.i
		sirve(k) = .not.(planta.es_ctr(k) .and. fuera)

		if (.not.planta.es_ctr(k) .and. planta.te_lim(k))
     >		    min_GT = min_GT + planta.potmax(k)-planta.ptlim(k)
	   end do

	   if (min_GT.gt.0.0) then
		call muestreo_1(min_GT)
	   	xxx(i).min_GT = min_GT
	   end if
	   IF (xxx(i).Phdmax .eq. 0.d0) then 
!
!		Si no hay hidraulica en el parque, hay que indicar
!		explicitamente POT.HIDR. = 0.0
!
	     xxx(i).max_Ho = 0.d0
	     if (min_GT .eq. 0.d0) then
	   	call observaciones_bis(monotona, esc, sumT, i)
	   	xxx(i).energy.a = 0.D0 
	   	xxx(i).energy.b = -1.D0	
	   	xxx(i).energy.c = Y(1,1)
	   	xxx(i).time.a = 0.D0 
	   	xxx(i).time.b = -1.D0	
	   	xxx(i).time.c = TMP(1,1)
	     else
		tipo = 2
		call observaciones(monotona, esc, i, .false., sirve)
	     end if
	   ELSE				!! T e n e m o s   h i d r o 

	     xxx(i).max_Ho = max_Ho
	     call muestreo_2(max_Ho)
	     if (min_GT .eq. xxx(i).Phdmax) then
		tipo = 1
		call observaciones_tris(monotona, esc, sumT, i)
	     else
		tipo = 3
		call observaciones(monotona, esc, i, .true., sirve)
	     end if
	   END IF

c	Resolver por mnimos cuadrados los dos ajustes
	   if (tipo .gt. 0) then
	      prob = 1					
	      call ajuste_conlim(z, prob, tipo, maxobs)

	      prob = 2				        
	      call ajuste_conlim(w, prob, tipo, maxobs)

	      xxx(i).energy.a = z(1)
	      xxx(i).energy.b = z(2)
	      xxx(i).energy.c = z(3)
	      xxx(i).time.a = w(1)
	      xxx(i).time.b = w(2)
	      xxx(i).time.c = w(3)
	   end if

	end do
      end
ccccccccccccccccccccccc
ccccccccccccccccccccccc
	real *8 function pendi(vecx,vecy,paso)
	implicit none

	real *8 vecx(*), vecy(*)
	integer paso, i

	i=1
	do while (vecy(i) .le. vecy(i+paso))
		i=i+1
	end do

	pendi = (vecy(i+paso)-vecy(i))/(vecx(i+paso)-vecx(i))
	end
ccccccccccccccccccccccc
ccccccccccccccccccccccc
	SUBROUTINE CARREGAR_BATERIES(NIN, NTERM, PHD, NFORMA, PPUN,
     >			NHORAS, PANAS, POTMAX, LIMCOM, ES_CTR, PTLIM,
     >			TE_LIM, IN_CTR, FI_CTR)

	implicit none
	include 'TadBezier.h'
	include 'varsbez.h'

	integer	nin, nterm, nforma(nin), i, in_ctr(nterm),
     >		fi_ctr(nterm)
	real *8	phd, nhoras(nin), ppun(nin), panas(nterm), 
     >		potmax(nterm), limcom(nterm), ptlim(nterm)
	logical es_ctr(nterm), te_lim(nterm)

	do i=1, nin
	   xxx(i).phdmax = phd
	   xxx(i).nforma = nforma(i)
	   xxx(i).ppunta = ppun(i)
	   xxx(i).nhores = nhoras(i)
	end do

	do i=1, nterm
	   planta.probpana(i) = panas(i)
	   planta.potmax(i) = potmax(i)
	   planta.limcomb(i) = limcom(i)
	   planta.ptlim(i) = ptlim(i)
	   planta.es_ctr(i) = es_ctr(i)
	   planta.in_ctr(i) = in_ctr(i)
	   planta.fi_ctr(i) = fi_ctr(i)
	   planta.te_lim(i) = te_lim(i)
	end do
	planta.ntermicas = nterm
	Nint = nin

	end
ccccccccccccccccc
c	Recuperar la curva i-sima
ccccccccccccccccc
	SUBROUTINE OBTENER_CURVA(I, CURVA)
	implicit none

	include 'TadBezier.h'
	include 'varsbez.h'
	integer i
	record /corbabez/ curva

        curva.nforma  = xxx(i).nforma
        curva.ppunta  = xxx(i).ppunta
        curva.nhores  = xxx(i).nhores
        curva.tbase  = xxx(i).tbase
        curva.phdmax  = xxx(i).phdmax
        curva.energia = xxx(i).energia
        curva.pmin    = xxx(i).pmin
	curva.energy.a = xxx(i).energy.a
	curva.energy.b = xxx(i).energy.b
	curva.energy.c = xxx(i).energy.c
	curva.time.a = xxx(i).time.a
	curva.time.b = xxx(i).time.b
	curva.time.c = xxx(i).time.c
	curva.max_Ho = xxx(i).max_Ho
	curva.min_GT = xxx(i).min_GT

      end !fi subroutine
c***********************************************************************
c     Lectura de monotonas			!!!! NOT USED ANYMORE
c***********************************************************************
	SUBROUTINE LEERFORMASMONOTONAS(FICHERO, MONOT, NQT, MAXQT)
        implicit none

c     
      integer*4 MAXQT, monot(MAXQT,*), nqt(*), nmonot, i, k

      character *(*) fichero
      character *100 pathformas

	call concat(pathformas, 
     >       '/users/onl4/term/4_artic/formas/', fichero)

	open(unit=11,name=pathformas,status='old')
	read(11,*) nmonot
	do i=1,nmonot
           read(11,*) nqt(i)
           read(11,*) (monot(k,i),k=1,nqt(i))
	end do
	close(11)
      
	end
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
	SUBROUTINE NORMALIZA_ESCALAS(T, PP, E_P, E_H)
	implicit none

	real *8 t, pp, e_p, e_h
	real *8 BASE, BASE10, x, y, z
	integer m, n

	BASE = 2.d1
	BASE10 = 2.d0
	x = t
	m = 0
	if (x .ge. BASE10) then
	    do while (x .ge. BASE)
		m = m+1
		x = x/1.d1
	    end do
	else
	    do while (x .lt. BASE10)
		m = m-1
		x = x*1.d1
	    end do
	end if
	y = pp
	n = 0
	if (y .ge. BASE10) then
	    do while (y .ge. BASE)
		n = n+1
		y = y/1.d1
	    end do
	else
	    do while (y .lt. BASE10)
		n = n-1
		y = y*1.d1
	    end do
	end if
c	{ t = x * 10^m   y   pp = y * 10^n,   con BASE10 <= x,y < BASE }

	z=x*y
	if (z .lt. BASE10) then
	    if (x .lt. y) m = m-1
	    if (x .ge. y) n = n-1
	else if (z .ge. BASE) then
	    if (x .lt. y) n = n+1
	    if (x .ge. y) m = m+1
	end if

	e_h = dble(10**m)
	e_p = dble(10**n)

	end

c2345678901c2345678901c2345678901c2345678901c2345678901c2345678901c2345678901
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cccccccccccccccccccccccccccccccccccccccccccccccc
c	Con esta rutina, hallamos una estimacin del tiempo base.
c	Estimacin, porque vara segun la Ho que ponemos como
c	pseudotrmica, pero como sabemos que lo hace proporcional
c	mente a la cantidad Ho, utilizaremos un valor esperado al
c	convolucionar la monotona ( y utilizaremos toda la trmica,
c	puesto que apenas tiene influencia )
C	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	REAL *8 FUNCTION TIEMPO_BASE(MONO, ESC, T, MAX_HO, PLANTA
     >			, PHDMAX)
	implicit none
	include 'TadBezier.h'
	include 'mono.h'

	record /mon_carga/ mono, mon2
	record /mon_carga_r/ mono_r
	record /incres/ esc
	record /park/ planta
	real *8 t, max_Ho, phdmax, pterm, hidro
	real *8 fallo(0:MAXPARQUE), ptmax(0:MAXPARQUE), 
     $    lcmb(0:MAXPARQUE)
	real *8 pot_termica
	integer i
	logical nul

	hidro = max_Ho/10.0
	pterm =  pot_termica( hidro, esc, mono, mon2, nul )

	fallo(planta.ntermicas) = 0.5d0
	lcmb(planta.ntermicas) = 9.9d9
	do i = 1, planta.ntermicas
		fallo(i-1) = planta.probpana(i)
		ptmax(i-1) = planta.potmax(i)
		lcmb(i-1) = planta.limcomb(i)
	end do
	
	ptmax(planta.ntermicas) = phdmax - hidro/t
	call panes_real(pterm, t, mon2, mono_r, planta.
     >               ntermicas+1, fallo, lcmb, ptmax, esc)
	tiempo_base = mono_r.filas(1) * esc.inc_t
	return

	end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccococo
c	Crear las variables independientes de Energia externa
c	y Tiempo externo ( o sea, sumT - suma(Tj) )
c	- - - - - - - - - - - - - - - - - - - - - - - - -
	SUBROUTINE MUESTREO_1(MINGT)
	implicit none

	include 'constbez.h'

	real *8 minGT, frac
	common /sitio/ hd, trm, y, tmp, prob
	real *8 hd(MAXOBS), trm(MAXOBS), y(MAXOBS,MAXOBS), 
     >		tmp(MAXOBS,MAXOBS)
	integer prob
	integer i

	do i=0, MAXOBS-1		!! 0 ... minGT/2
		frac = dble(i)/MAXOBS
		trm(i+1) = frac* minGT/2.0
	end do

	end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccococo
c	Crear las variables independientes de Energia externa
c	y Tiempo externo ( o sea, Ho )
c	- - - - - - - - - - - - - - - - - - - - - - - - -
	SUBROUTINE MUESTREO_2(MAX_Ho)
	implicit none

	include 'constbez.h'

	real *8 max_Ho, frac
	common /sitio/ hd, trm, y, tmp, prob
	real *8 hd(MAXOBS), trm(MAXOBS), y(MAXOBS,MAXOBS), 
     >		tmp(MAXOBS,MAXOBS)
	integer prob
	integer i

	do i=0, MAXOBS-1
		frac = dble(i)/MAXOBS
		hd(i+1) = frac* max_Ho
	end do

	end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccococo
c	Calcular los valores observados de Energia externa
c	y Tiempo externo (en funcion de las variables independientes)
c	- - - - - - - - - - - - - - - - - - - - - - - - -
	SUBROUTINE OBSERVACIONES(MONOTONA, ESC, IN, HIDRO, SIRVE)
	implicit none

	include 'TadBezier.h'
	include 'varsbez.h'
	include 'mono.h'

	record /incres/ esc
	common /sitio/ hd, trm, y, tmp, prob
	real *8 hd(MAXOBS), trm(MAXOBS), y(MAXOBS,MAXOBS), 
     >		tmp(MAXOBS,MAXOBS)
	integer prob
	integer in, i, j, d, m, nter
	record /mon_carga/ monotona, mon2
	record /mon_carga_r/ mono_r
	real *8 ptmax(0:MAXPARQUE, MAXOBS), r, pterm, pot_termica, w,
     >		fallo(0:MAXPARQUE), lcmb(0:MAXPARQUE), potalea, sumC, 
     >		integral_semi_discr, sumT, cof(MAXOBS)
	logical nul, hidro, sirve(*)

	sumT = 0.d0
	d = 0
	do i = 1, planta.ntermicas
	    if (sirve(i)) then
		d=d+1
		fallo(d) = planta.probpana(i)
		lcmb(d) = planta.limcomb(i)
		sumT = sumT+planta.potmax(i)
	    end if
	end do
	nter = d
	fallo(0) = 0.5d0
	lcmb(0) = 9.9d9
	r = 1.d0
	do j = 1, MAXOBS
	   d = 0
	   cof(j) = r
	   do i = 1, planta.ntermicas
	      if (sirve(i)) then
		 d=d+1
		 ptmax(d,j)=r*planta.potmax(i)+(1.d0-r)*planta.ptlim(i)
	      end if
	   end do
	   r = r-0.5d0/MAXOBS		!! minGT/2 ... minGT
	end do
 
	m = 1
	if (hidro) m = MAXOBS
	do i = 1, m
	     w = 0.d0
	     if (hidro) w = hd(i)
	     pterm =  POT_TERMICA( w, esc, monotona, mon2, nul )
	     potalea = xxx(in).phdmax - w/xxx(in).tbase
	     do j = 1, MAXOBS
		 ptmax(0, j) = cof(j)*potalea
		 sumC = sumT - trm(j) + ptmax(0,j)
		 if (hidro) then
		   call PANES_REAL(pterm, xxx(in).nhores, mon2, mono_r,
     >                nter+1, fallo, lcmb, ptmax(0,j), esc)
		 else
		   call PANES_REAL(pterm, xxx(in).nhores, mon2, mono_r,
     >                nter, fallo(1), lcmb(1), ptmax(1,j), esc)
		 end if
		 y(i,j)= INTEGRAL_SEMI_DISCR(sumC,pterm+sumC,mono_r,esc)
		 tmp(i,j) = mono_r.filas(mono_r.sigmat+1)*esc.inc_t
	     end do
	end do

	end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccococo
c	Una version que solo calcula sin hidraulica ni variacion de Y
c	- - - - - - - - - - - - - - - - - - - - - - - - -
	SUBROUTINE OBSERVACIONES_BIS(MONOTONA, ESC, SUMT, IN)
	implicit none

	include 'TadBezier.h'
	include 'varsbez.h'
	include 'mono.h'

	real *8 sumT
	record /incres/ esc
	common /sitio/ hd, trm, y, tmp, prob
	real *8 hd(MAXOBS), trm(MAXOBS), y(MAXOBS,MAXOBS), 
     >		tmp(MAXOBS,MAXOBS)
	integer prob
	integer in, i
	record /mon_carga/ monotona, mon2
	record /mon_carga_r/ mono_r
	real *8 ptmax(0:MAXPARQUE), pterm, pot_termica, 
     >	     fallo(0:MAXPARQUE), lcmb(0:MAXPARQUE), integral_semi_discr
	logical nul

	do i = 1, planta.ntermicas
		fallo(i) = planta.probpana(i)
		lcmb(i) = planta.limcomb(i)
		ptmax(i) = planta.potmax(i)
	end do

	pterm =  POT_TERMICA( 0.d0, esc, monotona, mon2, nul )
	call PANES_REAL(pterm, xxx(in).nhores, mon2, mono_r,
     >                planta.ntermicas, fallo, lcmb, ptmax(1), esc)
	y(1,1) = INTEGRAL_SEMI_DISCR(sumT,pterm+sumT,mono_r,esc)
	tmp(1,1) = mono_r.filas(mono_r.sigmat+1)*esc.inc_t

	end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccococo
c	En  esta version solo varia la hidro (y entran todas las
c	unidades termicas --no son contratos--
c	- - - - - - - - - - - - - - - - - - - - - - - - -
	SUBROUTINE OBSERVACIONES_TRIS(MONOTONA, ESC, SUMT, IN)
	implicit none

	include 'TadBezier.h'
	include 'varsbez.h'
	include 'mono.h'

	real *8 sumT
	record /incres/ esc
	common /sitio/ hd, trm, y, tmp, prob
	real *8 hd(MAXOBS), trm(MAXOBS), y(MAXOBS,MAXOBS), incg, poh,
     >		tmp(MAXOBS,MAXOBS)
	real*8	palea(MAXOBS)
	common /hasta2/ palea
	integer prob
	integer in, i, j
	record /mon_carga/ monotona, mon2
	record /mon_carga_r/ mono_r
	real *8 ptmax(0:MAXPARQUE), pterm, pot_termica, 
     >		fallo(0:MAXPARQUE), lcmb(0:MAXPARQUE), potalea, sumC, 
     >		integral_semi_discr
	logical nul

	do i = 1, planta.ntermicas
		fallo(i) = planta.probpana(i)
		lcmb(i) = planta.limcomb(i)
	end do
	fallo(0) = 0.5d0
	lcmb(0) = 9.9d9
	do i = 1, planta.ntermicas
	    ptmax(i) = planta.potmax(i)
	end do
 
	incg = xxx(in).phdmax/MAXOBS
	do i = 1, MAXOBS
	     pterm =  POT_TERMICA( hd(i), esc, monotona, mon2, nul )
	     potalea = xxx(in).phdmax - hd(i)/xxx(in).tbase
	     palea(i) = potalea
	     poh = 0.d0
	     j = 1
	     do while (poh .le. potalea)
	     	sumC = sumT + poh
	     	ptmax(0) = poh
	     	call PANES_REAL(pterm, xxx(in).nhores, mon2, mono_r,
     >                planta.ntermicas+1, fallo, lcmb, ptmax, esc)
	     	y(i,j) = INTEGRAL_SEMI_DISCR(sumC,pterm+sumC,mono_r,esc)
	     	tmp(i,j) = mono_r.filas(mono_r.sigmat+1)*esc.inc_t
		poh = poh+incg
		j = j+1
	     end do
	end do

	end
