!MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
!_,.-`"'-.,_,.-`"'-.,_,.-`"'-.,_,.-`"'-.,_,.-`"'-.,_,.-`"'-.,_,.-`"'-.,_
!WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
C
C	procedure for estimating quantiles. 
C	Input: sam, 'real' array with the sample
C	       nsam, size of sample
C	       k, # of commodities
C	       pr, probabilities of commodities
C	       xn, starting and reference point, provided by former estimation
C	       ka, a, ha, bl, bu, hs, pi, rc and z are arrays allocated by
C		the calling routines, but they do not contain nothing
C	       nwcore, size of z().
C	Output: xn, solution
C		ka, a, ha, bl, bu, hs, pi, rc and z has been already used
C		but they are not of interest (they can be disposed)
C		ok, 1 if no problems
Ccccccc

	subroutine quant_estimator(sam, nsam, k, pr, xn, ka, a, ha, 
     >			bl, bu, hs, pi, rc, z, nwcore, ok)
	implicit none

	integer	k, m, n, nb, ne, nncon, nnobj, nnjac, dummy, inform,
     >		mincor,  ns, ninf, nwcore, i, j, MAXART, nsam, ok,
     >          iobj, iaux
	integer	ka(*), ha(*), hs(*)
	real	sam(*), var, mean
	real*8	sinf, obj, epsi, Infty, objadd
	real*8	pr(*), a(*), bl(*), bu(*), xn(*), pi(*),
     >		rc(*), z(*)
	character*8	names(5)
	parameter (Infty = 1.d20, MAXART=20)
	real*8	ics(2*MAXART+1)
      integer*4 leniws, lencws
	parameter (leniws=1000, lencws=1000)
      character*8 cws(lencws)
      integer*4 iws(lencws), mincor2, mincor3
	external funcon, funobj_kwan

	common /com_qu_es/ x_init, prob, den, var
	real*8	x_init(MAXART), prob(0:MAXART), den(MAXART)
C	common /m1file/ iread, iprint, isumm
	integer iread, iprint, isumm

c
cc	Put data in common - - - - - - - - 
c
	do j=1, k
		x_init(j) = xn(j)
		ics(j) = xn(j)
		hs(j) = 0
	end do
	prob(0)=0.d0
	do j=1, k-1
		prob(j) = pr(j)
	end do
	prob(k)=0.d0
        do j=1, k
            if (dabs(xn(j)).gt.1.d0) then
			den(j) = dsqrt(1.d0/dabs(xn(j)))
	    else
			den(j) = 1.d0
	    end if
        end do

	names(1) = 'problem '
c	names(2) = 'lin.obj.'
c	names(3) = 'rhs     '
c	names(4) = 'ranges  '
c	names(5) = 'bounds  '

	call descriptiva(sam, %val(nsam), mean, var)
	if (var.lt.1.d-8) var = 1.0

	m = k
	n = k
	nb = n+m
	nncon = 0
	nnobj = n
	nnjac = 0
				!!!! LOWER BOUNDS: zero for all ...
	do i=1, n
		bl(i) = 0.d0
	end do
			        !!!! ... but for expectation constraint
	epsi = abs(mean)*2d-3
	bl(n+1) = (mean-epsi)
	do i=n+2, nb
		bl(i) = 0.d0
	end do
				!!!! UPPER BOUNDS: +oo for all ...
	do i=1, n
		bu(i) = Infty
	end do
	bu(n+1) = (mean+epsi)
	do i=n+2, nb
		bu(i) = Infty
	end do
				!!!! CONSTRAINT MATRIX
	ne = 0
	do i = 1, n
		ne = ne+1
		ka(i) = ne
		ha(ne) = 1
		a(ne) = 0.5*(prob(i-1)+prob(i))	!! coefs. for E[X]  constr.
		if (i.gt.1) then
			ne = ne+1
			ha(ne) = i
			a(ne) = 1.
		end if
		if (i.lt.n) then
			ne = ne+1
			ha(ne) = i+1
			a(ne) = -1.
		end if
	end do
	ka(n+1) = ne+1

	iread = 0
	iprint = 0
	isumm = 0
	call sninit(iprint, isumm, cws, lencws, iws, leniws, z, nwcore)

	iaux= 1
	iobj= 0
	objadd= 0.0d0
	call snopt( 'Cold', m, n, ne, iaux, nncon, nnobj, nnjac,
     >		    iobj, objadd, names(1), funcon, funobj_kwan, 
     >              a, ha, ka, bl, bu, dummy,
     >		    hs, ics, pi, rc, inform, mincor, mincor2, mincor3, 
     >		    ns, ninf, sinf, obj, cws, lencws, iws, leniws, 
     >              z, nwcore, cws, lencws, iws, leniws, z, nwcore )

	ok = 1
	if (inform .ne. 0 .and. inform .ne. 13) ok = 0

	do i= 1, n
		xn(i) = ics(i)
	end do
	end
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

	subroutine funobj_kwan( mode, n, x, f, g, nstate,
     >			cwsu, lencwsu, iwsu, leniwsu, rwsu, lenrwsu )

	implicit none
	double precision   f, x(*), g(*)
	integer		   n, mode, nstate
      integer*4 leniwsu, lencwsu, lenrwsu
      character*8 cwsu(lencwsu)
      integer*4 iwsu(lencwsu)
      real*8 rwsu(lenrwsu)

	real*8	root, aux, mu, v, penal
	integer	i, j, k, MAXART
	parameter (MAXART=20)
	common /com_qu_es/ x_init, pi, den, var
	real*8	x_init(MAXART), pi(0:MAXART), den(MAXART)
	real	var

	f = 0.d0
	do i=1, n
		root = den(i)
		aux = (x(i)-x_init(i))*root
		f = f + aux*aux
		g(i) = aux*root
	end do
	f = 0.5*f
	k=n			!!! PENALTY:  (Var[X] - s^2)^2 / s^2
	    mu = 0.d0
	    do j=1, k
		mu = mu + (pi(j-1)+pi(j))*x(j)
	    end do
	    v = 0.d0
	    do j=1, k-1
		v = v + pi(j)*(x(j+1)*(x(j+1)+x(j))+x(j)*x(j))
	    end do
	    penal = v/3.d0 - mu*mu/4.d0 - var
	    f = f + penal*penal/var
	    aux = 2.0*penal/var
	    g(1) = g(1) + aux*(pi(1)*(2.d0*x(1)+x(2))/3.d0 - 
     >				0.5d0*mu*pi(1))
	    do j=1, k-2
		    g(j+1) = g(j+1) + aux*((pi(j)*(2.d0*x(j+1)+x(j))
     >			     + pi(j+1)*(2.d0*x(j+1)+x(j+2)))/3.d0 
     >			     - 0.5d0*mu*(pi(j)+pi(j+1)))
	    end do
	    g(k) = g(k) + aux*(pi(k-1)*(2.d0*x(k)+x(k-1))/3.d0 - 
     >				0.5d0*mu*pi(k-1))

	end

	subroutine funcon()
	end
